/*----------------------------------------------------------------------
list.h - Header file for list.c, stacks and linked lists.

Author:     Michelle Yan
Revisions:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.
----------------------------------------------------------------------*/

#include <stdio.h>
#include <malloc.h>

#ifndef LINKED_LIST
#define LINKED_LIST


/*----------------------------------------------------------------------
Typedefs and prototypes for stack manipulation.       

The stack consists of a header of type stackType, which points to the
top and bottom of a singly linked list of nodes of type singleLinkType,
each of which points to a single item kept in the stack.  The header
also contains a pointer to the current node, which is used to scan
the stack without actually popping entries, and a count of the number
of items currently on the stack.
----------------------------------------------------------------------*/

/* A linked list node for a stack */
struct singleLinkType
{
  void *content;                /* Pointer to payload, or item stacked */
  struct singleLinkType *next;  /* Pointer to next link on stack */
};

typedef struct singleLinkType sLink;

/* Header for stack */
struct stackType
{
  sLink *top;       /* Pointer to the top of the stack */
  sLink *bottom;    /* Pointer to the bottom of the stack */
  sLink *current;   /* Pointer to the current element */
  int stackSize;    /* Number of elements in the stack */
};

typedef struct stackType stack_t;

/* Prototypes for functions to manipulate the stack */
void *LIST_newStack  ( void );
void  LIST_freeStack ( stack_t * );
void  LIST_resetStack( stack_t * );
void  LIST_pushStack ( stack_t *, void * );
void *LIST_popStack  ( stack_t * );
void *LIST_getStack  ( stack_t * );


/*----------------------------------------------------------------------
                      Doubly linked lists

This data structure consists of a header which points to the top and
bottom of a doubly linked list of nodes, each of which contains a 
pointer to the actual element in the list.  The header also contains
a pointer to a current element, which is used to scan the list without
modifying it, a count of the number of elements in the list, and pointers
to two functions to compare list elements and to free a list element.

----------------------------------------------------------------------*/

/* Typedefs for functions to compare list elements or free one */
typedef int  (*listelemcompfunc)( void *, void * );
typedef void (*listelemfreefunc)( void * );

/* Structure of a node in the doubly linked list */
struct doubleLinkType
{
  void *content;                /* Pointer to element, or payload */
  struct doubleLinkType *next;  /* Pointer to next node */
  struct doubleLinkType *prior; /* Pointer to previous node */
};

typedef struct doubleLinkType dLink;

/* Header of the doubly linked list */
struct listType
{
  dLink *top;                     /* Pointer to top (first) node */
  dLink *bottom;                  /* Pointer to bottom (last) node */
  dLink *current;                 /* Pointer to current node */
  int listSize;                   /* Number of elements in list */
  listelemcompfunc compare;       /* Function to compare elements */
  listelemfreefunc freeContent;   /* Function to free an element */
};

typedef struct listType list_t;

/* Function prototypes for doubly linked lists */
void *LIST_newList     ( listelemcompfunc, listelemfreefunc );
int   LIST_sizeList    ( list_t * );
void  LIST_topList     ( list_t * );
void  LIST_bottomList  ( list_t * );
int   LIST_isTopList   ( list_t * );
int   LIST_isBottomList( list_t * );
void  LIST_nextList    ( list_t * );
void  LIST_priorList   ( list_t * );
void *LIST_contentList ( list_t * );
void *LIST_contentAndNextList( list_t * );
int   LIST_findList    ( list_t *, void * );
void  LIST_insertList  ( list_t *, void * );
void *LIST_retrieveList( list_t * );
void  LIST_freeList    ( list_t * );

/*----------------------------------------------------------------------*/
/* Prototypes for miscellaneous functions.                              */
/*----------------------------------------------------------------------*/

int floatcmp( float *, float * );
int intcmp( int *, int * );
void defaultFreeContent( void * );
void nullFreeContent( void * );

#endif


