/*--------------------------------------------------------------------------
AdpKmean_3Dseg_Ebeta_test.c - Image segmentation by adaptive K-means

Programmer: Michelle Yan    
Revisions:  Dan Widyono, Paul Hughett                                      

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.
---------------------------------------------------------------------------*/


/*--------------------------------------------------------------------------
Includes the functions:
         adpKmean3D_1();                                                 
         adpKmean3D_2();                                                 

Some pre-RCS history:
  DRW 19980207: optimization
  DRW 19980220: kept interactive code in main(), placed rest in
      separate function which can in the future be called by Tcl/Tk
      or some other external program.
  PWH 19980310: Ported to Linux, fixed gcc warnings and errors.
     Also fixed apparent error of Gibbs3D = 3 for Gibbs3D == 3 in
     two conditional statements, now greppable with PWH 19880310.
--------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: AdpKmean_3Dseg_Ebeta_test.c,v 1.15 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";


/*--------------------------------------------------------------------------*/
/* Include proper constant, variable and function declarations              */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <string.h>
#include <math.h>
#include <stdlib.h>

#include "common.h"
#include "definition.h"
#include "image_io.h"
#include "memm.h"

#include "AdpKmean_3D.h"


/*-------------------------------------------------------------------------
Tuning parameters:

The following global variables are used to tune various aspects of the
segmentation algorithm to particular classes of images.  There are all
gathered here to help keep track of them, and also to strike fear into
the hearts of any persons that would dare to change this algorithm.

The code that chooses the values of these parameters is currently
scattered all over the source files but will eventually be gathered in
one place.  Note also that the user-provided values of Res_XY, Res_Z,
Gibbs3D, MinSp, NoRegion and ... should be (sort of) considered as
tuning parameters although they are currently passed as arguments.

Defining the tuning parameters as globals would be inappropriate in
subroutines for a multi-threaded environment, but that is not expected 
to be a common mode of operation for this program.  If that ever does
come to pass, the parameters can be bundled into a structure which is
passed as an argument.  It is more important now to get the tuning
parameters organized in some useful fashion.
--------------------------------------------------------------------------*/

/* User-provided parameters */
int echo_12;               /* Number of echoes: either 1 or 2 */
int NoRegion;              /* Number of regions (tissue types) */
int N, M;                  /* Number of rows and columns */
int Total;                 /* Number of slices in file */
int First, Last;           /* First and last slices to process (1-origin) */
float Res_XY, Res_Z;       /* Voxel dimensions */
char *Name1, *Name2;       /* File names for the two echoes */
int ImgType1, ImgType2;    /* Data types for the two echo files */
float STD_1, STD_2;        /* Standard deviation of image noise */
float Weight_1, Weight_2;  /* Relative weights for the two echoes */
int savemean;              /* Is a means image to be output? */

/* Derived parameters */
int SliceNo;               /* Number of slices to process */
int Gibbs3D = 0;           /* Aspect ratio of voxels */
int MinSp_XY;              /* Minimum half-width for local means */



/*--------------------------------------------------------------------------
adpKmean3D_1 - adaptively segment a single-echo MRI brain image

This source code segments an image corresponging to non-background into
several regions with an initial guess from K-mean. 

FIXME:  The arguments to this function need to be understood and
documented.                      
--------------------------------------------------------------------------*/

static void
adpKmean3D_1(
    float ***ImgP,	        /* Pointers to image and mask */
    unsigned char ***SegImgP,	/* Pointer to output image */
    int N,		        /* Number of rows, columns & slices */
    int M,
    int SliceNo,
    float ****AdpLocMn,         /* 4D array for local mean of each region */
    int *Sp_XY,	                /* Half window size of XY-, Z-axis */
    int *Sp_Z, 
    int MinSp,	                /* Minimum of half window size along XY */
    int NoRegion,	        /* Expected number of regions */
    int NoObj,	                /* Number of non-background pixels */
    float Res_XY,	        /* In-plane resolution & slice thickness */
    float Res_Z,
    float InVar,                /* ??? */
    int Gibbs3D )               /* Aspect ratio of voxels */
{
  int i;
  unsigned char	*Ptr_1;
  float *TotalMn, OldLike, NewLike, Gibbs=0.0;
  float Voxel;
  int   change_no, MinChange;
  int   k, IterNo, IterNo_Index;

    
  TotalMn = (float *) callock(NoRegion,sizeof(float));

  MinChange = SliceNo * max(N, M)/10.;
  *Sp_XY= (max(N, M) - 1)/2;	
  *Sp_Z = SliceNo - 1;

/*--------------------------------------------------------------------------*/
/* Refine segmentation of K_mean (ML)                                       */
/*--------------------------------------------------------------------------*/

  OldLike = Likelihood_1(ImgP, SegImgP, N, M, SliceNo, NoRegion, InVar);
  printf( "Likelihood from initial (ML) = %e\n", OldLike );

  Voxel = Res_XY * Res_XY * Res_Z;
  if ( Gibbs3D == 2 ) {
    /*	Gibbs = OldLike/(NoObj*Voxel*1.46); */
    Gibbs = OldLike/(NoObj*Voxel*1.2);
  }
  else if ( Gibbs3D == 1 ) {
    /*	Gibbs = OldLike/(NoObj*Voxel*2.); */
    Gibbs = OldLike/(NoObj*Voxel*2.1);
  }
  else {
    Gibbs = OldLike/(NoObj*Voxel*4.0);
  }

  do {
    *Sp_XY = *Sp_XY/2;
    *Sp_XY = max(*Sp_XY, MinSp);
    *Sp_Z = *Sp_Z/2;

    Create_LOTS (N, M, SliceNo, *Sp_XY, *Sp_XY, *Sp_Z);
    printf( "Update means with window size = %dX%dX%d (%e)\n", 
	   2*(*Sp_XY)+1, 2*(*Sp_XY)+1, 2*(*Sp_Z)+1, Gibbs); 

    IterNo = 0;

    if ( *Sp_XY <= MinSp && NoRegion == 5 && Gibbs3D == 3) {
      printf("Merging regions 2,3 -> 2 and 4,5 -> 3\n");
      for ( k=0; k<SliceNo; k++ ) { 
	Ptr_1=SegImgP[k][0];
	for ( i=N*M; i; i--, Ptr_1++ ) { 
	  if ( *Ptr_1 == 2 || *Ptr_1 == 3 )
	    *Ptr_1 = 2;
	  else if ( *Ptr_1 == 4 || *Ptr_1 == 5 ) 
	    *Ptr_1 = 3;
	}
      }
      NoRegion = 3;
    } 

    do {
      IterNo ++;

      Total_Mn(ImgP, SegImgP, N, M, SliceNo, TotalMn, NoRegion);
      Update_Mn (AdpLocMn, TotalMn, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP, N, M, SliceNo);

      IterNo_Index = 0;
      do { 
	IterNo_Index ++;
	change_no = Update_Seg1 (ImgP, SegImgP, N, M, SliceNo, 
				 AdpLocMn, TotalMn, *Sp_XY, *Sp_Z, 
				 NoRegion, Gibbs, InVar, Gibbs3D, 
				 &NewLike); 
	printf( "(%d--%d): Pixels changed (Likelihood) = %d  (%e)\n", 
	       IterNo, IterNo_Index, change_no, NewLike);
      } while ( Gibbs != 0.0 && change_no > MinChange && 
		IterNo_Index < 10);

    } while ( IterNo_Index > 1 && IterNo < 50);

    if ( *Sp_XY <= MinSp && NoRegion == 5 && Gibbs3D != 3 ) {
      printf("Merging regions 1,5 -> 1 and 3,4 -> 3\n");
      NoRegion = 3;
      for ( k=0; k<SliceNo; k++ ) { 
	for ( i=N*M, Ptr_1=SegImgP[k][0]; i; i--, Ptr_1++ ) { 
	  if ( *Ptr_1 == 1 || *Ptr_1 == 5 )
	    *Ptr_1 = 1;
	  else if ( *Ptr_1 == 3 || *Ptr_1 == 4 ) 
	    *Ptr_1 = 3;
	}
      } 

      IterNo ++; 

/*--------------------------------------------------------------------------*/
/*          Update means since regions are merged                           */
/*--------------------------------------------------------------------------*/

      Total_Mn(ImgP, SegImgP, N, M, SliceNo, TotalMn, NoRegion);
      Update_Mn (AdpLocMn, TotalMn, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP, N, M, SliceNo);

      IterNo_Index = 0;
      do { 
	IterNo_Index ++;
	change_no = Update_Seg1 (ImgP, SegImgP, N, M, SliceNo, 
				 AdpLocMn, TotalMn, *Sp_XY, *Sp_Z, 
				 NoRegion, Gibbs, InVar, Gibbs3D, 
				 &NewLike); 
	printf( "(%d--%d): Pixels changed (Likelihood) = %d  (%e)\n", 
	       IterNo, IterNo_Index, change_no, NewLike);
      } while ( Gibbs != 0.0 && change_no > MinChange && 
		IterNo_Index < 10); 

/*--------------------------------------------------------------------------*/
/*          Update means for output                                         */
/*--------------------------------------------------------------------------*/

      Total_Mn(ImgP, SegImgP, N, M, SliceNo, TotalMn, NoRegion);
      Update_Mn (AdpLocMn, TotalMn, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP, N, M, SliceNo);
    } 

    Gibbs = (Gibbs * NewLike)/OldLike;
    OldLike = NewLike;
    Free_LOTS ();
  } while ( *Sp_XY > MinSp );

  free(TotalMn);
} 



/*--------------------------------------------------------------------------
adpKmean3D_2 - adaptively segment a dual-echo brain image

This source code segments a dual-image corresponging to non-background
into several regions with an initial guess from K-mean.

FIXME:  The arguments to this function need to be understood and
documented.
--------------------------------------------------------------------------*/

static void
adpKmean3D_2(
    float ***ImgP_1,	      /* Pointers to image and mask */
    float ***ImgP_2,
    unsigned char ***SegImgP, /* Pointer to output image */
    int N,		      /* Number of rows and columns */
    int M,
    int SliceNo,              /* Number of slices??? */
    float ****AdpLocMn_1,     /* 4D array for local mean of each region*/
    float ****AdpLocMn_2,
    int *Sp_XY,	              /* Half window size of XY-, Z-axis */
    int *Sp_Z,
    int MinSp,	              /* Minimum of half window size of XY */
    int NoRegion,	      /* Expected number of regions */
    int NoObj, 	              /* Number of non-background pixels */ 
    float Res_XY,	      /* In-plane resolution & slice thickness */
    float Res_Z,
    float STD_1,              /* ??? */
    float STD_2,
    float Weight_1,           /* ??? */
    float Weight_2,
    int Gibbs3D )             /* ??? */
{
  int		i;
  unsigned char	*Ptr_1;
  float		tmpV_1, tmpV_2, minV_1, minV_2, *ptrV_1, *ptrV_2;

  float		 cosAngle, sinAngle, angle;
  float          *TotalMn_1, *TotalMn_2, OldLike, NewLike, 
                 Gibbs=NOT_UNINIT;
  float          Voxel, t2CSF_V, prMaxV;
  float          InVarW_1GW, InVarW_1C=NOT_UNINIT, InVarW_2C=NOT_UNINIT,
                 InVarW_2GW;
  float          csf_W1=NOT_UNINIT, csf_W2=NOT_UNINIT, DiffRatio;
  int            change_no, MinChange;
  int            k, IterNo, IterNo_Index;
  int            csfLabel=NOT_UNINIT;


  TotalMn_1 = (float *) callock(NoRegion,sizeof(float));
  TotalMn_2 = (float *) callock(NoRegion,sizeof(float));

  MinChange = SliceNo * max(N, M)/10.;
  *Sp_XY = (max(N, M) - 1)/2;
  *Sp_Z = SliceNo - 1;

  csf_W1 = csf_W2 = InVarW_1GW = InVarW_1C = InVarW_2C = InVarW_2GW = Gibbs = 0.0;


/*--------------------------------------------------------------------------*/
/* Compute weight for GM, WM and weight for CSF, Gibbs prior                */
/*--------------------------------------------------------------------------*/

  Voxel = Res_XY * Res_XY * Res_Z;
  InVarW_1GW = (0.5 / (STD_1*STD_1)) * Weight_1;
  InVarW_2GW = (0.5 / (STD_2*STD_2)) * Weight_2;

  if ( Gibbs3D == 6 && (int)(10*Res_XY+0.25) != 9 ) { 

    Total_Mn(ImgP_1, SegImgP, N, M, SliceNo, TotalMn_1, NoRegion);
    Total_Mn(ImgP_2, SegImgP, N, M, SliceNo, TotalMn_2, NoRegion);
    DiffRatio = abs( (TotalMn_1[1]-TotalMn_1[2])/
		     (TotalMn_2[1]-TotalMn_2[2]) );
    printf( "Difference ratio = %e \n", DiffRatio);

    if ( DiffRatio < 3.15 ) {    
      InVarW_1GW *= 1.70;
      InVarW_2GW *= 1.70;
      csf_W1 = 0.717;        /* overestimated cases                   */
      csf_W2 = 0.301;
    }
    else if ( DiffRatio >= 3.15 && DiffRatio < 11.9 ) {
      InVarW_1GW *= 1.55;
      InVarW_2GW *= 1.55;
      csf_W1 = 0.407;        /* good for mid range                    */
      csf_W2 = 0.270;
    }
    else if ( DiffRatio >= 11.9 ) {
      csf_W1 = 0.215;        /* underestimated cases                  */
      csf_W2 = 0.247;
    }
    InVarW_1C = (0.5 / (STD_1*STD_1)) * csf_W1;
    InVarW_2C = (0.5 / (STD_2*STD_2)) * csf_W2; 

    OldLike = Likelihood_2(ImgP_1, ImgP_2, SegImgP, N, M, SliceNo, NoRegion, 
			   InVarW_1GW, InVarW_2GW, InVarW_1C, InVarW_2C);
    printf( "Likelihood from initial (ML) = %e\n", OldLike);

    /*	Gibbs = OldLike/(NoObj*Voxel*7.0); */
    Gibbs = OldLike/(NoObj*Voxel*7.5);

  } /* Gibbs3D == 6 && (int)(10*Res_XY+0.25) != 9 */

  else if ( Gibbs3D == 6 && (int)(10*Res_XY+0.25) == 9 ) {

    csf_W1 = 0.40;             /* Finland protocol                      */
    csf_W2 = 0.60;
    InVarW_1C = (0.5 / (STD_1*STD_1)) * csf_W1;
    InVarW_2C = (0.5 / (STD_2*STD_2)) * csf_W2; 

    OldLike = Likelihood_2(ImgP_1, ImgP_2, SegImgP, N, M, SliceNo, NoRegion, 
			   InVarW_1GW, InVarW_2GW, InVarW_1C, InVarW_2C);
    printf( "Likelihood from initial (ML) = %e\n", OldLike);

    /*	Gibbs = OldLike/(NoObj*Voxel*5.5); */
    Gibbs = OldLike/(NoObj*Voxel*5.0);

  } /*  Gibbs3D == 6 && (int)(10*Res_XY+0.25) == 9 */

  else if ( Gibbs3D == 5 ) {

    csf_W1 = 0.30;             /* Brookhaven protocol                   */
    csf_W2 = 0.70; 
    InVarW_1C = (0.5 / (STD_1*STD_1)) * csf_W1;
    InVarW_2C = (0.5 / (STD_2*STD_2)) * csf_W2; 

    OldLike = Likelihood_2(ImgP_1, ImgP_2, SegImgP, N, M, SliceNo, NoRegion, 
			   InVarW_1GW, InVarW_2GW, InVarW_1C, InVarW_2C);
    printf( "Likelihood from initial (ML) = %e\n", OldLike);

    Gibbs = OldLike/(NoObj*Voxel*7.0);

  } /* Gibbs3D == 5 */

  else if ( Gibbs3D == 3 ) {       /* PWH 19980310:  Was Gibbs3D = 3 */

    csf_W1 = 0.60;
    csf_W2 = 0.40;
    InVarW_1C = (0.5 / (STD_1*STD_1)) * csf_W1;
    InVarW_2C = (0.5 / (STD_2*STD_2)) * csf_W2;
/*--------------------------------------------------------------------------*/
/*      Due large variation of quality in 3mm, TWO diff. procedures used    */
/*      (1) Max. means in T2 and Pr correspond to CSF:                      */
/*	    process it as 5mm;                                              */
/*      (2) Max. mean in Pr is not equal to CSF mean in Pr:                 */
/*          rodate the feature space by 22 degress, then process it as 5mm  */
/*--------------------------------------------------------------------------*/
    Total_Mn(ImgP_1, SegImgP, N, M, SliceNo, TotalMn_1, NoRegion);
    Total_Mn(ImgP_2, SegImgP, N, M, SliceNo, TotalMn_2, NoRegion);

    csfLabel = 0;
    for ( i=0, t2CSF_V = -1.0, prMaxV = -1.0; i<NoRegion; i++ ) {
      if ( TotalMn_1[i] > t2CSF_V ) {
	t2CSF_V = TotalMn_1[i];
	csfLabel = i+1;
      }
      if ( TotalMn_2[i] > prMaxV )
	prMaxV = TotalMn_2[i];
    }
    if ( prMaxV != TotalMn_2[csfLabel-1] ) {
      angle = 22.0;
      angle = angle*PI/180.0;
      cosAngle = cos(angle);
      sinAngle = sin(angle);
      for ( k=0; k<SliceNo; k++ ) {
	for ( i=0, ptrV_1=ImgP_1[k][0], ptrV_2=ImgP_2[k][0]; 
	      i<N*M; i++, ptrV_1++, ptrV_2++ ) {
	  tmpV_2 = cosAngle*(*ptrV_2) + sinAngle*(*ptrV_1);
	  tmpV_1 =-sinAngle*(*ptrV_2) + cosAngle*(*ptrV_1);
	  *ptrV_2 = tmpV_2;
	  *ptrV_1 = tmpV_1;
	}
      }

      minV_1 = ImgP_1[0][0][0];
      minV_2 = ImgP_2[0][0][0];
      for ( k=0; k<SliceNo; k++ ) {
	for ( i=0, ptrV_1=ImgP_1[k][0], ptrV_2=ImgP_2[k][0]; 
	      i<N*M; i++, ptrV_1++, ptrV_2++ ) {
	  if ( *ptrV_1 < minV_1 ) 
	    minV_1 = *ptrV_1;
	  if ( *ptrV_2 < minV_2 ) 
	    minV_2 = *ptrV_2;
	}
      }
      printf( "Minimum values of transformed set 1 & 2 = %e, %e\n", 
	     minV_1, minV_2);

      minV_1 = minV_1 - 0.5;
      minV_2 = minV_2 - 0.5;
      for ( k=0; k<SliceNo; k++ ) {
	for ( i=0, ptrV_1=ImgP_1[k][0], ptrV_2=ImgP_2[k][0], 
		Ptr_1=SegImgP[k][0]; i<N*M; i++, ptrV_1++, Ptr_1++ ) {
	  *ptrV_1 -= minV_1;
	  *ptrV_2 -= minV_2;
	  if ( *Ptr_1 == 0 ) 
	    *ptrV_1 = 0.0;
	}
      }
    }
    printf( "CSF means of set 1 and 2 --> %f, %f  (%f)\n", 
	   t2CSF_V, TotalMn_2[csfLabel-1], prMaxV); 

    OldLike = Likelihood_2(ImgP_1, ImgP_2, SegImgP, N, M, SliceNo, NoRegion, 
			   InVarW_1GW, InVarW_2GW, InVarW_1C, InVarW_2C);
    printf( "Likelihood from initial (ML) = %e\n", OldLike);

    /*	Gibbs = OldLike/(NoObj*Voxel*6.8); */
    Gibbs = OldLike/(NoObj*Voxel*7.0);

  } /* Gibbs3D == 3 */

  else {

    fprintf( stderr, "Unexpected value for Gibbs3D in adpKmean3D_2().\n" );
    fprintf( stderr, "Hard-coded expectations are 3, 5, 6, and 9 for 2-echo images.\n" );
    fprintf( stderr, "Current value is %d.  Please notify programmers.\n", Gibbs3D );
    fprintf( stderr, "mailto:sysadmin@bblmail.psycha.upenn.edu\n" );
    exit( 1 );

  }
  printf( "CSF weights for dual-echo = %f, %f\n", csf_W1, csf_W2);


/*--------------------------------------------------------------------------*/
/* Refine segmentation of K_mean (ML)                                       */
/*--------------------------------------------------------------------------*/

  do {
    *Sp_XY = *Sp_XY/2;
    *Sp_XY = max(*Sp_XY, MinSp);
    *Sp_Z = *Sp_Z/2;

    Create_LOTS (N, M, SliceNo, *Sp_XY, *Sp_XY, *Sp_Z);
    printf( "Update means with window size = %dX%dX%d (%e)\n", 
	   2*(*Sp_XY)+1, 2*(*Sp_XY)+1, 2*(*Sp_Z)+1, Gibbs);

    IterNo = 0;

    if ( *Sp_XY <= MinSp && NoRegion == 5 &&
	 (Gibbs3D == 5 || (Gibbs3D==6 && (int)(10*Res_XY+0.25)==9)) ) { 
      for ( k=0; k<SliceNo; k++ ) { 
	for ( i=0, Ptr_1=SegImgP[k][0]; i<N*M; i++, Ptr_1++ ) { 
	  if ( *Ptr_1 == 1 || *Ptr_1 == 5 )
	    *Ptr_1 = 2;
	  else if ( *Ptr_1 == 3 || *Ptr_1 == 4 ) 
	    *Ptr_1 = 3;
	  else if ( *Ptr_1 == 2 ) 
	    *Ptr_1 = 1;
	}
      }
      NoRegion = 3;
    }

    do {
      IterNo ++;

/*--------------------------------------------------------------------------*/
/*          Assume the first image is T2, so the CSF value is brightest     */
/*--------------------------------------------------------------------------*/
      Total_Mn(ImgP_1, SegImgP, N, M, SliceNo, TotalMn_1, NoRegion);
      Update_Mn (AdpLocMn_1, TotalMn_1, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP_1, N, M, SliceNo);
      Total_Mn(ImgP_2, SegImgP, N, M, SliceNo, TotalMn_2, NoRegion);
      Update_Mn (AdpLocMn_2, TotalMn_2, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP_2, N, M, SliceNo);

      csfLabel = 0;
      for ( i=0, t2CSF_V = -1.0; i<NoRegion; i++ ) 
	if ( TotalMn_1[i] > t2CSF_V ) {
	  t2CSF_V = TotalMn_1[i];
	  csfLabel = i+1;
	} 

      IterNo_Index = 0;
      do { 
	IterNo_Index ++;
	change_no = Update_Seg2 (ImgP_1, ImgP_2, SegImgP, 
				 N, M, SliceNo, AdpLocMn_1, 
				 AdpLocMn_2, TotalMn_1, TotalMn_2, 
				 *Sp_XY, *Sp_Z, NoRegion, Gibbs, 
				 InVarW_1GW, InVarW_1C, InVarW_2C, 
				 InVarW_2GW, Gibbs3D, &NewLike, 
				 csfLabel); 
	printf( "(%d--%d): Pixels changed = %d  (%e, %d)\n", 
	       IterNo, IterNo_Index, change_no, NewLike, csfLabel);
      } while ( Gibbs != 0.0 && change_no > MinChange && 
		IterNo_Index < 10);

    } while ( IterNo_Index > 1 && IterNo < 50);

    if ( *Sp_XY <= MinSp && NoRegion == 5 && 
	 (Gibbs3D == 3 || (Gibbs3D==6 && (int)(10*Res_XY+0.25)!=9)) ) {
      for ( k=0; k<SliceNo; k++ ) { 
	for ( i=0, Ptr_1=SegImgP[k][0]; i<N*M; i++, Ptr_1++ ) { 
	  if ( *Ptr_1 == 1 || *Ptr_1 == 5 )
	    *Ptr_1 = 2;
	  else if ( *Ptr_1 == 3 || *Ptr_1 == 4 ) 
	    *Ptr_1 = 3;
	  else if ( *Ptr_1 == 2 ) 
	    *Ptr_1 = 1;
	}
      }
      NoRegion = 3;

      IterNo ++; 

/*--------------------------------------------------------------------------*/
/*          Update means since regions are merged                           */
/*--------------------------------------------------------------------------*/
      Total_Mn(ImgP_1, SegImgP, N, M, SliceNo, TotalMn_1, NoRegion);
      Update_Mn (AdpLocMn_1, TotalMn_1, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP_1, N, M, SliceNo);
      Total_Mn(ImgP_2, SegImgP, N, M, SliceNo, TotalMn_2, NoRegion);
      Update_Mn (AdpLocMn_2, TotalMn_2, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP_2, N, M, SliceNo);
      for ( i=0, t2CSF_V=-1.0; i<NoRegion; i++ ) {
	if ( TotalMn_1[i] > t2CSF_V ) {
	  t2CSF_V = TotalMn_1[i];
	  csfLabel = i+1;
	}
      }
	    
      IterNo_Index = 0;
      do { 
	IterNo_Index ++;
	change_no = Update_Seg2 (ImgP_1, ImgP_2, SegImgP, 
				 N, M, SliceNo, AdpLocMn_1, 
				 AdpLocMn_2, TotalMn_1, TotalMn_2, 
				 *Sp_XY, *Sp_Z, NoRegion, Gibbs, 
				 InVarW_1GW, InVarW_1C, InVarW_2C, 
				 InVarW_2GW, Gibbs3D, &NewLike, 
				 csfLabel); 
	printf( "(%d--%d): Pixels changed = %d  (%e, %d)\n", 
	       IterNo, IterNo_Index, change_no, NewLike, csfLabel);
      } while ( Gibbs != 0.0 && change_no > MinChange && 
		IterNo_Index < 10); 
	    
/*--------------------------------------------------------------------------*/
/*          Update means for output                                         */
/*--------------------------------------------------------------------------*/

      Total_Mn(ImgP_1, SegImgP, N, M, SliceNo, TotalMn_1, NoRegion);
      Update_Mn (AdpLocMn_1, TotalMn_1, NoRegion, *Sp_XY, *Sp_Z, 
		 SegImgP, ImgP_1, N, M, SliceNo);
    }

    Gibbs = (Gibbs * NewLike)/OldLike;
    OldLike = NewLike;
    Free_LOTS (); 
  } while ( *Sp_XY > MinSp );
    
  free (TotalMn_1);
  free (TotalMn_2);
}



/*--------------------------------------------------------------------
adpkmean_3dseg - Do adaptive 3D brain segmentation, including file i/o

This function is called by the adpkm and AdpKmean_3Dseg_Ebeta programs
to do the actual file input/output and image segmentation once all the
parameters have been determined.

Its operation is not yet completely understood, but here are some
comments:

If First is set to zero, it is interpreted to mean the first slice
which containts at least one non-zero pixel.  If Last is set to zero,
it is interpreted to mean the last slice after First which contains at
least one non-zero pixel.

FIXME: Res_XY and Res_Z are used to choose tuning parameters for the
segmentation algorithms, which is NOT a reasonable and portable
method.

FIXME:  Several of the function arguments are not well understood.

--------------------------------------------------------------------*/


int
adpkmean_3dseg( 
   int echo_12,           /* Number of echoes: either 1 or 2 */
   int NoRegion,          /* Number of regions (tissue types) */
   const char *Name1,     /* File name for the first echo */
   int ImgType1,          /* Data type for the first echo file */
   float STD_1,           /* Standard deviation of noise for first echo */
   float Weight_1,        /* Relative weight of first echo */
   const char *Name2,     /* And likewise for the second echo... */
   int ImgType2,
   float STD_2,
   float Weight_2,
   int N,                 /* Number of rows in each image */
   int M,                 /* Number of columns */
   int First,             /* First slice to process (1-origin) */
   int Last,              /* Last slice to process */
   int Total,             /* Total number of slices in each file */
   float Res_XY,          /* Voxel dimensions, in-plane */
   float Res_Z,           /* Voxel dimensions, between slices */
   int savemean )         /* Save the means image? */
{
  char          *ptr;
  FILE          *File_1, *File_2=0, *File_Est, *File_Out, 
                *File_Mean=0, *File_Vol;
  float         ***ImageP_1=0, ***ImageP_2=0, **TmpImgP;
  unsigned char	*TmpImgRawP;
  float         ****AdpLocMn_1=0, ****AdpLocMn_2=0, *tmpMeans, *tmpTotal;
  int           SliceNo=0;
  int           Sp_XY, Sp_Z, NL, ML;
  int           Intensity_v, *NoVoxel, NoObj = 0;
  unsigned char ***SegImgP;
  unsigned char *Ptr_C;
  float         *Ptr_2, *Ptr_3;
  int           ir, i, j, k;
  int		retval = 0;

  int		FirstMarker = 0, LastMarker = 0;
  int len1, len2;     /* Length of each voxel in input image files */



  Gibbs3D = (int) (Res_Z/Res_XY+0.5);
  printf( "Gibbs3D = %d\n", Gibbs3D);


/*--------------------------------------------------------------------------*/
/*  Open input and output files  */
/*--------------------------------------------------------------------------*/

  ptr = callock ((strlen(Name1)+8), sizeof(char));

  /*  File_Est is the output from the Most Likely segmentation step,
   *  the Kmean_3Dseg algorithm.  */
  strcpy (ptr, "ml_");
  strcat (ptr, Name1);
  if (!(File_Est = open_file( ptr, "r" ))) {
    retval = -1;
    goto close_none;
  }

  /* Open the output file for the segmented image */
  strcpy (ptr, "Adpml_");
  strcat (ptr, Name1);
  if (!(File_Out = open_file (ptr, "w"))) {
    retval = -1;
    goto close_est;
  }

  /* Optionally open the means file, for the means image */
  printf("savemean is %d\n", savemean);
  if (savemean) {
    if (!(File_Mean = open_file ("Adpmean.img", "w"))) {
      retval = -1;
      goto close_out;
    }
  }

  /* Open the first imput file; record its datum length */
  if (!(File_1 = open_file(Name1, "r"))) {
    retval = -1;
    goto close_mean;
  }
  len1 = sizeof_ImgType(ImgType1);   /* FIXME:  Check for invalid type */
  
  if ((echo_12 == 2) && !(File_2 = open_file(Name2, "r"))) {
    retval = -1;
    goto close_one;
  }
  len2 = sizeof_ImgType(ImgType2);   /* FIXME:  Check for invalid type */


/*--------------------------------------------------------------------------*/
/*Allocate memory                                                           */
/*--------------------------------------------------------------------------*/

  MinSp_XY = 0;
  if ( echo_12 == 1 ) {
    if ( Gibbs3D == 1 || Gibbs3D == 2 )
      MinSp_XY = 3;
    else 
      MinSp_XY = 3;
  }
  else if ( echo_12 == 2 ) {
    /*	if ( Gibbs3D==3 || Gibbs3D == 5 || */
    if ( Gibbs3D==3 || (Gibbs3D==6 && (int)(10*Res_XY+0.25)!=9) )
      MinSp_XY = 3;
    else if ( Gibbs3D==5 || (Gibbs3D==6 && (int)(10*Res_XY+0.25)==9) )
      MinSp_XY = 7; 
    else {

      fprintf( stderr, "Unexpected value for Gibbs3D in adpkmean_3dseg().\n" );
      fprintf( stderr, "Hard-coded expectations are 3, 5, 6, and 9 for 2-echo images.\n" );
      fprintf( stderr, "Current value is %d.  Please notify programmers.\n", Gibbs3D );
      fprintf( stderr, "mailto:sysadmin@bblmail.psycha.upenn.edu\n" );
      exit( 1 );

    }
  }

  NL = (N-1)/3 + 1;
  ML = (M-1)/3 + 1; 
    

  TmpImgP = alloc_img(N, M);   /* Used as both float and uchar array */
  TmpImgRawP = (unsigned char *) *TmpImgP;

/*--------------------------------------------------------------------------*/
/*Skip image or segmentation slices from 1 to (FirstSlice - 1)              */
/*--------------------------------------------------------------------------*/

  /* Skip over the non-brain slices in the initial estimate image,
     and check that there isn't any brain there. */
  FirstMarker = (First ? First : Total);
  for (k = 1; k < FirstMarker; k++) {
    fread( TmpImgRawP, 1, N * M, File_Est );   /* Always unsigned char */
    if (!First) {
      printf("First not yet found.\n");
      for (i = N * M - 1; i >= 0; i--) {
	if (*(TmpImgRawP + i)) {
	  /* Oops, we have hit the first slice with brain; rewind to
             its start */
	  fseek( File_Est, -(N * M), SEEK_CUR );
	  First = k;
	  k = FirstMarker;
	  printf( "First slice: %d\n", First );
	  break;
	}
      }
    }
    fwrite(TmpImgRawP, 1, N * M, File_Out );  /* Always unsigned char */
  }

  /* Copy the corresponding slices from image 1 to the means file,
     or just skip over them; skip over the same slices in image 2 */
  if (savemean) {
    for (k = 1; k < First; k++) {
      fread ( TmpImgRawP, len1, N * M, File_1 );
      fwrite( TmpImgRawP, len1, N * M, File_Mean );
    }
  }
  else {
    fseek( File_1, N * M * (First - 1) * len1, SEEK_SET );
  }
  if (echo_12 == 2) {
    fseek( File_2, N * M * (First - 1) * len2, SEEK_SET );
  }

  /*
   * Since we may not know how many slices we're going to have yet,
   * let us allocate enough room for pointers for the rest of the slices.
   */

  if (!Last) {
    ImageP_1 = (float ***) callock( Total - First + 1, sizeof( void * ));
    if (echo_12 == 2) {
      ImageP_2 = (float ***) callock( Total - First + 1, sizeof( void * ));
    }
    SegImgP = (unsigned char ***) callock( Total - First + 1, sizeof( void * ));
  }
  else {
    SliceNo = Last - First + 1;
    ImageP_1 = alloc_3Dimg( N, M, SliceNo );
    if (echo_12 == 2) {
      ImageP_2 = alloc_3Dimg( N, M, SliceNo );
    }
    SegImgP = alloc_3Dchar( N, M, SliceNo );
  }

/*--------------------------------------------------------------------------*/
/* Read (LastSlice - FirstSlice + 1) slices into 3D array                   */
/*--------------------------------------------------------------------------*/

  Intensity_v = 250.0/NoRegion;
  LastMarker = (SliceNo ? SliceNo : Total - First + 1);
  for (k = 0; k < LastMarker; k++) {
    read_matrix(File_Est, TmpImgP, N, M, UNSIGNED_CHAR);
    if (!Last) {
      for (i = N * M - 1; i >= 0; i--)
	if (*(TmpImgRawP + i))
	  break;
      if (i < 0) {
	/* Oops, we have hit the first empty slice; rewind to its start */
	fseek( File_Est, -(N * M), SEEK_CUR );
	Last = First + k - 1;
	SliceNo = Last - First + 1; /* Yeah, it's k, but this is the real
				       meaning behind SliceNo. */
	k = LastMarker;
	printf( "Last slice: %d\n", Last);
      }
      else {
	ImageP_1[k] = alloc_img( N, M );
	if ( echo_12 == 2) {
	  ImageP_2[k] = alloc_img( N, M );
	}
	SegImgP[k] = alloc_char( N, M );
      }
    } /* if (!Last) */

    if (k < LastMarker) {
      read_matrix(File_1, ImageP_1[k], N, M, ImgType1);
      if (echo_12 == 2) {
	read_matrix(File_2, ImageP_2[k], N, M, ImgType2);
      }
      /*
       *  Initialize SegImgP to initial estimate, but map values to
       *    step function (e.g. 250->5, 200->4, 150->3, 100->2, 50->1).
       */
      Ptr_C = SegImgP[k][0];
      Ptr_2 = TmpImgP[0];
      for (i = 0; i < N * M; i++) {
	if ( *Ptr_2 != BACKGROUND ) {
	  NoObj ++;
	  *Ptr_C = *Ptr_2/Intensity_v;
	}
	else { 
	  *Ptr_C = 0;
	}
	Ptr_C++;
	Ptr_2++;
      }
    }
  }
  printf( "Total number of non-background voxels: %d (%0.3e mm^3)\n",
	  NoObj, 
	 (NoObj*Res_XY*Res_XY*Res_Z));

/*--------------------------------------------------------------------------
 *  Segment image(s) using adaptive K-mean (MAP) method: step 2
 *--------------------------------------------------------------------------*/

  AdpLocMn_1 = (float ****) callock(SliceNo, sizeof(void *));
  for ( k=0; k<SliceNo; k++ ) {
    AdpLocMn_1[k] = alloc_3Dimg(ML, NoRegion, NL);
  }
  if ( echo_12 == 2 ) {
    AdpLocMn_2 = (float ****) callock(SliceNo, sizeof(void *));
    for ( k=0; k<SliceNo; k++ ) {
      AdpLocMn_2[k] = alloc_3Dimg(ML, NoRegion, NL);
    }
  }

  if ( echo_12 == 1 ) {
    adpKmean3D_1(ImageP_1, SegImgP, N, M, SliceNo, AdpLocMn_1, 
		 &Sp_XY, &Sp_Z, MinSp_XY, NoRegion, NoObj, Res_XY, 
		 Res_Z, STD_1, Gibbs3D);
  }
  else if ( echo_12 == 2) {
    adpKmean3D_2(ImageP_1, ImageP_2, SegImgP, N, M, SliceNo, 
		 AdpLocMn_1, AdpLocMn_2, &Sp_XY, &Sp_Z, MinSp_XY, 
		 NoRegion, NoObj, Res_XY, Res_Z, STD_1, STD_2, 
		 Weight_1, Weight_2, Gibbs3D);
  }
    
  NoRegion = 3; 
  tmpMeans=(float *) callock(NoRegion,sizeof(float));
  tmpTotal=(float *) callock(NoRegion,sizeof(float));
  NoVoxel=(int *) callock(NoRegion,sizeof(int));

  Intensity_v = 250.0/NoRegion; 
  for ( k=0; k<NoRegion; k++ ) {
    NoVoxel[k] = 0;
  }
  Create_LOTS (N, M, SliceNo, Sp_XY, Sp_XY, Sp_Z);
  Total_Mn(ImageP_1, SegImgP, N, M, SliceNo, tmpTotal, NoRegion);
  for ( k=0; k<SliceNo; k++ ) {
    Ptr_C=SegImgP[k][0];
    Ptr_2=TmpImgP[0];
    Ptr_3=ImageP_1[0][0];
    for (i = 0; i<N; i++ ) {
      for ( j=0; j<M; j++, Ptr_C++, Ptr_2++, Ptr_3++ ) {
	if ( *Ptr_C != BACKGROUND ) {
	  ir = *Ptr_C;
	  NoVoxel[ir-1] ++;
	  Interp_Mn (i, j, k, AdpLocMn_1, tmpTotal, Sp_XY, Sp_Z, 
		     SegImgP, N, M, SliceNo, tmpMeans, NoRegion);
	  *Ptr_2 = tmpMeans[ir-1];
	  *Ptr_3 = ir*Intensity_v;
	}
	else {
	  *Ptr_2 = *Ptr_3 = 0.0;
	}
      }
    }
    write_matrix (File_Out, ImageP_1[0], N, M, UNSIGNED_CHAR);
    if (savemean) {
      write_matrix (File_Mean, TmpImgP, N, M, ImgType1);
    }
  }
  Free_LOTS();


  for ( k=0; k<NoRegion; k++ ) {
    printf( "Voxel No. of Tissue Type %d = %d\n", k+1, NoVoxel[k]);
  }

  if (savemean) {
    File_Vol = fopen ("Vol_Measure.dat", "a");
    if ( File_Vol != 0 ) {
      for ( k=0; k<NoRegion; k++ ) {
	fprintf(File_Vol, "%s: Voxel No. of tissue type %-d  = %-d\n",
		Name1, k+1, NoVoxel[k]);
      }
      fclose( File_Vol );
    }
    else {
      fprintf( stderr, "AdpKmean_3Dseg_Ebeta: WARNING: could not append to" );
      fprintf( stderr, " Vol_Measure.dat.\n" );
    }
  }


/*--------------------------------------------------------------------------
 *  Append image or segmentation slices from LastSlice+1 to the end
 *--------------------------------------------------------------------------*/

  if (savemean) {
    for (k = Last + 1; k <= Total; k++) {
      fread ( TmpImgP[0], len1, N * M, File_1 );
      fwrite( TmpImgP[0], len1, N * M, File_Mean );
    }
  }
  for (k = Last + 1; k <= Total; k++) {
    fread ( TmpImgP[0], 1, N * M, File_Est );
    fwrite( TmpImgP[0], 1, N * M, File_Out );
  }


/*--------------------------------------------------------------------------*/
/*Close opened files and release memory                                     */
/*--------------------------------------------------------------------------*/

  free(NoVoxel);
  free(tmpMeans);
  free(tmpTotal);
  free_img(TmpImgP);
  free_3Dchar(SegImgP, SliceNo);

  free_3Dimg(ImageP_1, SliceNo);
  for ( k=0; k<SliceNo; k++ ) {
    free_3Dimg(AdpLocMn_1[k], NL);
  }
  free(AdpLocMn_1);

  if ( echo_12 == 2) {
    fclose(File_2);
    free_3Dimg(ImageP_2, SliceNo);
    for ( k=0; k<SliceNo; k++ ) {
      free_3Dimg(AdpLocMn_2[k], NL);
    }
    free(AdpLocMn_2);
  }


close_one:
  fclose( File_1 );

close_mean:
  if (savemean) {
    fclose( File_Mean );
  }

close_out:
  fclose( File_Out );

close_est:
  fclose( File_Est );

close_none:
  free( ptr );

  return retval;
}
