/*--------------------------------------------------------------------------
BrainMask.c - Isolate or mask brain within a head image

Programmer: Michelle Yan
Revisions:  Paul Hughett                                      

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

   This program extracts brain volume as a pre-segmentation step,
which may need to further check by a expert.  Or it converts formats
to get brain volume with the original data format.
--------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: BrainMask.c,v 1.12 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";


#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>
#include <time.h>

#include "common.h"
#include "conn_component.h"
#include "definition.h"
#include "find_mask.h"
#include "image_inc.h"
#include "image_io.h"
#include "list.h"
#include "memm.h"
#include "region_grow.h"

#include "BrainMask.h"


/*--------------------------------------------------------------------------
bmsk - Brain masking, once command line has been processed

This function does one of three different operations, depending on the
values of OuterFirst_YN and Mask_YN.

If Mask_YN is 'y', it applies an existing mask file to the input image.
More precisely, the input image is set to 0 wherever the mask image is
equal to BACKGROUND, and then written out as the output image.

If Mask_YN is not 'y' and OuterFirst_YN is 'y', then this function
attempts to remove skull and scalp from an input T1 or proton image,
leaving only the brain.

Otherwise, this function attempts to remove the skull and scalp from a
PET or T2 image.

FIXME:  These three separate operations really ought to be three separate
functions, if not programs, since there are only vague resemblances in
the actual algorithms involved.

---------------------------------------------------------------------------*/
int
bmsk( char *Name,
      int ImgType,
      int N,
      int M,
      int Total,
      char *Mask_YN,
      char *MaskName,
      int MaskType,
      char *OuterFirst_YN,
      char *Overlay_YN,
      char *OutputSuffix )
{

  char          *ptr;
  FILE          *File_In,  *File_Out, *File_Overlay=0, *File_Mask;
  float         ***ImageIn, **tmpImg_1=0, **tmpImg_2=0, *ptr_f, *ptr_1, *ptr_2;
  unsigned char ***ScaledImg, ***Mask, ***tmpBin_1, ***tmpBin_2=0, *ptr_c1;
  int           i, k;
  int		retval = 0;
  clock_t       clock0, clock1;        /* To compute CPU run time */
  time_t        time0,  time1;         /* To compute elapsed time */
  double        cpu_time, elapsed;


  /* Record start times */
  clock0 = clock();
  time0  = time(NULL);

  /*-----------------------------------------------------------*/
  /*  Open i/o files for reading and writing                   */
  /*------------------------------------------------------------*/

  ptr = callock( (strlen(Name)+5), sizeof(char));
  strncpy (ptr, Name, strlen(Name)-4);   /* FIXME:  Assumes .img */
  strcat( ptr, OutputSuffix );
  strcat (ptr, ".img");

  if (!(File_In = open_file(Name, "r"))) {
    retval = -1;
    goto close_none;
  }

  if (!(File_Out = open_file(ptr, "w"))) {
    retval = -1;
    goto close_in;
  }


  /*------------------------------------------*/
  /*  Brain mask exists; apply to input file  */
  /*------------------------------------------*/

  if ( *Mask_YN == 'y' ) {
    if (!(File_Mask = open_file(MaskName, "r"))) {
      retval = -1;
      goto close_out;
    }
    tmpImg_1 = alloc_img(N, M);
    tmpImg_2 = alloc_img(N, M);
    for ( k=0; k<Total; k++) {
      read_matrix(File_In, tmpImg_1, N, M, ImgType);
      read_matrix(File_Mask, tmpImg_2, N, M, MaskType);
      for ( i=0, ptr_1=tmpImg_1[0], ptr_2=tmpImg_2[0]; 
	    i<N*M; i++, ptr_1++, ptr_2++ ) {
	if ( *ptr_2 == BACKGROUND )
	  *ptr_1 = 0;
      }
      write_matrix (File_Out, tmpImg_1, N, M, ImgType);
    }
    free_img (tmpImg_1);
    free_img (tmpImg_2);

    fclose( File_Mask );
    fclose( File_Out );
    fclose( File_In );
    free( ptr );
    return retval;
  }


  /*-------------------------------------------------*/
  /*  Brain mask does not exist; make one            */
  /*-------------------------------------------------*/

  ImageIn = alloc_3Dimg(N, M, Total);
  ScaledImg = alloc_3Dchar(N, M, Total);

  if ( *Overlay_YN == 'y' ) {
    strncpy (ptr, Name, strlen(Name)-4);
    *(Name + strlen(Name) - 4) = 0;
    strcat (ptr, "_bi.img");

    if (!(File_Overlay = open_file (ptr, "w"))) {
      printf( "Ignoring request for overlay image.\n" );
      *Overlay_YN = 'n';
    }
  } 

  /*----------------------------*/
  /*      Read a 3D array       */
  /*----------------------------*/

  for ( k=0; k<Total; k++) {
    read_matrix(File_In, ImageIn[k], N, M, ImgType);
  }
  printf("Finished reading 3D image.\n");


  /*----------------------------------*/
  /*      Search a 3D brain volume    */
  /*----------------------------------*/

  Scale2Char3D (ImageIn, ScaledImg, N, M, Total, NOSMOOTH);

  tmpBin_1 = alloc_3Dchar(N, M, Total);
  BinaryImg3D (ScaledImg, tmpBin_1, N, M, Total, "n");
	
  Mask = alloc_3Dchar(N, M, Total);
  if ( *OuterFirst_YN == 'y' ) {
    tmpBin_2 = alloc_3Dchar(N, M, Total);
    ExtractVOI( tmpBin_1, tmpBin_2, N, M, Total, "y" );
    BinaryImg3D( ScaledImg, tmpBin_1, N, M, Total, "y" );
    InnerMask( tmpBin_1, tmpBin_2, Mask, N, M, Total );
    CleanUpMask( Mask, tmpBin_1, N, M, Total );
  }
  else {
    ExtractVOI (tmpBin_1, Mask, N, M, Total, "n"); 
  }

  if ( *Overlay_YN == 'y' ) { 

  /*----------------------------------------------------*/
  /*      Write overlayed image (original + contour)    */
  /*----------------------------------------------------*/

    tmpImg_1 = alloc_img(N, M);
    tmpImg_2 = alloc_img(N, M);
    for ( k=0; k<Total; k++ ) { 
      ptr_c1 = Mask[k][0];
      ptr_f = tmpImg_1[0];
      for ( i=0; i<N*M; ptr_c1++, ptr_f++, i++) {
	*ptr_f = *ptr_c1;
      }
      find_mask (tmpImg_1, tmpImg_2, N, M, "CONTOUR");
      ptr_c1 = ScaledImg[k][0];
      ptr_f = tmpImg_2[0];
      for ( i=0; i<N*M; ptr_c1++, ptr_f++, i++) {
	if ( *ptr_f == BOUNDARY )
	  *ptr_f = OBJECT;
	else 
	  *ptr_f = *ptr_c1;
      } 
      write_matrix (File_Overlay, tmpImg_2, N, M, UNSIGNED_CHAR); 
    }
  }
  for ( k=0; k<Total; k++ ) { 
    ptr_c1 = Mask[k][0];
    ptr_f = ImageIn[k][0];
    for ( i=0; i<N*M; ptr_c1++, ptr_f++, i++) {
      if ( *ptr_c1 == BACKGROUND ) 
	*ptr_f = 0.0;
    }
    write_matrix (File_Out, ImageIn[k], N, M, ImgType);
  } 

  /*---------------------------------------------*/
  /*      Close opened files and release memory  */
  /*---------------------------------------------*/

  free_3Dimg (ImageIn, Total);
  if ( *Overlay_YN == 'y' ) {
    fclose(File_Overlay);
    free_img (tmpImg_1);
    free_img (tmpImg_2);
  }
  free_3Dchar(Mask, Total);
  free_3Dchar(tmpBin_1, Total);
  if (*OuterFirst_YN == 'y') {
    free_3Dchar(tmpBin_2, Total);
  }


close_out:
  fclose(File_Out);

close_in:
  fclose(File_In);

close_none:
  free (ptr);

  /* Report the CPU and elapsed time used */
  clock1 = clock();
  time1  = time(NULL);
  cpu_time = (clock1 - clock0) / (double)CLOCKS_PER_SEC;
  elapsed  = (double)(time1 - time0);
  fprintf(stderr, "CPU run time was %8.3f seconds.\n", cpu_time);
  fprintf(stderr, "Elapsed time was %8.3f seconds.\n", elapsed);

  return retval;
}
