/*--------------------------------------------------------------------------
BrainMask_main.c - Extract brain volume from a head image

Programmer: Michelle Yan
Revisions:  Paul Hughett                                     

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


   It extract brain volume as a pre-segmentation step, which may need to  
further check by a expert. Or it converts formats to get brain volume    
with the original data format.                                           
--------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: BrainMask_main.c,v 1.11 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

/*--------------------------------------------------------------------------*/
/* Include proper constant, variable and function declarations              */
/*--------------------------------------------------------------------------*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <unistd.h>     /* For getopt and friends */
#include <sys/types.h>
#include <sys/stat.h>

#include "BrainMask.h"

#define MAX_INPUT_LEN 255


/*----------------------------------------------------------------------------
showusage - Describe the command arguments to the user

-----------------------------------------------------------------------------*/

static void
showusage( char * argv0 )
{
  fprintf( stderr, "Usage:\n\n" );

  fprintf( stderr, "Batch mode:\n" );
  fprintf( stderr, "	%s [-r rows] [-c cols] [-d] [-o]\n", argv0 );
  fprintf( stderr, "		[-t type] [-o output_suffix]\n" );
  fprintf( stderr, "		filename ...\n\n" );

  fprintf( stderr, "Interactive / Parameter file mode (old method):\n" );
  fprintf( stderr, "	%s < {parameter_filename}\n\n", argv0 );

  fprintf( stderr, "	%s\n\n", argv0 );

  fprintf( stderr, "	-r, -c	# of rows and columns; default is 256\n" );
  fprintf( stderr, "	-d	debug\n" );
  fprintf( stderr, "	-p	peel off outer layer first (use for\n" );
  fprintf( stderr, "		T1-weighted images)\n" );
  fprintf( stderr, "	-t	image data type (see table below)\n" );
  fprintf( stderr, "		default is 2\n" );
  fprintf( stderr, "	-o	Output suffix.  default is  _ns\n" );

  fprintf( stderr, "\nImage types:\n");
  fprintf( stderr, "	 1 = One Byte Per Voxel,   Unsigned\n");
  fprintf( stderr, "	 2 = Two Bytes Per Voxel,  Unsigned\n");
  fprintf( stderr, "	-2 = Two Bytes Per Voxel,  Signed\n");
  fprintf( stderr, "	 4 = Four Bytes Per Voxel, Signed\n");
  fprintf( stderr, "	-4 = Floating Point Voxel Data\n");
  fprintf( stderr, "(Usually 2 is correct for raw MRI data)\n");
}


/*----------------------------------------------------------------------------
main - main program for BrainMask

-----------------------------------------------------------------------------*/

int
main( int argc, char *argv[] )
{

  char		Name[128], MaskName[128];
  char		nextline[MAX_INPUT_LEN];
  char		Mask_YN[MAX_INPUT_LEN],
    Overlay_YN[MAX_INPUT_LEN],
    OuterFirst_YN[MAX_INPUT_LEN],
    OutputSuffix[MAX_INPUT_LEN];
  int		ImgType, MaskType;
  int		N, M, Total; 
  struct stat	filestat;

  int		c;
  int		debugflag = 0;

  /* Default settings */

  *Mask_YN = 'n';		/* create a mask (one does not yet exist) */
  *Overlay_YN = 'n';		/* do not output overlay image */
  *OuterFirst_YN = 'n';		/* do not remove outer layer first (T2) */
  strcpy( OutputSuffix, "_ns" );

  ImgType = MaskType = 2;	/* unsigned 2-byte */
  N = M = 256;			/* 256 x 256 pixels per slice */
  Total = 0;			/* calculate # slices using file size */


  /* Process command line options, or tell user the right way to do it */
  while ((c = getopt( argc, argv, "dpr:c:t:o:")) != EOF) {
    switch (c) {
    case 'd':
      debugflag = 1;
      break;
    case 'p':
      *OuterFirst_YN = 'y';
      break;
    case 'r':
      if (sscanf( optarg, "%d", &N ) != 1)
	N = 256;
      break;
    case 'c':
      if (sscanf( optarg, "%d", &M ) != 1)
	M = 256;
      break;
    case 't':
      if (sscanf( optarg, "%d", &ImgType ) != 1)
	ImgType = 2;
      break;
    case 'o':
      strncpy( OutputSuffix, optarg, MAX_INPUT_LEN );
      break;
    case '?':
      showusage( argv[0] );
      exit( 1 );
    }
  }


  /* at least one filename argument provided */
  if (optind < argc) {

    int retval = 0;

    /* silence stdout during normal operation */
    if (!debugflag)
      freopen( "/dev/null", "w", stdout );

    while (optind < argc) {

      if (stat( argv[optind], &filestat ) < 0) {
	perror( argv[optind] );
	putchar( '\n' );
	exit( -1 );
      }
      Total = filestat.st_size / (N * M * abs(ImgType));

      printf("\nImage name & type --> %s (%d)\n", argv[optind], ImgType);
      printf("\nImage dimensions = %d X %d X %d\n", N, M, Total);
      printf("\nRemove outer layer of MR image first --> %c\n",
	     (*OuterFirst_YN == 'y' ? 'y' : 'n'));

      retval += bmsk( argv[optind], ImgType, N, M, Total,
		      Mask_YN, MaskName, MaskType,
		      OuterFirst_YN, Overlay_YN,
		      OutputSuffix );
      optind ++;

    }

    return retval;
  }


  /* No filename given, but options were given */
  if (argc > 2) {
    showusage( argv[0] );
    exit( 1 );
  }

  if (argc == 2 && !debugflag) {
    showusage( argv[0] );
    exit( 1 );
  }

  /*
   * No filename(s) were given, and no options were given.
   * Revert to old method of getting parameters from stdin
   */

  /* silence stdout for batch mode (e.g. "bmsk < parameter_file") */
  if (!(isatty( fileno( stdin ))) && !debugflag)
    freopen( "/dev/null", "w", stdout );
  else
    showusage( argv[0] );

  printf("Please enter original image filename (*.img) & type.\n\n");
  fgets( nextline, MAX_INPUT_LEN, stdin );
  sscanf( nextline, "%126s %d", Name, &ImgType );
  printf("\nImage name & type --> %s (%d)\n", Name, ImgType);

  printf("\nPlease enter image dimensions:  Rows  Cols  Total_slices\n");
  printf("(If Total_slices is unknown, enter 0)\n\n");
  fgets( nextline, MAX_INPUT_LEN, stdin );
  sscanf( nextline, "%d %d %d", &N, &M, &Total );

  if (!Total) {
    if (stat( Name, &filestat ) < 0) {
      perror( Name );
      putchar( '\n' );
      exit( -1 );
    }
    Total = filestat.st_size / (N * M * abs(ImgType));
  }
  printf("\nImage dimensions = %d X %d X %d\n", N, M, Total);

  printf("\nDo you want to apply an existing brain-mask image (y/n) [n] ? ");
  fgets( Mask_YN, MAX_INPUT_LEN, stdin );
  if ( *Mask_YN == 'y' ) {
    printf("\nA brain-mask image exists, and will be applied to the\n");
    printf("  above-entered original image.\n\n");

    printf("Please enter mask image filename & type.\n\n");
    scanf("%126s %d", MaskName, &MaskType );
    printf("\nMask name & type --> %s (%d)\n", MaskName, MaskType);
  }

  else {
    printf("\nOkay, we will create a brain-mask image.\n");
    printf("<< For T2-weighted sequences, respond with n: >>\n");

    printf("\nRemove outer layer of MR image first (y/n) [n] ? ");
    fgets( OuterFirst_YN, MAX_INPUT_LEN, stdin );
    printf("\nRemove outer layer of MR image first --> %c\n",
	   (*OuterFirst_YN == 'y' ? 'y' : 'n'));

    printf("\nWould you also like to create an image with\n");
    printf("  the masking contour overlayed on top (y/n) [n] ? ");
    fgets( Overlay_YN, MAX_INPUT_LEN, stdin );
    printf("\nOutput (original + contour) image --> %c\n",
	   (*Overlay_YN == 'y' ? 'y' : 'n'));
  }

  return bmsk( Name, ImgType, N, M, Total,
	       Mask_YN, MaskName, MaskType,
	       OuterFirst_YN, Overlay_YN,
	       OutputSuffix );
}
