/***************************************************************************
brmask1.c - Brain masking for a T1 or proton image

Author:  Paul Hughett, based on an earlier program (BrainMask) written by
Michelle Yan.

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


   brmask1   image   [ size ]

This program reads a raw image file containing an MRI T1- or proton-weighted
image of the human head and (attempts to) remove the skull and scalp, leaving
only the brain and surrounding csf.  It writes the result to an output file.
The input image is assumed to be raw pixel values, each a two-byte big endian
unsigned integer, in the usual raster scan order; the output file is in the
same format, with non-brain voxels set to 0.  The size, if given, is taken
as the number of rows and the number of columns; if omitted, it is assumed
to be 256.  The algorithm used is probably suitable only for images obtained
using the Gur protocol.

This program uses the same algorithm as Michelle Yan's BrainMask program
with the OuterFirst_YN option enabled, but may have been revised for
greater efficiency.  The newer program brmask3 is intended for the same
purpose and uses an improved (we hope) but imcompatible algorithm.

***************************************************************************/

static const char rcsid[] = "$Id: brmask1.c,v 1.6 2000/06/05 17:33:57 hughett Exp $ Release $Name: segm-0-90 $";


#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include <time.h>
#include <unistd.h>
#include <sys/stat.h>

#include "common.h"
#include "conn_component.h"
#include "definition.h"
#include "errm.h"
#include "find_mask.h"
#include "image_inc.h"
#include "image_io.h"
#include "list.h"
#include "memm.h"
#include "param.h"
#include "region_grow.h"


int verbosity = 1;            /* For increasingly verbose commentary */

int
main(int argc, char *argv[])
{
    /* Parameters possibly set by the user */
    Param_list params;             /* Contents of the param file */
    char *Input;                   /* Name of the input image file */
    char *Output;                  /* Name of the output mask image file */
    char *Overlay;                 /* Name of the overlay image output file */
    int ImgType = 2;               /* Image type:  2 => 2-byte unsigned int */
    int nrows, ncols;              /* Number of rows and columns */
    int nslices;                   /* Number of slices */
    char Overlay_YN[] = "n";       /* Overlay contour on original image? */
    char ParamSuffix[]   = ".par";      /* Suffix for param file */
    char InputSuffix[]   = ".img";      /* Suffix for input file */
    char OutputSuffix[]  = "_ns.img";   /* Suffix for (primary) output file */
    char OverlaySuffix[] = "_bi.img";   /* Suffix for overlay output file */

    /* Parameters derived from the settable parameters */
    int nbytes = 2;                /* Number of bytes per pixel */
    int M, N, Total;               /* Image dimensions */

    /* Other variables */
    FILE *File_In;                 /* Input image */
    struct stat imgstat;           /* Information on the image file */
    FILE *File_Out;                /* Output mask image */
    FILE *File_Overlay=0;          /* (Optional) overlay image */
    float ***ImageIn;              /* Input gray scale image */
    float **tmpImg_1=0;            /* */
    float **tmpImg_2=0;
    float *ptr_f;
    unsigned char ***ScaledImg;    /* Input image scaled to 0..255 */
    unsigned char ***Mask;         /* Output brain mask image */
    unsigned char ***tmpBin_1;
    unsigned char ***tmpBin_2=0;
    unsigned char *ptr_c1;
    int i, k, status, len;
    char *s;
    const char *s1, *s2;
    clock_t clock0, clock1;        /* To compute CPU run time */
    time_t  time0,  time1;         /* To compute elapsed time */
    double  cpu_time, elapsed;

    /* Record start times */
    clock0 = clock();
    time0  = time(NULL);

    /* Read the param file */
    if (argc != 2)
	fatal("Invalid number of arguments");
    params = read_param_file(argv[1]);

    /* Determine the name of the input file */
    Input = (char *)get_param(params, argv[0], "input");
    if (Input == 0) {
	len = strlen(argv[1]) - strlen(ParamSuffix) + strlen(InputSuffix); 
	Input = mallock(len);
	strcpy(Input, argv[1]);
	s = Input + strlen(argv[1]) - strlen(ParamSuffix);
	if (strcmp(s, ParamSuffix) != 0)
	    fatal("Param file name does not end in .par");
	strcpy(s, InputSuffix);
    }

    /* Determine the name of the output file */
    Output = (char *)get_param(params, argv[0], "output");
    if (Output == 0) {
	len = strlen(argv[1]) - strlen(InputSuffix) + strlen(OutputSuffix); 
	Output = mallock(len);
	strcpy(Output, Input);
	s = Output + strlen(Output) - strlen(InputSuffix);
	if (strcmp(s, InputSuffix) != 0)
	    fatal("Input file name does not end in .img");
	strcpy(s, OutputSuffix);
    }

    /* Determine the name of the overlay output file */
    Overlay = (char *)get_param(params, argv[0], "overlay");
    if (Overlay == 0) {
	len = strlen(argv[1]) - strlen(InputSuffix) + strlen(OverlaySuffix); 
	Overlay = mallock(len);
	strcpy(Overlay, Input);
	s = Overlay + strlen(Overlay) - strlen(InputSuffix);
	if (strcmp(s, InputSuffix) != 0)
	    fatal("Input file name does not end in .img");
	strcpy(s, OverlaySuffix);
    }

    /* Determine the image type */
    s1 = get_param(params, argv[0], "type");
    if (s1 == 0)  ImgType = 2;
    else          ImgType = strtol(s1, 0, 10);
    nbytes = sizeof_ImgType(ImgType);
    if (nbytes == 0)
	fatal("Invalid image type");

    /* Determine number of rows and columns */
    s1 = get_param(params, argv[0], "nrows");
    s2 = get_param(params, argv[0], "ncols");
    if (s1 == 0 && s2 == 0) 
	nrows = ncols = 256;
    else if (s1 == 0)
	nrows = ncols = strtol(s2, 0, 10);
    else if (s2 == 0)
	ncols = nrows = strtol(s1, 0, 10);
    else {
	nrows = strtol(s1, 0, 10);
	ncols = strtol(s2, 0, 10); }
    N = nrows;
    M = ncols;
    if (nrows < 0 || ncols < 0)
	fatal("Invalid number of rows or columns");

    /* Determine number of slices in input file */
    s1 = get_param(params, argv[0], "nslices");
    if (s1 != 0)
	nslices = strtol(s1, 0, 10);
    else {
	status = stat(Input, &imgstat);
	if (status != 0) {
	    fprintf(stderr, "Stat failed on image file.\n");
	    exit(0); }
	nslices = imgstat.st_size / ((long)N*M*nbytes);
    }
    Total = nslices;

    /* Tell the user what parameter values he has chosen or defaulted */
    if (verbosity >= 1) {
	fprintf(stderr, "%s.input   = %s\n", argv[0], Input);
	fprintf(stderr, "%s.output  = %s\n", argv[0], Output);
	fprintf(stderr, "%s.overlay = %s\n", argv[0], Overlay);
	fprintf(stderr, "%s.type    = %d\n", argv[0], ImgType);
	fprintf(stderr, "%s.nrows   = %d\n", argv[0], N);
	fprintf(stderr, "%s.ncols   = %d\n", argv[0], M);
	fprintf(stderr, "%s.nslices = %d\n", argv[0], Total);
    }

    /* Allocate working memory */
    fprintf(stderr, "Allocating image memory...\n");
    ImageIn   = alloc_3Dimg (N, M, Total);
    ScaledImg = alloc_3Dchar(N, M, Total);
    Mask      = alloc_3Dchar(N, M, Total);
    tmpBin_1  = alloc_3Dchar(N, M, Total);
    tmpBin_2  = alloc_3Dchar(N, M, Total);

    /* Read the 3D input image */
    if (!(File_In = open_file(Input, "r"))) {
        fprintf(stderr, "Unable to open input file.\n");
        exit(EXIT_FAILURE);
    }
    fprintf(stderr, "Reading the input image...\n");
    for ( k=0; k<Total; k++) {
        read_matrix(File_In, ImageIn[k], N, M, ImgType);
    }
    fclose(File_In);


    /* Extract the brain from the input image */
    fprintf(stderr, "Preparing scaled binary image...\n");
    Scale2Char3D   ( ImageIn,   ScaledImg,  N, M, Total, NOSMOOTH );
    BinaryImg3D  ( ScaledImg, tmpBin_1,       N, M, Total, "n" ); 
    if (0) put_volume_uchar("binary.img", tmpBin_1, N, M, Total);

    fprintf(stderr, "Extracting the head volume...\n");
    ExtractHead  ( tmpBin_1,  tmpBin_2,       N, M, Total );
    /* tmpBin_2 now contains the head mask image */
    if (0) put_volume_uchar("hdmask.img", tmpBin_2, N, M, Total);

    fprintf(stderr, "Removing skull and scalp...\n");
    BinaryImg3D  ( ScaledImg, tmpBin_1,       N, M, Total, "y" );
    if (0) put_volume_uchar("binmask.img", tmpBin_1, N, M, Total);
    InnerMask    ( tmpBin_1,  tmpBin_2, Mask, N, M, Total );
    if (0) put_volume_uchar("inmask.img", Mask, N, M, Total);
    CleanUpMask  ( Mask,      tmpBin_1,       N, M, Total );


    /* Write the output image, slice by slice */
    File_Out = open_file(Output, "w");
    if ( File_Out == 0 )
        fatal("Unable to open output file.");
    fprintf(stderr, "Writing the output image...\n");
    for ( k=0; k<Total; k++ ) { 
        ptr_c1 = Mask[k][0];
        ptr_f  = ImageIn[k][0];
        for ( i=0; i<N*M; ptr_c1++, ptr_f++, i++) {
            if ( *ptr_c1 == BACKGROUND )  *ptr_f = 0.0;
        }
        write_matrix (File_Out, ImageIn[k], N, M, ImgType);
    } 
    fclose(File_Out);


    /* Optionally compute and write the overlay image (input + contour) */
    if ( *Overlay_YN == 'y' ) {
        if (!(File_Overlay = open_file (Overlay, "w"))) {
            fprintf(stderr, "Unable to open file for overlay image.\n" );
            exit(EXIT_FAILURE);
        }
        fprintf(stderr, "Computing and writing the overlay image...\n");
        tmpImg_1 = alloc_img(N, M);
        tmpImg_2 = alloc_img(N, M);
        for ( k=0; k<Total; k++ ) { 
            ptr_c1 = Mask[k][0];
            ptr_f = tmpImg_1[0];
            for ( i=0; i<N*M; ptr_c1++, ptr_f++, i++) {
	        *ptr_f = *ptr_c1;
            }
            find_mask (tmpImg_1, tmpImg_2, N, M, "CONTOUR");
            ptr_c1 = ScaledImg[k][0];
            ptr_f  = tmpImg_2[0];
            for ( i=0; i<N*M; ptr_c1++, ptr_f++, i++) {
	        if ( *ptr_f == BOUNDARY )  *ptr_f = OBJECT;
	        else                       *ptr_f = *ptr_c1;
            } 
            write_matrix (File_Overlay, tmpImg_2, N, M, UNSIGNED_CHAR); 
        }
        fclose(File_Overlay);
        free_img (tmpImg_1);
        free_img (tmpImg_2);
    }


    /* Close open files and release memory
      (FIXME:  much of this could be done sooner) */
    free_3Dimg (ImageIn,  Total);
    free_3Dchar(Mask,     Total);
    free_3Dchar(tmpBin_1, Total);
    free_3Dchar(tmpBin_2, Total);

    /* Report the CPU and elapsed time used */
    clock1 = clock();
    time1  = time(NULL);
    cpu_time = (clock1 - clock0) / (double)CLOCKS_PER_SEC;
    elapsed  = (double)(time1 - time0);
    fprintf(stderr, "CPU run time was %8.3f seconds.\n", cpu_time);
    fprintf(stderr, "Elapsed time was %8.3f seconds.\n", elapsed);

    return 0;
}

