/*--------------------------------------------------------------------------
conn_component.c - find connected component of images of any type.          

Author:     Michelle Yan
Revisions:  Paul Hughett

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


   This source code contains some basic functions for 3D region grow.     

   CC_flood: grow region from a seed pixel.                               
   CC_label: perform connected component labeling.                        
----------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: conn_component.c,v 1.11 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";


#include "conn_component.h"
#include "errm.h"
#include "memm.h"

/* -------------------------------------------------------------------------*/
/* Constants for tuning CC_label                                            */
/* -------------------------------------------------------------------------*/

#define MAX_USHORT 65535          /* Maximum number of connected components */
#define SMALL_COMPONENT_SIZE 100  /* Maximum component size to relabel */


/* -------------------------------------------------------------------------*/
/* Prototypes for internal functions                                        */
/* -------------------------------------------------------------------------*/

static stack_t * CC_label( void ***, int, int, int, compfunc );
static int CC_flood( void ***, int, int, int, int, int, int, void *, void *,
   compfunc, copyfunc );
static void CC_add_to_neighbors( int, int, int, stack_t * );
static int CC_get_from_neighbors( int *, int *, int *, stack_t * );
static void CC_add_all_neighbors( void ***, int, int, int, int, int, int,
   stack_t *, void *, compfunc );



/****************************************************************************/
/* Part 1: miscellaneous application-specific lblcmp and lblcpy functions.  */
/*                                                                          */
/* Contrary to what you might expect, cmp_* returns 0 for equality and      */
/* 1 for inequality.  The reason for this is unknown.                       */
/*                                                                          */
/* The copy routines verify that the destination has a specified value      */
/* before copying; they return 0 if the specified value is absent (and      */
/* do not copy) and 1 if it is.                                             */
/*                                                                          */
/* NOTE: when cmp_* is not defined strictly by equality, the line           */
/*           *curLabel = inVec[index];                                      */
/*       is needed in cpy_* before "inVec[index] =* newLabel;"              */
/****************************************************************************/

/* Compare unsigned char labels */
static int
cmp_uchar_pix( unsigned char *vec1,
	       int index1,
	       unsigned char *vec2,
	       int index2 )
{
  if (vec1[index1] == vec2[index2])
    return 0;
  else
    return 1;    
}


/* Compare unsigned short labels */
static int
cmp_ushort_pix( unsigned short *vec1,
		int index1,
		unsigned short *vec2,
		int index2 )
{
  if (vec1[index1] == vec2[index2])
    return 0;
  else
    return 1;    
}


/* Copy an unsigned char label */
static int
cpy_uchar_pix( unsigned char *inVec,
	       int index,
	       unsigned char *curLabel,
	       unsigned char *newLabel )
{
  if (inVec[index] != *curLabel)
    return 0;
  inVec[index] = *newLabel;
  return 1;
}


/* Copy an unsigned short label */
static int
cpy_ushort_pix( unsigned short *inVec,
		int index,
		unsigned short *curLabel,
		unsigned short *newLabel )
{
  if (inVec[index] != *curLabel)
    return 0;
  inVec[index] = *newLabel;
  return 1;
}





/****************************************************************************/
/* part 2: type-specific public functions                                   */
/****************************************************************************/


/*--------------------------------------------------------------------------*/
/* CC_flood_uchar - Flood connected region in a uchar volume image          */
/*                                                                          */
/* This function starts from a given seed pixel and relabels all connected  */
/* pixels containing the value curLabel to the new value newLabel.  It      */
/* returns the number of voxels in the region.                              */
/*--------------------------------------------------------------------------*/

int
CC_flood_uchar( unsigned char ***inImg,   /* Volume image to modify */
		int numSlice,             /* Dimensions of the image */
		int numRow,
		int numCol, 
		int seedSlice,            /* Seed pixel for flooding */
		int seedRow,
		int seedCol,
		unsigned char *curLabel, /* pointers to labels */
		unsigned char *newLabel )
{
  unsigned char tmpCurLabel=*curLabel, tmpNewLabel=*newLabel;

  return CC_flood ((void ***) inImg, numSlice, numRow, numCol, 
	           seedSlice, seedRow, seedCol, &tmpCurLabel, 
		   &tmpNewLabel, (compfunc) cmp_uchar_pix,
                   (copyfunc) cpy_uchar_pix);
}


/*--------------------------------------------------------------------------*/
/* CC_flood_ushort - Flood connected region in a ushort volume image        */
/*                                                                          */
/* This function starts from a given seed pixel and relabels all connected  */
/* pixels containing the value curLabel to the new value newLabel.  It      */
/* returns the number of voxels in the region.                              */
/*--------------------------------------------------------------------------*/

int
CC_flood_ushort( unsigned short ***inImg,   /* Image to be modified */
		 int numSlice,              /* Dimensions of the image */
		 int numRow,
		 int numCol, 
		 int seedSlice,             /* Location of seed pixel */
		 int seedRow,
		 int seedCol,
		 unsigned short *curLabel, /* pointers to labels */
		 unsigned short *newLabel )
{
  unsigned short tmpCurLabel=*curLabel, tmpNewLabel=*newLabel;

  return CC_flood ((void ***) inImg, numSlice, numRow, numCol, 
	           seedSlice, seedRow, seedCol, &tmpCurLabel, 
		   &tmpNewLabel, (compfunc) cmp_ushort_pix,
                   (copyfunc) cpy_ushort_pix);
}



/*--------------------------------------------------------------------------*/
/* CC_label_uchar - Find connected components of a uchar volume image       */
/*                                                                          */
/* This function finds the connected components of an unsigned char volume  */
/* image and returns a stack which contains a pointer to a seed pixel for   */
/* each region found.   See CC_label for more information.                  */
/*--------------------------------------------------------------------------*/

stack_t *
CC_label_uchar( void ***inImg,
		int numSlice,
		int numRow,
		int numCol )
{
  return CC_label (inImg, numSlice, numRow, numCol, 
		   (compfunc) cmp_uchar_pix);
}




/*--------------------------------------------------------------------------*/
/* CC_label_ushort - Find connected components of a ushort volume image     */
/*                                                                          */
/* This function finds the connected components of an unsigned char volume  */
/* image and returns a stack which contains a pointer to a seed pixel for   */
/* each region found.  See CC_label for more information.                   */
/*--------------------------------------------------------------------------*/

stack_t *
CC_label_ushort( void ***inImg,
		 int numSlice,
		 int numRow,
		 int numCol )
{
  return CC_label (inImg, numSlice, numRow, numCol, 
		   (compfunc) cmp_ushort_pix);
}




/****************************************************************************/
/* part 3: generic private functions                                        */
/****************************************************************************/


/*--------------------------------------------------------------------------*/
/* CC_add_to_neighbors - Push a pixel onto the neighbors stack.             */
/*                                                                          */
/* The algorithm for finding connected components of a image works by       */
/* recursively finding neighbors of a seed pixel.  This function pushes a   */
/* pixel location onto a stack of pixels whose neighbors are yet to be      */
/* examined.                                                                */
/*--------------------------------------------------------------------------*/

static void
CC_add_to_neighbors( int slice,
		     int row,
		     int col,
		     stack_t *neighbors )
{

    CC_pixel3d_t *pixel;


    pixel = (CC_pixel3d_t *) mallock(sizeof(CC_pixel3d_t));
    pixel->slice = slice;
    pixel->row = row;
    pixel->col = col;
    LIST_pushStack (neighbors, pixel);
}



/*--------------------------------------------------------------------------*/
/* CC_get_from_neighbors - Pop a pixel from the neighbors stack             */
/*                                                                          */
/* The algorithm for finding connected components of a image works by       */
/* recursively finding neighbors of a seed pixel.  This function pops the   */
/* location of an unexamined pixel from the neighbors stack.                */
/*--------------------------------------------------------------------------*/

static int
CC_get_from_neighbors( int *slice,
		       int *row,
		       int *col,
		       stack_t *neighbors )
{

  CC_pixel3d_t *pixel;


  if ((pixel = LIST_popStack (neighbors))) {
    *slice = pixel->slice;
    *row = pixel->row;
    *col = pixel->col;
    free (pixel);
    return 1;       
  }
  else
    return 0;
}



/*--------------------------------------------------------------------------*/
/* CC_add_all_neighbors - Push all matching neighbors onto the stack        */
/*                                                                          */
/* This function examines all neighbors of a given pixel and adds all       */
/* those bearing the label curLabel to the neighbors stack for later        */
/* (recursive) examination within the connected components algorithm.       */
/*--------------------------------------------------------------------------*/

static void
CC_add_all_neighbors( 
   void ***inImg,       /* Image to find connected components in */
   int numSlice,        /* Dimenstions of the image */
   int numRow,
   int numCol, 
   int slice,           /* Location of the given pixel */
   int row,
   int col,
   stack_t *neighbors,  /* A stack of unexamined pixels */
   void *curLabel,      /* Label to check for */
   compfunc lblcmp )    /* function comparing labels  */
{

  if (slice>0 && ! (*lblcmp)(inImg[slice-1][row], col, curLabel, 0))
    CC_add_to_neighbors (slice-1, row, col, neighbors);
  if (row>0 && ! (*lblcmp)(inImg[slice][row-1], col, curLabel, 0))
    CC_add_to_neighbors (slice, row-1, col, neighbors);
  if (col>0 && ! (*lblcmp)(inImg[slice][row], col-1, curLabel, 0))
    CC_add_to_neighbors (slice, row, col-1, neighbors);

  if (col<numCol-1 && ! (*lblcmp)(inImg[slice][row], col+1, curLabel, 0))
    CC_add_to_neighbors (slice, row, col+1, neighbors);
  if (row<numRow-1 && ! (*lblcmp)(inImg[slice][row+1], col, curLabel, 0))
    CC_add_to_neighbors (slice, row+1, col, neighbors);
  if (slice<numSlice-1 && ! (*lblcmp)(inImg[slice+1][row], col, curLabel, 0))
    CC_add_to_neighbors (slice+1, row, col, neighbors);
}



/*--------------------------------------------------------------------------*/
/* CC_flood - Flood connected component given a seed pixel                  */
/*                                                                          */
/* Given the location of a seed pixel and a current label, this function    */
/* finds all pixels recursively connected to the seed pixel and bearing     */
/* the current label and relabels them with a new label.  It returns the    */
/* number of pixels in the region.                                          */
/*--------------------------------------------------------------------------*/

static int
CC_flood( void ***inImg,        /* Image to examine and modify */
	  int numSlice,         /* Dimensions of the image */
	  int numRow,
	  int numCol,
	  int seedSlice,        /* Location of the seed pixel */
	  int seedRow,
	  int seedCol,
	  void *curLabel,	/* pointers to labels.        */
	  void *newLabel,
	  compfunc lblcmp,	/* function comparing labels  */
	  copyfunc lblcpy )	/* function copying labels     */
{
  int     slice, row, col, regSize = 0;
  stack_t *neighbors = LIST_newStack();


  /*  Initialize of neighbors stack using the seed pixel */
  if (! (*lblcpy)(inImg[seedSlice][seedRow], seedCol, curLabel, newLabel))
    return 0;
  regSize ++;
  CC_add_all_neighbors (inImg, numSlice, numRow, numCol, seedSlice, seedRow, 
	               seedCol, neighbors, curLabel, lblcmp);

  /*  Recursively change all connected curLabel pixels to newLabel */
  while (CC_get_from_neighbors (&slice, &row, &col, neighbors)) {
    if ( (*lblcpy)(inImg[slice][row], col, curLabel, newLabel)) {
      regSize ++;
      CC_add_all_neighbors (inImg, numSlice, numRow, numCol, 
	  	           slice, row, col, neighbors, curLabel, lblcmp);
    }
  }

  /* Free the stack and return the number of pixels in the region */
  LIST_freeStack (neighbors);
  return regSize;
}


/*-------------------------------------------------------------------------*/
/* CC_eqLabel - Find canonical label for a connected component             */
/*                                                                         */
/* The algorithm used in CC_label will sometimes give different labels to  */
/* two regions that are later discovered to be connected.  When this       */
/* happens, one of the labels is chosen as the canonical label and         */
/* recorded in the table eqLabelTable. This function, which is used only   */
/* by CC_label, finds the canonical equivalent of a given label.           */
/*-------------------------------------------------------------------------*/

static unsigned short
CC_eqLabel( unsigned short label,
	    unsigned short *eqLabelTable )
{

  if (eqLabelTable[label] == label)
    return label;
  else
    return CC_eqLabel (eqLabelTable[label], eqLabelTable);
}


/*--------------------------------------------------------------------------*/
/* CC_label - Find connected components of a volume image                   */
/*                                                                          */
/* This generic function implements CC_label_uchar and CC_label_ushort, and */
/* is specialized to the desired image data type by providing a suitable    */
/* function to compare labels.  It returns a stack of pixel locations, which*/
/* are seed pixels for the regions found.  The maximum number of distinct   */
/* labels used must not exceed MAX_USHORT (not to be confused with the ANSI */
/* C macro USHORT_MAX).  Regions that initially appear distinct will be     */
/* relabelled if smaller than SMALL_COMPONENT_SIZE; if so, the redundant    */
/* label will be recycled.                                                  */
/*                                                                          */
/* The stack generated and returned consists of pointers to CC_reg_seed_t   */
/* structs, which contain the coordinates of a seed pixel for the region    */
/* and the number of pixels in the region.                                  */
/*--------------------------------------------------------------------------*/

/* FIXME:  With only a little more effort, this function could return a
   image with the connected components already labelled, saving the caller
   the effort of recomputing them from the seed pixels.   */

static stack_t *
CC_label( void ***inImg,        /* Image to be processed */
	  int numSlice,         /* Dimensions of the image */
	  int numRow,
	  int numCol,
	  compfunc lblcmp )	/* function comparing labels */
{
  int            slice, row, col, index, numPix=numSlice*numRow*numCol;
  int            sliceSize=numRow*numCol;
  int            *eqSizeTable=0, notAssigned;
  unsigned short *eqLabelTable=0, ***tmpLabels, *ptr_t=0, **ptr_t2=0;
  unsigned short curLabel, tmpEqLabel, curEqLabel;
  void           *ptr_s, *ptr_s2=0;
  CC_reg_seed_t  *seed;
  stack_t        *spareLabels = LIST_newStack();
  stack_t        *regionSeeds = LIST_newStack();


  /* Allocate a volume image of temporary region labels */
  tmpLabels = (unsigned short ***) mallock(numSlice*sizeof (void *));
  ptr_t2 = (unsigned short **) mallock(numSlice*numRow*sizeof(void *));
  ptr_t = (unsigned short *) mallock(numPix*sizeof(short));

  /* Define pointer arrays for multi-dimensional indexing */
  for (slice=0; slice<numSlice; slice++) {
    tmpLabels[slice] = ptr_t2;
    for (row=0; row<numRow; row++, ptr_t2++, ptr_t+=numCol)
      *ptr_t2 = ptr_t;
  }

  /* Allocate tables for canonical label and size of each region */
  eqLabelTable = (unsigned short *) mallock(MAX_USHORT*sizeof(short));
  eqSizeTable = (int *) mallock( MAX_USHORT*sizeof(int));
  memset (tmpLabels[0][0], 0, numPix*sizeof(short));
  memset (eqLabelTable, 0, MAX_USHORT*sizeof(short));
  memset (eqSizeTable, 0, MAX_USHORT*sizeof(int));

/*--------------------------------------------------------------------------*/
/* Find all connected components                                            */
/*--------------------------------------------------------------------------*/

  /*** Loop over all the pixels in the image ***/
  curLabel=1;
  for (slice=0; slice<numSlice; slice++) {
    ptr_s = inImg[slice][0];  
    if (slice)
      ptr_s2 = inImg[slice-1][0];
    ptr_t = tmpLabels[slice][0];
    for (row=0, index=0; row<numRow; row++) {
      for (col=0; col<numCol; col++, index++, ptr_t++) {
        notAssigned = 1;

        /* If this pixel matches the one to its left, assign it to 
           the same region as the pixel to the left. */
	if (col && ! (*lblcmp)(ptr_s, index, ptr_s, index-1)) {
	  tmpEqLabel = CC_eqLabel (*(ptr_t-1), eqLabelTable);
	  *ptr_t = tmpEqLabel;
	  eqSizeTable[tmpEqLabel] ++;
	  notAssigned = 0;
	}

        /* Compare this pixel to the one above it */
	if (row && ! (*lblcmp)(ptr_s, index, ptr_s, index-numCol)) {
	  tmpEqLabel = CC_eqLabel (*(ptr_t-numCol), eqLabelTable);

          /* If the labels match and this pixel hasn't already been
             assigned, assign it to the same region as the pixel above */
          if (notAssigned) {
	    *ptr_t = tmpEqLabel;
	    eqSizeTable[tmpEqLabel] ++;
	  }

          /* If the labels match, but this pixel has already been
             assigned to a different region, it means that the two
             regions aren't really different and should be merged. */
          else if (*ptr_t != tmpEqLabel) {

            /* If either of the regions is small, relabel its pixels
               and recyle its old label.  */
	    if (eqSizeTable[*ptr_t] <= eqSizeTable[tmpEqLabel] &&
	        eqSizeTable[*ptr_t] < SMALL_COMPONENT_SIZE) {
              if (eqLabelTable[*ptr_t] != *ptr_t) {
		  fatal ("CC_mergeSmallReg: check code (1).");
	      }
	      eqSizeTable[tmpEqLabel] += eqSizeTable[*ptr_t];
	      eqSizeTable[*ptr_t] = 0;
	      LIST_pushStack (spareLabels, &(eqLabelTable[*ptr_t]));
	      curEqLabel = *ptr_t;
	      CC_flood ((void ***) tmpLabels, numSlice, numRow, numCol,
		        slice, row, col, &curEqLabel, &tmpEqLabel,
			(compfunc) cmp_ushort_pix,
			(copyfunc) cpy_ushort_pix);
	    }
	    else if (eqSizeTable[tmpEqLabel] < eqSizeTable[*ptr_t] &&
		     eqSizeTable[tmpEqLabel] < SMALL_COMPONENT_SIZE) {
              if (eqLabelTable[tmpEqLabel] != tmpEqLabel) {
		  fatal ("CC_mergeSmallReg: check code (2).");
	      }
  	      eqSizeTable[*ptr_t] += eqSizeTable[tmpEqLabel];
	      eqSizeTable[tmpEqLabel] = 0;
	      LIST_pushStack (spareLabels, &(eqLabelTable[tmpEqLabel]));
	      curEqLabel = *ptr_t;
	      CC_flood ((void ***) tmpLabels, numSlice, numRow, numCol,
			slice, row-1, col, &tmpEqLabel, &curEqLabel, 
			(compfunc) cmp_ushort_pix,
			(copyfunc) cpy_ushort_pix);
	    }

            /* If both regions are too large to relabel efficiently,
               define the canonical label as the smaller label and
               update the table of canonical labels. */
	    else if (*ptr_t < tmpEqLabel) {
	      eqLabelTable[tmpEqLabel] = *ptr_t;
	      eqSizeTable[*ptr_t] += eqSizeTable[tmpEqLabel];
	      eqSizeTable[tmpEqLabel] = 0;
	    }
	    else if (*ptr_t > tmpEqLabel) {
	      eqLabelTable[*ptr_t] = tmpEqLabel;
	      eqSizeTable[tmpEqLabel] += eqSizeTable[*ptr_t];
	      eqSizeTable[*ptr_t] = 0;
	    }
	  }
	  notAssigned = 0;
	}

        /* Compare the current pixel to the one in the previous slice
           and treat in the same fashion as for the pixel above. */
	if (slice && ! (*lblcmp)(ptr_s, index, ptr_s2, index)) {
          tmpEqLabel = CC_eqLabel (*(ptr_t-sliceSize), eqLabelTable);
	  if (notAssigned) {
	    *ptr_t = tmpEqLabel;
	    eqSizeTable[tmpEqLabel] ++;
	  }
	  else if ((curEqLabel=CC_eqLabel(*ptr_t, eqLabelTable)) != 
		   tmpEqLabel) {
	    if (eqSizeTable[curEqLabel] <= eqSizeTable[tmpEqLabel] &&
	        eqSizeTable[curEqLabel] < SMALL_COMPONENT_SIZE) {
              if (eqLabelTable[*ptr_t] != *ptr_t) {
		  fatal ("CC_mergeSmallReg: check code (3).");
	      }
	      eqSizeTable[tmpEqLabel] += eqSizeTable[*ptr_t];
	      eqSizeTable[*ptr_t] = 0;
	      LIST_pushStack (spareLabels, &(eqLabelTable[*ptr_t]));
	      CC_flood ((void ***) tmpLabels, numSlice, numRow, numCol,
		        slice, row, col, &curEqLabel, &tmpEqLabel,
			(compfunc) cmp_ushort_pix,
			(copyfunc) cpy_ushort_pix);
	    }
	    else if (eqSizeTable[tmpEqLabel] < eqSizeTable[curEqLabel] &&
		     eqSizeTable[tmpEqLabel] < SMALL_COMPONENT_SIZE) {
              if (eqLabelTable[tmpEqLabel] != tmpEqLabel) {
		  fatal ("CC_mergeSmallReg: check code (4).");
	      }
  	      eqSizeTable[curEqLabel] += eqSizeTable[tmpEqLabel];
	      eqSizeTable[tmpEqLabel] = 0;
	      LIST_pushStack (spareLabels, &(eqLabelTable[tmpEqLabel]));
	      CC_flood ((void ***) tmpLabels, numSlice, numRow, numCol,
			slice-1, row, col, &tmpEqLabel, &curEqLabel, 
			(compfunc) cmp_ushort_pix,
			(copyfunc) cpy_ushort_pix);
	    }
	    else if (curEqLabel < tmpEqLabel) {
	      eqLabelTable[tmpEqLabel] = curEqLabel;
	      eqSizeTable[curEqLabel] += eqSizeTable[tmpEqLabel];
	      eqSizeTable[tmpEqLabel] = 0;
	    }
	    else if (curEqLabel > tmpEqLabel) {
	      eqLabelTable[curEqLabel] = tmpEqLabel;
	      eqSizeTable[tmpEqLabel] += eqSizeTable[curEqLabel];
	      eqSizeTable[curEqLabel] = 0;
	    }
	  }
	  notAssigned = 0;
	}

	/* Get the next region label, either recycling one that was
           found to be redundant, or taking the next one never used. */
	if (notAssigned) {
          unsigned short *ptmp = LIST_popStack (spareLabels);
	  if (ptmp) {
	    *ptr_t = *ptmp;
 	    eqLabelTable[*ptmp] = *ptmp;
	    eqSizeTable[*ptmp] = 1;
	  }
	  else {
	    *ptr_t = curLabel;
	    eqLabelTable[curLabel] = curLabel;
	    eqSizeTable[curLabel] = 1;
	    curLabel ++;
	    if (curLabel >= MAX_USHORT) {
		fatal ("CC_label: too many components.");
	    }
	  }
	}	

      }
    }
  }
  /*** End of loop over all the pixels ***/


  /* Generate a stack of seed pixels for all the connected components. */
  ptr_t = tmpLabels[0][0];
  for (slice=0; slice<numSlice; slice++) {
    for (row=0; row<numRow; row++) {
      for (col=0; col<numCol; col++, ptr_t++) {
	if (eqSizeTable[*ptr_t] > 0) {
          seed = (CC_reg_seed_t *) mallock(sizeof(CC_reg_seed_t));
	  seed->slice = slice;
	  seed->row = row;
	  seed->col = col;
	  seed->regSize = eqSizeTable[*ptr_t];
	  LIST_pushStack (regionSeeds, seed);
	  eqSizeTable[*ptr_t] = 0;
	}
      }
    }
  }
  free (eqSizeTable);
  free (eqLabelTable); 
  free (tmpLabels[0][0]);
  free (tmpLabels[0]);
  free (tmpLabels);
  while (LIST_popStack(spareLabels));
  LIST_freeStack (spareLabels);
  return regionSeeds;
}

