/*--------------------------------------------------------------------------
find_mask.c                                                              

Programmer: Michelle Yan                                          

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

   This source code contains some basic functions for morphological       
filtering, tracing contours and finding a mask.                          
   Including:                                                             
   void B_Coordinate3D()                 : compute sequences of 3D        
                                           BOUNDARY coordinates and store 
                                           them into 2D array.            
   int find_mask()                       : find BACKGROUND, BOUNDARY      
                                           and OBJECT of 2D binary image  
   int boundary_pixel()                  : find a sequence of coordinates 
                                           boundary pixels, it should be  
                                           used after find_mask().        
   void dilate()                         : dilate a NXM image with 4      
                                           unfixed elements.              
   void dilate_8()                       : dilate a NXM image with 8 fixed
                                           elements.                      
   void dilate3D()                       : dilate a NXMXL image with rad. 
                                           of square root of 2.           
   void dilateLabel3D()                  : dilate a NXMXL label image with
                                           radius of square root of 2.    
   void erode()                          : erode a NXM image with 4       
                                           unfixed elements.              
   void erode_8()                        : erode a NXM image with 8 fixed 
                                           elements.                      
   void erode3D()                        : erode a NXMXL image with radius
                                           of square root of 2.           
   struct coordinate2 rotate_direction() : find an adjacent neighbor      
                                           having OBJECT value.           
   struct coordinate2 next_direction()   : find next dir. by rotating     
                                           the current one 90 degree.     
   struct coordinate2 search_contour()   : trace a OBJECT contour, return 
                                           coordinate of an interior      
                                           point.                         
   int PutN_Stack()       : put a new neighbor pixel at the top of        
                            N_Stack.                                      
   int PotN_Stack()       : get the top pixel from N_Stack.               
   int UpdateN_Stack()    : add neighbor pixels of a given pixel to       
                            N_Stack.                                      
   int FillContour()      : fill OBJECT pixels enclosed by a contour      
--------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: find_mask.c,v 1.12 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>

#include "common.h"
#include "definition.h"
#include "errm.h"
#include "find_mask.h"
#include "image_inc.h"
#include "image_io.h"
#include "memm.h"
#include "region_grow.h"


/* Local functions */
static struct coordinate2  rotate_direction(struct coordinate2 cur_dir);
static struct coordinate2  next_direction(struct coordinate2 cur_dir);
static struct coordinate2  search_contour (float **LabelP, int N, int M, 
   int *ContourSize);
static int  PutN_Stack(int i, int j);
static int  PopN_Stack(int *i, int *j);
static int  UpdateN_Stack(float **LabelP, int row, int col, int i, int j);
static int  FillContour (float **LabelP, int RowNum, int ColNum, int Seed_i, 
   int Seed_j, int FeatureV);



/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*    it finds sequences of BOUNDARY coordinates, and store them into 2D    */
/* array, write the 2D array into a file. Output (original + contour) image,*/
/* whose image type = ImgType_In, is an option                              */
/*--------------------------------------------------------------------------*/

void
B_Coordinate3D ( char *Name_In,
		 int ImgType_In,
		 char *Overlay_YN,
		 char *OuterFirst_YN, 
		 char *UpperT_YN,
		 int SmoothOption,
		 int TotalSlice,
		 int FirstSlice, 
		 int LastSlice,
		 int N,		/* Number of rows        */
		 int M,		/*  and columns  */
		 float Res_XY,	/* Voxel size (mm)^3 of input image   */
		 float Res_Z,
		 struct coordinate2 ***B_Coordinate3P,
		 int **B_SizeP ) /* Its length = total used slice
				    No. + 1, First element is stored
				    total used slice no. */

{

    char          *Ptr;
    FILE          *FileP_In, *FileP_BC, *FileP_Overlay=0;
    float         ***Image,           /* Pointer to 3D image                */
                  **Contour, 
                  **TmpImg;
    float         Parameter[7];       /* Image dimension array for output   */
    float         *ptr_1, *ptr_2;
    int           NameLen;
    int           BoundarySize, SliceNo;
    int           i, k; 
    unsigned char ***ScaledImg, ***Mask_1, ***Mask_2, *ptr_c;
    

/*--------------------------------------------------------------------------*/
/*Open input file for reading                                               */
/*--------------------------------------------------------------------------*/

    FileP_In = open_file(Name_In, "r");

    NameLen = strlen(Name_In) - 4;
    if ( (Ptr = calloc (NameLen + 8, sizeof(char))) == NULL) {
	fatal("B_Coordinate3D: cannot allocate memory for a file name.");
    }
    strncpy (Ptr, Name_In, NameLen);
    strcat (Ptr, ".BC");
    FileP_BC = open_file(Ptr, "w");

    if ( strcmp ( Overlay_YN, "y") == 0 ) {
	strncpy (Ptr, " ", NameLen+8);
	strncpy (Ptr, Name_In, NameLen);
	strcat ( Ptr, "_bi.img");
	FileP_Overlay = open_file (Ptr, "w");
    }
    free(Ptr);

/*--------------------------------------------------------------------------*/
/*Allocate memory for NXM & NXMXL & arrays for BOUNDARY size & coordinates  */
/*--------------------------------------------------------------------------*/
    Contour = alloc_img(N, M);
    TmpImg = alloc_img(N, M);
    Image = alloc_3Dimg(N, M, TotalSlice);
    ScaledImg = alloc_3Dchar(N, M, TotalSlice);
    Mask_1 = alloc_3Dchar(N, M, TotalSlice);
    Mask_2 = alloc_3Dchar(N, M, TotalSlice);

    SliceNo = LastSlice - FirstSlice + 1;
    if ( (Ptr = calloc(SliceNo + 1, sizeof( int ))) == NULL ) {
	fatal("B_Coordinate3D: cannot allocate memory for BOUNDARY size.");
    }
    *B_SizeP = (int *) Ptr;
    (*B_SizeP)[0] = SliceNo;
    if ( (Ptr = calloc(SliceNo, sizeof( struct coordinate2 * ))) == NULL ) {
	fatal("B_Coordinate3D: cannot allocate memory for coordinates.");
    }
    *B_Coordinate3P = (struct coordinate2 **) Ptr;

/*--------------------------------------------------------------------------*/
/* Read the entire 3D image & generate a binary image & extract VOI         */
/*--------------------------------------------------------------------------*/
    for ( k=0; k<TotalSlice; k++) {
	read_matrix(FileP_In, Image[k], N, M, ImgType_In);
    }
    
    Scale2Char3D (Image, ScaledImg, N, M, TotalSlice, SmoothOption);
    BinaryImg3D (ScaledImg, Mask_1, N, M, TotalSlice, UpperT_YN);
    free_3Dimg (Image, TotalSlice);

    ExtractVOI (Mask_1, Mask_2, N, M, TotalSlice, OuterFirst_YN);
    free_3Dchar (Mask_1, TotalSlice);

/*--------------------------------------------------------------------------*/
/* write image slices from 1 to (FirstSlice - 1)                             */
/*--------------------------------------------------------------------------*/
    if ( strcmp ( Overlay_YN, "y") == 0 ) {
	for ( k=0; k< FirstSlice-1; k++ ) {
	    for ( i=0, ptr_c=ScaledImg[k][0], ptr_1=TmpImg[0];
		  i<N*M; i++, ptr_c++, ptr_1++ ) 
		*ptr_1 = *ptr_c;
	    write_matrix (FileP_Overlay, TmpImg, N, M, ImgType_In);
	}
    }

/*--------------------------------------------------------------------------*/
/* Start at FirstSlice, find contours for (LastSlice-FirstSlice+1)          */
/*--------------------------------------------------------------------------*/
    for ( k=FirstSlice-1; k<LastSlice; k++) {
	for ( i=0, ptr_c=Mask_2[k][0], ptr_1=TmpImg[0]; 
	      i<N*M; i++, ptr_c++, ptr_1++ ) {
	    *ptr_1 = *ptr_c;
	}
	BoundarySize = find_mask(TmpImg, Contour, N, M, "CONTOUR");
/*--------------------------------------------------------------------------*/
/*      writing out overlayed image (original + contour)                    */
/*--------------------------------------------------------------------------*/
	if ( strcmp(Overlay_YN, "y") == 0 ) {
	    ptr_c = ScaledImg[k][0];
	    ptr_1 = Contour[0];
	    ptr_2 = TmpImg[0];
	    for ( i=0; i<N*M; i++,  ptr_c++, ptr_1++, ptr_2++ ) {
		if ( *ptr_1 == BOUNDARY )
		    *ptr_2 = OBJECT;
		else 
		    *ptr_2 = *ptr_c;
	    }
	    write_matrix (FileP_Overlay, TmpImg, N, M, ImgType_In);
	}

/*--------------------------------------------------------------------------*/
/*      allocate memory & search a sequence of BOUNDARY coordinates.        */
/*--------------------------------------------------------------------------*/
	(*B_SizeP)[k-FirstSlice+2] = BoundarySize;
	printf("Total number of pixels at %d-th BOUNDARY: %d\n", 
	       k+1, BoundarySize);

	Ptr = calloc(BoundarySize, sizeof(struct coordinate2));
	if ( Ptr == NULL ) {
	    fatal("B_Coordinate3D: cannot allocate memory B_coordinates.");
	}
	(*B_Coordinate3P)[k-FirstSlice+1] = (struct coordinate2 *) Ptr;
	boundary_pixel(Contour, (*B_Coordinate3P)[k-FirstSlice+1], N, M);
    }
/*--------------------------------------------------------------------------*/
/*  write image slices from LastSlice+1 to TotalSlice                       */
/*--------------------------------------------------------------------------*/
    if ( strcmp ( Overlay_YN, "y") == 0 ) {
	for ( k=LastSlice; k<TotalSlice; k++ ) {
	    for ( i=0, ptr_c=ScaledImg[k][0], ptr_1=TmpImg[0];
		  i<N*M; i++, ptr_c++, ptr_1++ ) 
		*ptr_1 = *ptr_c;
	    write_matrix (FileP_Overlay, TmpImg, N, M, ImgType_In);
	}
    }
/*--------------------------------------------------------------------------*/
/*  Write Dimension, BOUNDARY size and coordinates to a file                */
/*--------------------------------------------------------------------------*/
    Parameter[0] = N;
    Parameter[1] = M;
    Parameter[2] = FirstSlice;
    Parameter[3] = LastSlice;
    Parameter[4] = TotalSlice;
    Parameter[5] = Res_XY;
    Parameter[6] = Res_Z;

    fwrite(Parameter, 7*sizeof(float), 1, FileP_BC);
    fwrite(*B_SizeP, (SliceNo + 1)*sizeof(float), 1, FileP_BC);

    for ( i=0, BoundarySize=(*B_SizeP)[1]; i<(*B_SizeP)[0]; i++ ) {
	if ( (*B_SizeP)[i+1] > BoundarySize ) 
	    BoundarySize = (*B_SizeP)[i+1];
    }

    if ( (ptr_1=(float *)calloc(BoundarySize, sizeof(float))) == NULL) {
	fatal("B_Coordinate3D: cannot allocate memory I_coordinates.");
    }
    if ( (ptr_2=(float *)calloc(BoundarySize, sizeof(float))) == NULL) {
	fatal("B_Coordinate3D: cannot allocate memory J_coordinates.");
    }
    for ( k=0; k<(*B_SizeP)[0]; k++ ) {
	for ( i=0; i<(*B_SizeP)[k+1]; i++) {
	    ptr_1[i] = (*B_Coordinate3P)[k][i].i;
	    ptr_2[i] = (*B_Coordinate3P)[k][i].j;
	}
	fwrite(ptr_1, (*B_SizeP)[k+1]*sizeof(float), 1, FileP_BC);
	fwrite(ptr_2, (*B_SizeP)[k+1]*sizeof(float), 1, FileP_BC);
    }

/*--------------------------------------------------------------------------*/
/*Close opened files and release memory                                     */
/*--------------------------------------------------------------------------*/
    fclose (FileP_In);
    fclose (FileP_BC);
    if ( strcmp ( Overlay_YN, "y") == 0 ) {
	fclose (FileP_Overlay);
    }

    free(ptr_1);
    free(ptr_2);
    free_3Dchar(ScaledImg, TotalSlice);
    free_3Dchar(Mask_2, TotalSlice);
    free_img(Contour);
    free_img(TmpImg);
}



/*--------------------------------------------------------------------------
find_mask - Construct 2D mask image showing background, object, and boundary

This source code generates a 2D mask image indicating boundary,
object enclosed by boundary and background.                     

SUGGESTION: It should be used after Binary3D() & ExtractVOI()

The "CONTOUR" option seems to be the only one ever used.

??? As near as I can make out, this function does repeated open/close
morphological operations to smooth the mask and then removes (exactly)
three small regions to generate the final mask.  I am not sure what
happens to the boundary points.  -PWH

??? This function is probably tuned to some specific application, since
it has exactly three passes of morphological smoothing and removes exactly
three objects; the constants in MinSize may also be task-specific.
---------------------------------------------------------------------------*/

int
find_mask(Mask_In, Mask_Out, N, M, CONTOUR_MASK)
   char       *CONTOUR_MASK;          /* "MASK" or "CONTOUR" */
   float      **Mask_In, **Mask_Out;  /* Pointers to input & output mask */
   int        N, M;                   /* Number of rows and columns */
{
    float              **TmpImg, *ptr_1, *ptr_2;
    int                ObjSize, CntSize, RegSize=UNDEFINED, 
                       Size=UNDEFINED, MinSize=UNDEFINED, NoRepeat; 
    int                i, k;
    struct coordinate2 interior_pix;      /* seed for region growing        */

    TmpImg = alloc_img(N, M);

   /* Apply morphological operations to smooth boundaries, fill holes and */
   /* remove isolated pixels                                              */

    memcpy (Mask_Out[0], Mask_In[0], N*M*sizeof(float));
    for (k=0; k<3; k++) {
	erode(Mask_Out, TmpImg, N, M);  
	dilate(TmpImg, Mask_Out, N, M);

	dilate(Mask_Out, TmpImg, N, M);
	erode(TmpImg, Mask_Out, N, M);  
    }


    /*** Repeatedly look for objects within the binary image ***/
    /* If I have this right, this loop removes exactly three small
       regions from the mask and then exits.  The rationale is
       obscure.  -PWH */
    memcpy (TmpImg[0], Mask_Out[0], N*M*sizeof(float));
    NoRepeat = 0;
    do {

        /* If there are fewer than 10 OBJECT pixels left in the
           image, exit immediately.  FIXME:  Shouldn't this clean
           up first, like freeing TmpImg? */ 
	ObjSize = 0;
	for ( i=0, ptr_1=Mask_Out[0]; i<N*M; i++, ptr_1++ ) {
	    if ( *ptr_1 == OBJECT )
		ObjSize ++;
	}
	if ( ObjSize < 10 ) 
	    return 0;

        /* Decide the appropriate minimum size, depending on
           whether I am looking for a MASK or a CONTOUR. */
	if ( (strcmp(CONTOUR_MASK, "MASK") == 0 ) )
	    MinSize = (int) (ObjSize * 0.1);
	else if ( (strcmp(CONTOUR_MASK, "CONTOUR") == 0 ) )
	    MinSize = (int) (sqrt(ObjSize*0.1) * 4.0);

        /* Find a seed and a BOUNDARY of object:  search_contour
           will find and label an OBJECT/BACKGROUND boundary, label
           its pixels as type BOUNDARY, and return an interior
           OBJECT pixel as a seed.  */
	interior_pix = search_contour (Mask_Out, N, M, &CntSize);

/*	printf("Seed for filling inside of contour = (%d, %d) \n", 
	       interior_pix.i, interior_pix.j); */

        /* If a contour less than the minimum size, or any mask,
           change all OBJECT pixels to BACKGROUND and fill the
           region inside the BOUNDARY contour with OBJECT pixels.
           FIXME:  If this isn't the right region, we've just
           erased all the other objects in the image. */
	if ( (strcmp(CONTOUR_MASK, "CONTOUR") == 0 && CntSize<MinSize)
	     || (strcmp(CONTOUR_MASK, "MASK") == 0) ) { 

            /* Fill inside of BOUNDARY by OBJECT */
	    if ( interior_pix.i != UNDEFINED && 
		 interior_pix.j != UNDEFINED ) {
		for (ptr_1=Mask_Out[0], i=0; i<N*M; i++, ptr_1++) {
		    if (*ptr_1 == OBJECT) {
			*ptr_1 = BACKGROUND;
		    }
		}
		RegSize = FillContour(Mask_Out, N, M, interior_pix.i, 
				      interior_pix.j, OBJECT);
	    }
	    else { 
		RegSize = UNDEFINED;
	    }
	}

        /* The size is either the number of points in the contour, or
           the number of pixels in the region. */
	if ( strcmp(CONTOUR_MASK, "CONTOUR") == 0 )
	    Size = CntSize;
	else if ( strcmp(CONTOUR_MASK, "MASK") == 0 )
	    Size = RegSize;

        /* If the region is small enough, convert it to background
           and restore the rest of the image from TmpImg. */
	if ( Size < MinSize && NoRepeat < 3 ) {
	    NoRepeat ++;
	    if ( RegSize != UNDEFINED ) {
		for ( ptr_1=Mask_Out[0], ptr_2=TmpImg[0], i=0; 
 		      i<N*M; ptr_1++, ptr_2++, i++ ) {
		    if ( *ptr_1 != BACKGROUND ) {
			*ptr_1 = BACKGROUND;
			*ptr_2 = BACKGROUND;
		    }
		    else {
			*ptr_1 = *ptr_2;
		    }
		}
	    }
	    else {
		for ( ptr_1=Mask_Out[0], ptr_2=TmpImg[0], i=0; 
		      i<N*M; ptr_1++, ptr_2++, i++ ) {
		    if ( *ptr_1 == BOUNDARY ) {
			*ptr_1 = BACKGROUND;
			*ptr_2 = BACKGROUND;
		    }
		}
	    }
	}

    } while ( Size < MinSize && NoRepeat < 3 );
    /*** End of loop looking for objects ***/

    free_img(TmpImg);

    return Size;
}


int
boundary_pixel(LabelP, B_Coordinate, N, M)
/*--------------------------------------------------------------------------*/
/* Purpose:                                                                 */
/*    find a sequence of (i,j) coordinates of boundary pixels and return    */
/* number of pixels on BOUNDARY                                             */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
float                **LabelP;
int                  N, M;
struct coordinate2   *B_Coordinate;
{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    float               Centroid_I, Centroid_J, Sum_I = 0, Sum_J = 0;
    int                 ContourSize = 0, PixelNum = 0, i, j;
    int                 end_of_search = FALSE, end_loop;

    struct coordinate2  init_pos, cur_pos, cur_dir;
/*--------------------------------------------------------------------------*/
/* Compute the centroid of BOUNDARY                                         */
/*--------------------------------------------------------------------------*/
    for ( i=0; i<N; i++ ) {
	for ( j=0; j<M; j++ ) {
	    if ( LabelP[i][j] == BOUNDARY ) {
		Sum_I += i;
		Sum_J += j;
		ContourSize++;
	    }
	}
    }
    Centroid_I = Sum_I/ContourSize;
    Centroid_J = Sum_J/ContourSize;

/*--------------------------------------------------------------------------*/
/* Find an initial BOUNDARY pixel                                           */
/*--------------------------------------------------------------------------*/
    init_pos.j = (int) ( Centroid_J + 0.5);
    init_pos.i = 0;
    while ( init_pos.i<N && LabelP[init_pos.i][init_pos.j] != BOUNDARY ) {
	init_pos.i ++;
    }
    if ( init_pos.i == N ) {
	fatal ("boundary_pixel: cannot find a starting point on BOUNDARY.");
    }

    B_Coordinate[PixelNum].i = init_pos.i;
    B_Coordinate[PixelNum].j = init_pos.j;
    PixelNum ++;
    cur_pos.i = init_pos.i;
    cur_pos.j = init_pos.j;
    cur_dir.i = -1;
    cur_dir.j = 1;

/*--------------------------------------------------------------------------*/
/* Trace contour starting from initial boundary point.                      */
/*--------------------------------------------------------------------------*/
    while (!end_of_search) {
	end_loop = FALSE;
	while (!end_loop) {
	    if (LabelP[cur_pos.i+cur_dir.i][cur_pos.j+cur_dir.j] == BOUNDARY)
		end_loop = TRUE;
	    else
		cur_dir = rotate_direction(cur_dir);
	}
	cur_pos.i = cur_pos.i + cur_dir.i;
	cur_pos.j = cur_pos.j + cur_dir.j;
	if (cur_pos.i == init_pos.i && cur_pos.j == init_pos.j)
	    end_of_search = TRUE;
	else {
	    B_Coordinate[PixelNum].i = cur_pos.i; 
	    B_Coordinate[PixelNum].j = cur_pos.j; 
	    PixelNum ++;
	    cur_dir = next_direction (cur_dir);
	}
    }
    if ( PixelNum != ContourSize ) {
	printf("%d %d \n", PixelNum, ContourSize);
	fatal ("boundary_pixel: pixel numbers differ from search_contour.");
    }
    return ContourSize;
}


void
dilate(InImgP, OutImgP, N, M)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*    Dilate image                                                          */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
float   **InImgP, **OutImgP;
int     N, M;                               /* Number of rows and columns   */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j;
    float         sum, three_obj;
    float         *Ptr_in, *Ptr_out;

    three_obj = 3.0*OBJECT;
    Ptr_in = InImgP[0];
    Ptr_out = OutImgP[0];
    for(i=0; i<N*M; Ptr_in++, Ptr_out++, i++)
	*Ptr_out = *Ptr_in;
/*--------------------------------------------------------------------------*/
/* Dilate the middle (N-2)X(M-2) BACKGROUND, if 4 or more neighbors are     */
/* OBJECT.                                                                  */
/*--------------------------------------------------------------------------*/
    for(i=1; i<N-1; i++) 
	for (j=1; j<M-1; j++) 
	    if (InImgP[i][j] == BACKGROUND) {
		sum = InImgP[i-1][j-1] + InImgP[i-1][j] + InImgP[i-1][j+1] +
		      InImgP[i][j-1] + InImgP[i][j+1] +
		      InImgP[i+1][j-1] + InImgP[i+1][j] + InImgP[i+1][j+1];
		if (sum > three_obj) 
		    OutImgP[i][j] = OBJECT;
	    }
}


void
dilate_8(InImgP, OutImgP, N, M)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*    Dilate image with fixed 8 neighbors.                                  */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
float   **InImgP, **OutImgP;
int     N, M;                               /* Number of rows and columns   */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j;
    float         sum;
    float         *Ptr_in, *Ptr_out;

    Ptr_in = InImgP[0];
    Ptr_out = OutImgP[0];
    for(i=0; i<N*M; Ptr_in++, Ptr_out++, i++)
	*Ptr_out = *Ptr_in;
/*--------------------------------------------------------------------------*/
/* Dilate the middle (N-2)X(M-2) BACKGROUND, if one or more neighbors are   */
/* OBJECT.                                                                  */
/*--------------------------------------------------------------------------*/
    for(i=1; i<N-1; i++) 
	for (j=1; j<M-1; j++) 
	    if (InImgP[i][j] == BACKGROUND) {
		sum = InImgP[i-1][j-1] + InImgP[i-1][j] + InImgP[i-1][j+1] +
		      InImgP[i][j-1] + InImgP[i][j+1] +
		      InImgP[i+1][j-1] + InImgP[i+1][j] + InImgP[i+1][j+1];
		if (sum > BACKGROUND) 
		    OutImgP[i][j] = OBJECT;
	    }
}


void
dilate3D(InImgP, OutImgP, N, M, L)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*    Dilate image with a fixed square root of 2 radius neighbors.          */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/

     int N, M, L;	/* # of rows, cols & slices     */
     unsigned char ***InImgP;
     unsigned char ***OutImgP;

{

/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j, k;
    int           sum;
    unsigned char *Ptr_in, *Ptr_out;
    
  for ( k=0; k<L; k++ ) { 
    Ptr_in = InImgP[k][0];
    Ptr_out = OutImgP[k][0];
    for (i = 0; i < N*M; Ptr_in++, Ptr_out++, i++) {
      *Ptr_out = *Ptr_in;
    }
  }

/*--------------------------------------------------------------------------*/
/* Dilate the middle (N-2)X(M-2) BACKGROUND, if one or more neighbors are   */
/* OBJECT.                                                                  */
/*--------------------------------------------------------------------------*/

  for ( k=0; k<L; k++ ) { 
    for( i=1; i<N-1; i++ ) {
      for ( j=1; j<M-1; j++ ) {
	if (InImgP[k][i][j] == BACKGROUND) {
	  sum = (InImgP[k][i-1][j-1] + InImgP[k][i-1][j] + 
		 InImgP[k][i-1][j+1] + InImgP[k][i][j-1] + 
		 InImgP[k][i][j+1] + InImgP[k][i+1][j-1] + 
		 InImgP[k][i+1][j] + InImgP[k][i+1][j+1]);
	  if ( k>0 ) {
	    sum += (InImgP[k-1][i-1][j] + InImgP[k-1][i][j-1] + 
		    InImgP[k-1][i][j] + InImgP[k-1][i][j+1] + 
		    InImgP[k-1][i+1][j]);
	  }
	  if ( k<L-1 ) {
	    sum += (InImgP[k+1][i-1][j] + InImgP[k+1][i][j-1] + 
		    InImgP[k+1][i][j] + InImgP[k+1][i][j+1] + 
		    InImgP[k+1][i+1][j]);
	  }
	  if (sum > BACKGROUND) 
	    OutImgP[k][i][j] = OBJECT;
	}
      }
    }
  }
}


void
dilateLabel3D(InImgP, OutImgP, N, M, L)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*    Dilate labeled image with fixed 8 neighbors.                          */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
int           N, M, L;                      /* Number of rows and columns   */
unsigned char ***InImgP, ***OutImgP;

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j, k;
    unsigned char sum;
    unsigned char *Ptr_in, *Ptr_out;

    for ( k=0; k<L; k++ ) { 
	Ptr_in = InImgP[k][0];
	Ptr_out = OutImgP[k][0];
	for(i=0; i<N*M; Ptr_in++, Ptr_out++, i++) {
	    *Ptr_out = *Ptr_in;
	}
    }
/*--------------------------------------------------------------------------*/
/* Dilate the middle (N-2)X(M-2) BACKGROUND, if one or more neighbors are   */
/* labeled.                                                                 */
/*--------------------------------------------------------------------------*/
    for ( k=0; k<L; k++ ) { 
	for( i=1; i<N-1; i++ ) {
	    for ( j=1; j<M-1; j++ ) {
		if (InImgP[k][i][j] == BACKGROUND) {
		    sum = InImgP[k][i-1][j-1] + InImgP[k][i-1][j] + 
			  InImgP[k][i-1][j+1] + InImgP[k][i][j-1] + 
			  InImgP[k][i][j+1] + InImgP[k][i+1][j-1] + 
			  InImgP[k][i+1][j] + InImgP[k][i+1][j+1];
		    if ( k>0 ) {
			sum += (InImgP[k-1][i-1][j] + InImgP[k-1][i][j-1] + 
				InImgP[k-1][i][j] + InImgP[k-1][i][j+1] + 
				InImgP[k-1][i+1][j]);
		    }
		    if ( k<L-1 ) {
			sum += (InImgP[k+1][i-1][j] + InImgP[k+1][i][j-1] + 
				InImgP[k+1][i][j] + InImgP[k+1][i][j+1] + 
				InImgP[k+1][i+1][j]);
		    }

		    if (sum > BACKGROUND) {
                        /*--------------------------------------------------*/
			/*First-order of 3D neighbor                        */
                        /*--------------------------------------------------*/
			if ( InImgP[k][i][j-1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i][j-1];
			else if ( InImgP[k][i-1][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i-1][j];
			else if ( InImgP[k][i][j+1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i][j+1];
			else if ( InImgP[k][i+1][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i+1][j];
			else if ( k>0 && InImgP[k-1][i][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k-1][i][j];
			else if ( k<L-1 && InImgP[k+1][i][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k+1][i][j];
                        /*--------------------------------------------------*/
			/*Second-order of 3D neighbor, mid slice            */
                        /*--------------------------------------------------*/
			else if ( InImgP[k][i-1][j-1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i-1][j-1];
			else if ( InImgP[k][i-1][j+1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i-1][j+1];
			else if ( InImgP[k][i+1][j+1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i+1][j+1];
			else if ( InImgP[k][i+1][j-1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k][i+1][j-1];
                        /*--------------------------------------------------*/
			/*Second-order of 3D neighbor, upper slice          */
                        /*--------------------------------------------------*/
			else if ( k>0 && InImgP[k-1][i][j-1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k-1][i][j-1];
			else if ( k>0 && InImgP[k-1][i-1][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k-1][i-1][j];
			else if ( k>0 && InImgP[k-1][i][j+1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k-1][i][j+1];
			else if ( k>0 && InImgP[k-1][i+1][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k-1][i+1][j];
                        /*--------------------------------------------------*/
			/*Second-order of 3D neighbor, lower slice          */
                        /*--------------------------------------------------*/
			else if (k<L-1 && InImgP[k+1][i][j-1] != BACKGROUND) 
			    OutImgP[k][i][j] = InImgP[k+1][i][j-1];
			else if (k<L-1 && InImgP[k+1][i-1][j] != BACKGROUND) 
			    OutImgP[k][i][j] = InImgP[k+1][i-1][j];
			else if (k<L-1 && InImgP[k+1][i][j+1] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k+1][i][j+1];
			else if (k<L-1 && InImgP[k+1][i+1][j] != BACKGROUND ) 
			    OutImgP[k][i][j] = InImgP[k+1][i+1][j];
		    }
		}
	    }
	}
    }
}


void
erode(InImgP, OutImgP, N, M)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/* erode image                                                              */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
float  **InImgP, **OutImgP;
int    N, M;                               /* Number of rows and columns    */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j;
    float         sum, five_obj;
    float         *Ptr_in, *Ptr_out;

    five_obj = 5.0*OBJECT;
    Ptr_in = InImgP[0];
    Ptr_out = OutImgP[0];
    for(i=0; i<N*M; Ptr_in++, Ptr_out++, i++)
	*Ptr_out = *Ptr_in;
/*--------------------------------------------------------------------------*/
/* erode the middle (N-2)X(M-2) OBJECT, if 4 or more neighbors are          */
/* BACKGROUND.                                                              */
/*--------------------------------------------------------------------------*/
    for(i=1; i<N-1; i++) 
	for (j=1; j<M-1; j++) 
	    if (InImgP[i][j] == OBJECT) {
		sum = InImgP[i-1][j-1] + InImgP[i-1][j] + InImgP[i-1][j+1] +
		      InImgP[i][j-1] + InImgP[i][j+1] +
		      InImgP[i+1][j-1] + InImgP[i+1][j] + InImgP[i+1][j+1];
		if (sum < five_obj) 
		    OutImgP[i][j] = BACKGROUND;
	    }
}


void
erode_8(InImgP, OutImgP, N, M)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*   erode image with fixed 8 neighbors.                                    */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
float  **InImgP, **OutImgP;
int    N, M;                               /* Number of rows and columns    */

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j;
    float         sum, eight_obj;
    float         *Ptr_in, *Ptr_out;

    eight_obj = 8.0*OBJECT;
    Ptr_in = InImgP[0];
    Ptr_out = OutImgP[0];
    for(i=0; i<N*M; Ptr_in++, Ptr_out++, i++)
	*Ptr_out = *Ptr_in;
/*--------------------------------------------------------------------------*/
/* erode the middle (N-2)X(M-2) OBJECT, if one or more neighbors are        */
/* BACKGROUND.                                                              */
/*--------------------------------------------------------------------------*/
    for(i=1; i<N-1; i++) 
	for (j=1; j<M-1; j++) 
	    if (InImgP[i][j] == OBJECT) {
		sum = InImgP[i-1][j-1] + InImgP[i-1][j] + InImgP[i-1][j+1] +
		      InImgP[i][j-1] + InImgP[i][j+1] +
		      InImgP[i+1][j-1] + InImgP[i+1][j] + InImgP[i+1][j+1];
		if (sum < eight_obj) 
		    OutImgP[i][j] = BACKGROUND;
	    }
}


void
erode3D(InImgP, OutImgP, N, M, L)
/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*   erode image with a fixed square root of 2 raduis neighbors.            */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/
int           N, M, L;                     /* Number of rows and columns    */
unsigned char ***InImgP, ***OutImgP;

{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int           i, j, k;
    int           sum, Obj_18=4590, Obj_8=2040, Obj_13=3315;
    unsigned char *Ptr_in, *Ptr_out;

    for ( k=0; k<L; k++ ) { 
	Ptr_in = InImgP[k][0];
	Ptr_out = OutImgP[k][0];
	for(i=0; i<N*M; Ptr_in++, Ptr_out++, i++) {
	    *Ptr_out = *Ptr_in;
	}
    }
/*--------------------------------------------------------------------------*/
/* erode the middle (N-2)X(M-2) OBJECT, if 4 or more neighbors are          */
/* BACKGROUND.                                                              */
/*--------------------------------------------------------------------------*/
    for ( k=0; k<L; k++ ) { 
	for( i=1; i<N-1; i++ ) {
	    for ( j=1; j<M-1; j++ ) {
		if ( InImgP[k][i][j] == OBJECT ) {
		    sum = InImgP[k][i-1][j-1] + InImgP[k][i-1][j] + 
			  InImgP[k][i-1][j+1] + InImgP[k][i][j-1] + 
			  InImgP[k][i][j+1] + InImgP[k][i+1][j-1] + 
			  InImgP[k][i+1][j] + InImgP[k][i+1][j+1];
		    if ( k>0 ) {
			sum += (InImgP[k-1][i-1][j] + InImgP[k-1][i][j-1] + 
				InImgP[k-1][i][j] + InImgP[k-1][i][j+1] + 
				InImgP[k-1][i+1][j]);
		    }
		    if ( k<L-1 ) {
			sum += (InImgP[k+1][i-1][j] + InImgP[k+1][i][j-1] + 
				InImgP[k+1][i][j] + InImgP[k+1][i][j+1] + 
				InImgP[k+1][i+1][j]);
		    }
		    if ( L == 1 && sum < Obj_8) 
			OutImgP[k][i][j] = BACKGROUND;
		    else if ( L>1 && k>0 && k<L-1 && sum < Obj_18) 
			OutImgP[k][i][j] = BACKGROUND;
		    else if ( L>1 && (k==L-1 || k==0) && sum < Obj_13) 
			    OutImgP[k][i][j] = BACKGROUND;
		}
	    }
	}
    }
}
 


/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */ 
/*    Rotate direction vector 45 degree clockwise to find an adjacent       */
/* neighbor with value OBJECT                                               */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/

static struct coordinate2
rotate_direction (
   struct coordinate2 cur_dir)
{ 
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int                i, j;
    struct coordinate2 rotate_dir;
	
    i = cur_dir.i;
    j = cur_dir.j;

    if ((i==-1) && (j==0)) {
	i=-1;
	j=1;
    }
    else if ((i==-1) && (j==1)) {
	i=0;
	j=1;
    }
    else if ((i==0) && (j==1)) {
	i=1;
	j=1;
    }
    else if ((i==1) && (j==1)) {
	i=1;
	j=0;
    }
    else if ((i==1) && (j==0)) {
	i=1;
	j=-1;
    }
    else if ((i==1) && (j==-1)) {
	i=0;
	j=-1;
    }
    else if ((i==0) && (j==-1)) {
	i=-1;
	j=-1;
    }
    else if ((i==-1) && (j==-1)) {
	i=-1;
	j=0;
    }
    rotate_dir.i = i;
    rotate_dir.j = j;
    return rotate_dir;
}



/*--------------------------------------------------------------------------*/
/* Purpose:                                                                 */
/*   find next direction vector by rotating the current direction vector 90 */
/* degree counter clockwise to find the next moving direction               */
/*--------------------------------------------------------------------------*/
/*Input and output variable definitions                                     */
/*--------------------------------------------------------------------------*/

static struct coordinate2
next_direction (
   struct coordinate2 cur_dir)
{
/*--------------------------------------------------------------------------*/
/*Local variable definitions                                                */
/*--------------------------------------------------------------------------*/
    int                i, j;
    struct coordinate2 next_dir;
	
    i = cur_dir.i;
    j = cur_dir.j;
    if ((i==-1) && (j==0)) {
	i=0;
	j=-1; 
    }
    else if ((i==-1) && (j==1)) {
	i=-1;
	j=-1;
    }
    else if ((i==0) && (j==1)) {
	i=-1;
	j=0;
    }
    else if ((i==1) && (j==1)) {
	i=-1;
	j=1;
    }
    else if ((i==1) && (j==0)) {
	i=0;
	j=1;
    }
    else if ((i==1) && (j==-1)) {
	i=1;
	j=1;
    }
    else if ((i==0) && (j==-1)) {
	i=1;
	j=0;
    }
    else if ((i==-1) && (j==-1)) {
	i=1;
	j=-1;
    }
    next_dir.i = i;
    next_dir.j = j;
    return next_dir;
}



/*--------------------------------------------------------------------------
search_contour - find an object contour and return an interior point

This function searches a label image for an initial boundary point and
traces out the boundary, relabelling OBJECT points as BOUNDARY points
as it goes.  It returns an interior OBJECT point (which can be used as
a seed for region growing) and sets ContourSize to the number of
pixels in the boundary.  The initial boundary point is an OBJECT point
with another OBJECT point N or E of it and no OBJECT point to the W,
SW, S, or SE; if more than one point qualifies, the most southwesterly
one is taken.  If no interior point can be found, its coordinates will
be set to UNDEFINED.  Various other error conditions cause an error
message and may also abort the program.
--------------------------------------------------------------------------*/

struct coordinate2
search_contour (
   float   **LabelP,       /* Label image to be boundary traced */
   int     N, int M,       /* Dimensions of the image */
   int     *ContourSize)   /* (out) Number of pixels in boundary */
{
    int          i, j, max_size, rotate_num, InitPix = FALSE;
    int          end_of_search = FALSE, end_loop;
    struct coordinate2    cur_pos, cur_dir, init_pos, interior_pos;
    struct coordinate2    inter_dir;

    /* Search the entire image to find an initial boundary point, which 
       is an OBJECT point with another OBJECT point N or E of it and no 
       OBJECT point to the W, SW, S, or SE.   FIXME:  This is one of the
       admittedly rare places where a goto is the right solution. */
    max_size = N*M;
    *ContourSize = 0;
    init_pos.i = init_pos.j = 0;
    for ( i=1; i<N-1 && InitPix == FALSE; i++) {
	for ( j=1; j<M-1 && InitPix == FALSE; j++) {
	    if ( LabelP[i][j] == OBJECT &&
		 (LabelP[i][j+1] == OBJECT || LabelP[i+1][j] == OBJECT) ) { 
		if ( init_pos.i == 0 && init_pos.j == 0 ) {
		    InitPix = TRUE;
		    init_pos.i = i;
		    init_pos.j = j;
		}
	    }
	}
    }
    if ( InitPix == FALSE ) 
	fatal ("search_contour: no OBJECT pixel in the image.");

    /* Trace the contour starting from the initial boundary point, and 
       search for an interior point for region growing along the way. */
    cur_pos.i = init_pos.i;
    cur_pos.j = init_pos.j;
    LabelP[cur_pos.i][cur_pos.j] = BOUNDARY;
    (*ContourSize) ++;
    cur_dir.i = -1;
    cur_dir.j = 1;

    interior_pos.i = UNDEFINED;
    interior_pos.j = UNDEFINED;
    while (!end_of_search) {
	end_loop = FALSE;
	rotate_num = 0;
	while (!end_loop) {
	    rotate_num++;
	    if (rotate_num > 8) {
		fatal ("search_contour: isolated object pixel.");
	    }
	    if ( (cur_pos.i+cur_dir.i)>=0 && (cur_pos.i+cur_dir.i)<N && 
		 (cur_pos.j+cur_dir.j)>=0 && (cur_pos.j+cur_dir.j)<M &&
		 LabelP[cur_pos.i+cur_dir.i][cur_pos.j+cur_dir.j] 
		 != BACKGROUND ) {
		end_loop = TRUE;
	    }
	    else {
		cur_dir = rotate_direction(cur_dir);
	    }
	}
	
	if ( interior_pos.i == UNDEFINED && interior_pos.j == UNDEFINED) {
	    inter_dir = rotate_direction ( cur_dir );
	    if ( LabelP[cur_pos.i+inter_dir.i][cur_pos.j+inter_dir.j] 
		 == OBJECT ) {
		interior_pos.i = cur_pos.i + inter_dir.i;
		interior_pos.j = cur_pos.j + inter_dir.j;
	    }
	}

	cur_pos.i = cur_pos.i + cur_dir.i;
	cur_pos.j = cur_pos.j + cur_dir.j;
	if (cur_pos.i == init_pos.i && cur_pos.j == init_pos.j)
	    end_of_search = TRUE;
	else {
	    if (LabelP[cur_pos.i][cur_pos.j] == BOUNDARY) {
		printf("search_contour: double boundary point.\n");
/*		fatal("search_contour: double boundary point."); */
	    }
	    if ( cur_pos.i == interior_pos.i && 
		 cur_pos.j == interior_pos.j ) {
		interior_pos.i = UNDEFINED;
		interior_pos.j = UNDEFINED;
	    }
	    LabelP[cur_pos.i][cur_pos.j] = BOUNDARY;
	    cur_dir = next_direction (cur_dir);
	    (*ContourSize) ++;
	    if ( *ContourSize > max_size) {
		fatal ("search_contour: cannot find contour.");
	    }
	}
    }
/*    printf("Number of boundary pixels = %d \n", *ContourSize); */
    return interior_pos;
}



/*--------------------------------------------------------------------------*/
/* Structure for a stack of pixels neighboring of a given region.           */
/*--------------------------------------------------------------------------*/
struct NeighborPix {
   int                i, j;         /* Pixel coordinate;                    */
   struct NeighborPix *Next;        /* Pointer to next element of N_Stack.  */
};

/* -------------------------------------------------------------------------*/
/* Static variables declaration                                             */
/* -------------------------------------------------------------------------*/
static struct NeighborPix 
                 *N_Stack;           /* Pointer to the top of N_Stack.      */
   
static int       NS_Size=0;          /* Number of elements of N_Stack.      */
static int       sizeof_NP;          /* equal to sizeof(struct NeighborPix).*/
/*--------------------------------------------------------------------------*/
/*--------------------------------------------------------------------------*/


int
PutN_Stack (i, j)
/*--------------------------------------------------------------------------*/
/* Purpose :                                                                */
/*     It put an new element at the top of N_Stack. N_Stack may have        */
/* repeated entries.                                                        */
/*--------------------------------------------------------------------------*/
/* Definition of Input/Output Variables:                                    */
/*--------------------------------------------------------------------------*/
int                i, j;              /* Coordinate of to be added pixel.   */
{
/*--------------------------------------------------------------------------*/
/*  Definition of static/local variables :                                  */
/*------- ------------------------------------------------------------------*/
    struct NeighborPix *New_NP;
/*--------------------------------------------------------------------------*/
/*  Put the new element at the top of N_Stack.                              */
/*--------------------------------------------------------------------------*/
    if ((New_NP=(struct NeighborPix *) malloc (sizeof_NP)) != NULL) {
	New_NP->i = i;
	New_NP->j = j;
	New_NP->Next = N_Stack;
	N_Stack = New_NP;
	NS_Size++;
    }
    else
	fatal ("InsN_Stack: not enough memory.");
    return(0);
}


int
PopN_Stack (i, j)
/*------ -------------------------------------------------------------------*/
/* Purpose :                                                                */
/*     It pops the top element from N_Stack.                                */
/*--------------------------------------------------------------------------*/
/* Definition of Input/Output Variables:                                    */
/*--------------------------------------------------------------------------*/
int                *i, *j;      /* Ptrs to coordinate of N_Stack top. */
{
/*------- ------------------------------------------------------------------*/
/*  Definition of static/local variables :                                  */
/*--------------------------------------------------------------------------*/
    struct NeighborPix     *NP_t;
/*--------------------------------------------------------------------------*/
/*  Pop the top element of N_Stack.                                         */
/*--------------------------------------------------------------------------*/
    if (NS_Size == 0 || N_Stack == NULL)
	fatal ("PopN_Stack: empty stack of NeighborPix.");
    else {
	NP_t = N_Stack;
	N_Stack = NP_t->Next;
	NS_Size--;
	*i = NP_t->i;
	*j = NP_t->j;
	free (NP_t);
    }
    return(0);
}


int
UpdateN_Stack (LabelP, RowNum, ColNum, i, j)
/*--------------------------------------------------------------------------*/
/* Purpose :                                                                */
/*     It puts all neighbor pixels of (i,j) which are marked as             */
/* BACKGROUND to N_Stack.                                                   */
/*--------------------------------------------------------------------------*/
/* Definition of Input/Output Variables:                                    */
/*--------------------------------------------------------------------------*/
float              **LabelP;          /* Image of pixel status.             */ 
int                RowNum,            /* Number of rows of the image.       */
                   ColNum;            /* Number of columns of the image.    */
int                i, j;              /* Coordinate of to be added pixel.   */
{
/*--------------------------------------------------------------------------*/
/*  Definition of static/local variables :                                  */
/*--------------------------------------------------------------------------*/

    if (i > 0 && LabelP[i-1][j] == BACKGROUND)
	PutN_Stack (i-1, j);
    if (i < RowNum-1 && LabelP[i+1][j] == BACKGROUND)
	PutN_Stack (i+1, j);
    if (j > 0 && LabelP[i][j-1] == BACKGROUND) 
	PutN_Stack (i, j-1);
    if (j < ColNum-1 && LabelP[i][j+1] == BACKGROUND)
	PutN_Stack (i, j+1);
    return(0);
}



/*--------------------------------------------------------------------------
FillContour - Fill in holes in an object

Given the position of a seed pixel of type BACKGROUND, this function
finds all the connected pixels of type BACKGROUND and relabels them as
type FeatureV.  If the seed pixel is not BACKGROUND, just exit
immediately.  The number of pixels relabelled is returned.  This
function is often used to fill in holes in an OBJECT extracted from an
image.

FIXME:  This function is recursive and can overflow its internal stack
(actually a linked list) if the connected region is too large.
----------------------------------------------------------------------------*/

/*--------------------------------------------------------------------------*/
/* An old and apparently bogus description of FillContour:                  */
/*     It fills the OBJECT of a region enclosed by a contour. A seed        */
/* pixel which is in the OBJECT of the BOUNDARY is provided.                */ 
/*--------------------------------------------------------------------------*/

static int
FillContour (
   float   **LabelP,     /* Label image.                     */ 
   int     RowNum,       /* Number of rows of the image.     */
   int     ColNum,       /* Number of columns of the image.  */
   int     Seed_i,       /* Coordinate of seed pixel         */
   int     Seed_j,
   int     FeatureV)     /* Value to use as new label        */
{
    int    i, j, RegSize=0;

    /* Initialize the stack by popping off all the elements or
       dying noisily if there are less than none!? FIXME, please! */
    sizeof_NP = sizeof(struct NeighborPix);
    if (NS_Size < 0)
	fatal("FillContour: number of elements is negative.");
    while (NS_Size > 0)
	PopN_Stack (&i, &j);
    N_Stack = NULL;

    /* If the seed pixel isn't BACKGROUND, do nothing */
    if (LabelP[Seed_i][Seed_j] != BACKGROUND)
	return 0;

    /* Otherwise recursively relabel the seed pixel and all its
       connected neighbors of type BACKGROUND. */
    LabelP[Seed_i][Seed_j] = (float) FeatureV;
    RegSize ++;
    UpdateN_Stack(LabelP, RowNum, ColNum, Seed_i, Seed_j);
    while (NS_Size > 0) {
	PopN_Stack(&i, &j);
	if (LabelP[i][j] == BACKGROUND) {
	    LabelP[i][j] = (float) FeatureV;
	    RegSize ++;
	    UpdateN_Stack(LabelP, RowNum, ColNum, i, j);
	}
    }
    return RegSize;
}

