/*-------------------------------------------------------------------------
image_io.c - Functions for image i/o and memory management

Programmer       : Michelle Yan                                   
Revisions        : Paul Hughett                                   

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This source file contains some basic functions for allocating memory
and handling image I/O.  The functions callock, mallock, and reallock
previously included in this file have been moved to memm.c.  Most of
these functions are now deprecated and should not be used in new code.
                                                                   
   The included functions are                                      
   alloc_img      : allocates memory for floating NXM array        
   alloc_char     : allocates memory for unsigned char NXM array   
   alloc_3Dimg    : allocates memory for floating NXMXK array      
   alloc_3Dchar   : allocates memory for unsigned char NXMXK array 
   alloc_imgdouble: allocates memory for a double NXM array        
   open_file      : open a file for reading or writing.            
   read_matrix    : reads a NXM image from an opened file, whose   
                    type can be:                                   
                    unsigned char, short, unsigned short, int and  
                    float. It ought be used after alloc_img(), and 
                    open_file().                                   
   get_matrix     : opens a file, reads a NXM array and closes file
   write_matrix   : writes a NXM image to an opened file. It should
                    be used after open_file().                     
   put_matrix     : opens a file, writes a NXM array and close it  
   put_volume_uchar : open file, write unsigned char image         
   copy_matrix    : copies floating NxM image to another           
   copy_char      : copies unsigned char NxM image to another      
   err_msg        : display error message and exit  (OBSOLETE)               
   free_img       : frees memory occupied by a 2D array            
   free_3Dimg     : frees memory occupied by a 3D array            
   free_char      : frees memory used by a unsigned char 2D array  
   free_3Dchar    : frees memory used by a unsigned char 3D array  
   sizeof_ImgType : return byte size of a given data type       
---------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: image_io.c,v 1.13 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <limits.h>
#include <math.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "definition.h"

#include "dconv.h"
#include "errm.h"
#include "memm.h"
#include "image_io.h"



/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It allocated memory for a floating NXM array.                   */
/*-------------------------------------------------------------------*/

float **
alloc_img( int N,
	   int M )
{
  int			i;
  float		*Ptr;
  float		**Array_P;


  Array_P = (float **) mallock( N * sizeof(void *) );
  Ptr = (float *) callock( N * M, sizeof(float) );
  for (i = 0; i < N; i++) {
    *(Array_P + i) = Ptr;
    Ptr += M;
  }

  return Array_P;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It allocated memory for a double NXM array.                     */
/*-------------------------------------------------------------------*/

double **
alloc_imgdouble( int N,
		 int M )
{
  int			i;
  double		*Ptr;
  double		**Array_P;


  Array_P = (double **) mallock( N * sizeof(void *) );
  Ptr = (double *) callock( N * M, sizeof(double) );
  for (i = 0; i < N; i++) {
    *(Array_P + i) = Ptr;
    Ptr += M;
  }

  return Array_P;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It allocated memory for a unsigned char NXM array.              */
/*-------------------------------------------------------------------*/

unsigned char **
alloc_char( int N,
	    int M )
{
  int			i;
  unsigned char	*Ptr; 
  unsigned char	**Array_P;


  Array_P = (unsigned char **) mallock( N * sizeof(void *) );
  Ptr = (unsigned char *) callock( N * M, sizeof(unsigned char) );
  for (i = 0; i < N; i++) {
    *(Array_P + i) = Ptr;
    Ptr += M;
  }

  return Array_P;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It allocated memory for a floating NXMXK 3D array.              */
/*-------------------------------------------------------------------*/

float ***
alloc_3Dimg( int N,
	     int M,
	     int K )
{
  int			i;
  float		***Array_P;


  Array_P = (float ***) mallock( K * sizeof(void *) );
  for (i = 0; i < K; i++) {
    *(Array_P + i) = alloc_img( N, M );
  }

  return Array_P;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It allocated memory for a unsigned char NXMXK 3D array.         */
/*-------------------------------------------------------------------*/

unsigned char ***
alloc_3Dchar( int N,
	      int M,
	      int K )
{
  int			i;
  unsigned char	***Array_P;


  Array_P = (unsigned char ***) mallock( K * sizeof(void *) );
  for (i = 0; i < K; i++) {
    *(Array_P + i) = alloc_char( N, M );
  }

  return Array_P;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*   It opens a file for read "r" or write "w", and return a pointer */
/* of the file                                                       */
/*-------------------------------------------------------------------*/

FILE *
open_file( const char * FileName,
	   const char * OpenType )
{

  FILE           *retval;

  /*  FIXME: if input comes from file (batch processing), then if
      image file can not be found, we are stuck in endless loop
      waiting for input.  DRW  19980202  */

  retval = fopen( FileName, OpenType );
  if (retval) return retval;

  fprintf( stderr, "open_file() ERROR: can not open \"%s\" to ",
	   FileName );
  if (*OpenType == 'r') fprintf( stderr, "read.\n" );
  if (*OpenType == 'w') fprintf( stderr, "write.\n" );
  if (*OpenType == 'a') fprintf( stderr, "update.\n" );
  return retval;
}


/*------------------------------------------------------------------
sizeof_ImgType - Report datum length of an external representation

This function returns the length in bytes of one datum in a given 
external representation, or zero if the given type is invalid.
------------------------------------------------------------------*/

int
sizeof_ImgType(
   int ImgType)
{
  switch (ImgType) {
     case UCHAR:  return 1;
     case UINT2:  return 2;
     case INT2:   return 2;
     case INT4:   return 4;
     case FLOAT4: return 4;
     default:     return 0;
     }
}


/*-------------------------------------------------------------------
read_matrix - Read matrix from a binary file

This function reads N*M binary data from a file, converts them to
float, and stores it in a given buffer.  The possible binary types 
(for the external representation used in the file) are UCHAR, UINT2, 
INT2, INT4, and (on Suns only) FLOAT4; they are defined in definition.h.  
The data stored in the buffer is always float.

FIXME:  This routine should have better error checking for the i/o.

FIXME:  The float format on the Suns appears to be IEEE 754; if this
is so, then it should be reasonably portable (with only endian changes).
-------------------------------------------------------------------*/

void
read_matrix( FILE * FileP,	/* Pointer to an open file */
	     float ** ImageP,   /* Buffer for the data read */
	     int N,             /* Dimensions of matrix */
	     int M,
	     int ImgType )      /* Image type */
{
  float	*Ptr_float = *ImageP;
  unsigned char *tmp;
  long count, length;

  /* Compute number of elements and length of buffer */
  count = M*N;
  length = count*sizeof_ImgType(ImgType);
  if (length == 0) {
     fatal ("read_matrix: unknown image type.");
     return;
  }

  /* Allocate temporary buffer and read the data as unsigned chars*/
  tmp = mallock(length);
  fread(tmp, 1, length, FileP);

  /* Now convert the data according to its type */
  switch (ImgType) {

      case UCHAR:    /* unsigned character */
        uchar_to_float(count, Ptr_float, tmp);
	break;

      case UINT2:    /* unsigned 2-byte integer big endian */ 
        uint2_to_float(count, Ptr_float, tmp);
	break;

      case INT2:     /* signed 2-byte integer big endian */
        int2_to_float(count, Ptr_float, tmp);
        break;

      case INT4:     /* Two's comp 4-byte integer big endian */
        int4_to_float(count, Ptr_float, tmp);
	break;

#ifdef sparc         /* Implemented only on the Suns */
      case FLOAT4:   /* 4-byte float, Sun format, big endian */
	{ long i;
          for (i = 0; i < count; i++)
             Ptr_float[i] = tmp[i];
	  break;
	}
#endif

  } /* switch( ImgType ) */

  free(tmp);
  return;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It opens a file, and reads it into a NXM array, then close the */
/* file.                                                             */
/*-------------------------------------------------------------------*/

void
get_matrix( char * FileName,
	    float ** ImageP,
	    int N,
	    int M,
	    int ImgType )
{

    FILE       *FileP;

    
    FileP = open_file(FileName, "r");
    read_matrix (FileP, ImageP, N, M, ImgType);
    fclose(FileP);
}


/*-------------------------------------------------------------------
write_matrix - Write a matrix to a binary file

This function reads N*M binary data from a float array, converts them
to the appropriate external representation, and writes them to a binary
file.  The possible external representations are UCHAR, UINT2, INT2,
INT4, and (for Suns only) FLOAT4; they are defined in definition.h.

FIXME:  This routine should do better error checking on the i/o.
---------------------------------------------------------------------*/

void
write_matrix( 
   FILE * FileP,     /* Pointer to an opened file */
   float ** ImageP,  /* Buffer for the data to be written */
   int N,            /* Dimensions of the array */
   int M,
   int ImgType )     /* Image type */
{
  float	*Ptr_float = *ImageP;
  unsigned char *tmp;
  long count, length;

  /* Compute number of elements and length of buffer; allocate buffer */
  count = M*N;
  length = count*sizeof_ImgType(ImgType);
  if (length == 0) {
     fatal ("write_matrix: unknown image type.");
     return;
  }
  tmp = mallock(length);

  /* Now convert the data according to its type */
  switch (ImgType) {

      case UCHAR:    /* unsigned character */
        float_to_uchar(count, tmp, Ptr_float);
	break;

      case UINT2:    /* unsigned 2-byte integer big endian */ 
        float_to_uint2(count, tmp, Ptr_float);
	break;

      case INT2:     /* signed 2-byte integer big endian */
        float_to_int2(count,tmp, Ptr_float);
        break;

      case INT4:     /* Two's comp 4-byte integer big endian */
        float_to_int4(count, tmp, Ptr_float);
	break;

#ifdef sparc         /* Implemented only on the Suns */
      case FLOAT4:   /* 4-byte float, Sun format, big endian */
	{ long i;
          for (i = 0; i < count; i++)
             tmp[i] = Ptr_float[i];
	  break;
	}
#endif

  } /* switch( ImgType ) */

  /* Finally write the data, free the buffer, and return */
  fwrite(tmp, 1, length, FileP);
  free(tmp);
  return;

}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It opens a file, and writes it into a NXM array, then close    */
/* the file.                                                         */
/*-------------------------------------------------------------------*/

void
put_matrix( char * FileName,
	    float ** ImageP,
	    int N,
	    int M,
	    int ImgType )
{

    FILE       *FileP;

    
    FileP = open_file(FileName, "w");
    write_matrix (FileP, ImageP, N, M, ImgType);
    fclose(FileP);
}


/*-------------------------------------------------------------------
put_volume_uchar - Write an unsigned char volume image to a file

This function opens a file, writes an unsigned character image to
it in raw format, and closes the file.
-------------------------------------------------------------------*/

void
put_volume_uchar(
    char * FileName,            /* Name of file to write to */
    unsigned char ***ImageP,    /* Volume image to write out */
    int N,                      /* Dimensions of the volume image */
    int M,
    int L)
{
    FILE       *FileP;
    int k;
    
    FileP = open_file(FileName, "w");
    for (k = 0; k < L; k++)
        fwrite(&ImageP[k][0][0], 1, (long)N*M, FileP);
    fclose(FileP);
    return;
}



/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It copies a NXM array to another.                              */
/*-------------------------------------------------------------------*/

void
copy_matrix( float ** InImgP,
	     float ** OutImgP,
	     int N,
	     int M )
{

    int            i;
    float          *ptr1, *ptr2;


    for (i=0, ptr1=InImgP[0], ptr2=OutImgP[0]; i<N*M; 
	 i++, ptr1++, ptr2++)
	*ptr2 = *ptr1;
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It copies an unsigned char NXM array to another.               */
/*-------------------------------------------------------------------*/

void
copy_char( unsigned char ** Label_In, /* Pointer to input pointer array */
	   unsigned char ** Label_Out, /* Pointer to output pointer array*/
	   int N,
	   int M )
{

  int            i;
  unsigned char  *ptr_i, *ptr_o;

  for (i=0, ptr_i=Label_In[0], ptr_o=Label_Out[0]; i<N*M; 
       i++, ptr_i++, ptr_o++)
    *ptr_o = *ptr_i;
}


/*---------------------------------------------------------------------
err_msg - Display error message and abort program

An old function for reporting a fatal error; new code should use fatal()
or panic().
----------------------------------------------------------------------*/

#if 0         /* FIXME:  Have I changed all the users to fatal? */
void
err_msg( const char * message )
{
  fprintf( stderr, "\nERROR:\n        %s\n", message );
  exit(-1);
}
#endif


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It frees occupied memory by a 2D array                         */
/*-------------------------------------------------------------------*/

void
free_img( float ** ImageP )
{
  if (ImageP) {
    free( *ImageP );
    free( ImageP );
  }
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It frees occupied memory by a 2D double array                  */
/*-------------------------------------------------------------------*/

void
free_imgdouble( double ** ImageP )
{
  if (ImageP) {
    free( *ImageP );
    free( ImageP );
  }
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It frees occupied memory by a unsigned char 2D array           */
/*-------------------------------------------------------------------*/

void
free_char( unsigned char ** ImageP )
{
  if (ImageP) {
    free( *ImageP );
    free( ImageP );
  }
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It frees occupied memory by a floating 3D array                */
/*-------------------------------------------------------------------*/

void
free_3Dimg( float *** ImageP,
	    int K )		/* Number of slices */
{
  int       i;

  if (ImageP) {
    for (i = 0; i < K; i++) {
      free_img( *(ImageP + i) );
    }
    free( ImageP );
  }
}


/*-------------------------------------------------------------------*/
/* purpose:                                                          */
/*    It frees occupied memory by a unsigned char 3D array           */
/*-------------------------------------------------------------------*/

void
free_3Dchar( unsigned char *** ImageP,
	     int K )		/* Number of slices */
{
  int       i;

  if (ImageP) {
    for (i = 0; i < K; i++) {
      free_char( *(ImageP + i) );
    }
    free( ImageP );
  }
}
