/***************************************************************************
kmeans.c - segment a volume image using the K-means clustering algorithm

Programmer: Paul Hughett

Copyright (c) 1999 Trustees of the University of Pennsylvania
See the file COPYING for license terms.

This program segments a volume image using a plain (non-adaptive) K-means
clustering algorithm on the gray levels.  It is an adaptation of and eventual
replacement for Michelle Yan's Kmeans_3Dseg program, but is not guaranteed
to produce identical results.  See the man page for usage instructions.

FIXME:  This program already prints out the number of voxels in each class;
it shouldn't be hard to also print out the total volume for each class.

FIXME:  Read a verbosity parameter and do something reasonable with it.

****************************************************************************/

static const char rcsid[] = "$Id: kmeans.c,v 1.5 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <time.h>
#include <unistd.h>   /* For getopt and friends */

#include <sys/types.h>
#include <sys/stat.h>

#include "config.h"

#include "errm.h"
#include "Kmean_3Dseg.h"
#include "param.h"


int
main( int argc, char *argv[] )
{
    Param_list params;             /* Contents of param file */
    char *stage = "kmeans";        /* Default stage name for params */
    const char *image1, *image2;   /* Filenames for the two echo images */
    int	savemean = 0;              /* Save the means image? */
    int	echo_12;                   /* Number of echoes: 1 or 2 */
    int NoRegion;                  /* Number of regions to classify into */
    int	ImgType1, ImgType2;        /* Types of 1st and 2nd images */
    int nrows, ncols;              /* Number of rows and columns */
    int nslices;                   /* Number of slices */
    int First, Last;               /* First and last slice to process */
    float Res_XY, Res_Z;           /* Physical voxel dimensions [mm] */
    float Perturb;                 /* Value of perturbation */
    float Epsilon;                 /* Value of distortion threshold */
    float Gibbs;                   /* Value of Gibbs potential */
    float std1;                    /* Standard deviation of noise in image 1 */
    float std2;                    /* Standard deviation of noise in image 2 */
    int flag;                      /* Flag returned by segmentation engine */
    struct stat	filestat;
    const char *s;
    clock_t clock0, clock1;        /* To compute CPU run time */
    time_t  time0,  time1;         /* To compute elapsed time */
    double  cpu_time, elapsed;

    /* Record start times */
    clock0 = clock();
    time0  = time(NULL);

    /* Read the param file named on the command line */
    if (argc < 2)
	fatal("No param file specified");
    params = read_param_file(argv[1]);

    /* Extract the necessary parameter values */
    image1 = get_param(params, stage, "image1");
    image2 = get_param(params, stage, "image2");
    ImgType1 = (s = get_param(params, stage, "type1"   ) ) ? atoi(s) :   2;
    ImgType2 = (s = get_param(params, stage, "type2"   ) ) ? atoi(s) :   2;
    nrows    = (s = get_param(params, stage, "nrows"   ) ) ? atoi(s) : 256;
    ncols    = (s = get_param(params, stage, "ncols"   ) ) ? atoi(s) : 256;
    nslices  = (s = get_param(params, stage, "nslices" ) ) ? atoi(s) :   0;
    First    = (s = get_param(params, stage, "first"   ) ) ? atoi(s) :   1;
    Last     = (s = get_param(params, stage, "last"    ) ) ? atoi(s) :  -1;
    echo_12  = (s = get_param(params, stage, "nechoes" ) ) ? atoi(s) :   1;
    NoRegion = (s = get_param(params, stage, "nregions") ) ? atoi(s) :   5;
    Perturb  = (s = get_param(params, stage, "perturb" ) ) ? atof(s) : 0.01;
    std1     = (s = get_param(params, stage, "std1"    ) ) ? atof(s) : 10.0;
    std2     = (s = get_param(params, stage, "std2"    ) ) ? atof(s) : 10.0;
    Epsilon  = (s = get_param(params, stage, "epsilon" ) ) ? atof(s) : 0.02;
    Res_XY   = (s = get_param(params, stage, "xspace"  ) ) ? atof(s) : 0.9375;
    Res_XY   = (s = get_param(params, stage, "yspace"  ) ) ? atof(s) : Res_XY;
    Res_XY   = (s = get_param(params, stage, "xyspace" ) ) ? atof(s) : Res_XY;
    Res_Z    = (s = get_param(params, stage, "zspace"  ) ) ? atof(s) : 1.0;
    Gibbs    = (s = get_param(params, stage, "gibbs"   ) ) ? atof(s) : 0.0;


    /* FIXME:  Provide reasonable defaults for input and output file names */

    /* FIXME: Add consistency checks for the parameters */

    /* FIXME: The reason for the following code is rather obscure, 
       so just comment out for now rather than delete. -PWH 19990121 */
#if 0
    /* silence stdout during normal operation */
    if (!debugflag)
	freopen( "/dev/null", "w", stdout );
#endif

    /* Determine nslices from size of image file */
    if (nslices <= 0) {
	if (stat( image1, &filestat ) < 0) 
	    fatal ("Unable to read size of image1 %s", image1);
	nslices = filestat.st_size / (nrows * ncols * abs(ImgType1));
    }

    /* Last defaults to the number of slices */
    if (Last == -1)   
	Last = nslices;

    /* Print out the actual parameter values for inspection */
    printf("Number of echoes, number of regions = %d, %d\n", 
	   echo_12, NoRegion);
    printf("Perturbation, distortion & potential = %0.3e, %0.3e, %0.3e\n",
	   Perturb, Epsilon, Gibbs);
    printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	   image1, ImgType1, std1);
    std1 = 0.5 / ( std1*std1);   /* FIXME:  Hunh?  What's this? */
    if ( echo_12 == 2 ) {
	printf("Image name, type & standard deviation: %s (%d), %0.3f\n",
	       image2, ImgType2, std2);
	std2 = 0.5 / (std2 * std2);
    }
    printf("Image dimension = %d X %d; %d -> %d  (%d)\n", 
	   nrows, ncols, First, Last, nslices);
    printf("XY- & Z-resolution = %0.3e %0.3e\n", Res_XY, Res_Z);

    /* Call the segmentation engine to do all the real work */
    flag = kmean_3dseg( echo_12, NoRegion,
			image1, ImgType1, std1,
			image2, ImgType2, std2,
			nrows, ncols, First, Last, nslices,
			Res_XY, Res_Z,
			Perturb, Epsilon, Gibbs,
			savemean );

    /* Report the CPU and elapsed time used */
    clock1 = clock();
    time1  = time(NULL);
    cpu_time = (clock1 - clock0) / (double)CLOCKS_PER_SEC;
    elapsed  = (double)(time1 - time0);
    fprintf(stderr, "CPU run time was %8.3f seconds.\n", cpu_time);
    fprintf(stderr, "Elapsed time was %8.3f seconds.\n", elapsed);

    return flag;
}
