/*--------------------------------------------------------------------------
region_grow.c                                                            

Programmer       : Michelle Yan                                          
Revisions        : Paul Hughett                                          

Copyright (c) 1998 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


This source code contains some basic functions for 3D region growing.
  
   int PutN_Stack3D()     : put a new neighbor pixel at the top of        
                            N_Stack.                                      
   int PopN_Stack3D()     : get the top pixel from N_Stack.               
   int UpdateN_Stack3D()  : add neighbor pixels of a given pixel to       
                            N_Stack.                                      
   int FillRegion()       : fill all connected current label pixels by a  
                            new label.                                    
   int ConnectComponent() : label connected components of a binary image  
                            using FillRegion().                           
   void TakeLabel()       : compute binary image by assigning input value 
                            as OBJECT or BACKGROUND.                      
   void And2Mask()        : compute binary image by taking AND operation  
                            on 2 Masks.                                   
   void ExtractVOI()      : find the largest 3D connected OBJECT or fill  
                            holes not connected to BACGROUND, used as an  
                            initial volume of interest (VOI).             
   void InnerMask()       : find mask for inner VOI based on the result   
                            from ExtractVOI.                              
   void CleanUpMask()     : clean up and refine VOI based on the result   
                            from InnerMask.                               
   int MaskPropagate()    : propagate reference mask to test one using    
                            overlap portion of the two.                   
   void DistMap()         : compute 3D distance map using ChamferDist2D.  
   int  T_DistMap()       : select optimal threshold for 3D distance map. 
   void ChamferDist2D()   : compute Chamfer distance for a 2D OBJECT mask.
   int RegionGrow3D()     : find a connected volume based on a seed, and  
                            assign a new label.                           
   NOTE: It has problem for large OBJECT, due to recursive use of it.     
--------------------------------------------------------------------------*/

static const char rcsid[] = "$Id: region_grow.c,v 1.14 2000/06/05 17:17:59 hughett Exp $ Release $Name: segm-0-90 $";

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <math.h>

#include "common.h"
#include "conn_component.h"
#include "definition.h"
#include "errm.h"
#include "find_mask.h"
#include "image_inc.h"
#include "image_io.h"
#include "list.h"
#include "memm.h"
#include "region_grow.h"


/*-------------------------------------------------------------------------

                        Neighborhood stack

Connected components of an image are found by recursively finding
neighbors of an initial seed pixel; the neighborhood stack contains
a randomly ordered list of pixels whose neighbors have not yet been
found.  Each entry in the stack contains the coordinates of a pixel
and a pointer to the next entry.

There is only a single neighborhood stack, so you shouldn't try to
use this code in a multi-threaded or recursive environment.

FIXME: The functions PutN_Stack3D, PopN_Stack3D, and UpdateN_Stack3D
are apparently older code that is now never used in Michelle's code.
If a user is eventually discovered, these functions should probably be
rewritten for efficiency (don't call mallock for every entry, don't
maintain both size count and list itself, don't use a variable for
sizeof an entry).

----------------------------------------------------------------------------*/


/*--------------------------------------------------------------------------*/
/* Structure for a stack of pixels neighboring a given region.              */
/*--------------------------------------------------------------------------*/
typedef struct NeighborPix 
{
   short              i, j, k;      /* Pixel coordinate.                    */
   struct NeighborPix *Next;        /* Pointer to next element of N_Stack.  */
} NeighborPix;


/* -------------------------------------------------------------------------*/
/* Static variables declaration                                             */
/* -------------------------------------------------------------------------*/
static NeighborPix *N_Stack;         /* Pointer to the top of N_Stack.      */
static int         NS_Size = 0;      /* Number of elements of N_Stack.      */
static int         sizeof_NP;        /* equal to sizeof(NeighborPix).       */
/*--------------------------------------------------------------------------*/



/*--------------------------------------------------------------------------*/
/* PutN_Stack3D - Push element onto the neighborhood stack                  */
/*                                                                          */
/*     It put an new element at the top of N_Stack. N_Stack may have        */
/* repeated entries.                                                        */
/*--------------------------------------------------------------------------*/
int
PutN_Stack3D( int i,		/* Coordinate of pixel to be added.   */
	      int j,
	      int k )
{

    NeighborPix    *New_NP;


    New_NP = (NeighborPix *) mallock(sizeof_NP);
    New_NP->i = i;
    New_NP->j = j;
    New_NP->k = k;
    New_NP->Next = N_Stack;
    N_Stack = New_NP;
    NS_Size++;

    return 0;
}


/*--------------------------------------------------------------------------*/
/* PopN_Stack3D - Pop element off the neighborhood stack.                   */
/*                                                                          */
/*     It pops the top element from N_Stack.                                */
/*--------------------------------------------------------------------------*/

int
PopN_Stack3D( int * i,		/* coordinate of pixel from N_Stack top */
	      int * j,
	      int * k )
{

    NeighborPix    *NP_t;


    if (NS_Size == 0 || N_Stack == NULL)
	fatal ("PopN_Stack: empty stack of NeighborPix.");
    else {
	NP_t = N_Stack;
	N_Stack = NP_t->Next;
	NS_Size --;
	*i = NP_t->i;
	*j = NP_t->j;
	*k = NP_t->k;
	free (NP_t);
    }

    return 0;
}


/*--------------------------------------------------------------------------*/
/* UpdateN_Stack3D - Push all neighbors of a pixel onto neighborhod stack   */
/*                                                                          */
/*     It puts all neighbor pixels of (i,j,k) which are marked as CurLabel  */
/* to N_Stack.                                                              */
/*--------------------------------------------------------------------------*/

int
UpdateN_Stack3D( unsigned char *** LabImg, /* 3D label image. */ 
		 int N,		/* Number of rows, cols and slices.   */
		 int M,
		 int L,
		 int i,		/* Coordinate of pixel to be added.   */
		 int j,
		 int k,
		 int CurLabel )
{
    if ( i > 0 && LabImg[k][i-1][j] == CurLabel )
	PutN_Stack3D (i-1, j, k);
    if ( i < N-1 && LabImg[k][i+1][j] == CurLabel )
	PutN_Stack3D (i+1, j, k);
    if ( j > 0 && LabImg[k][i][j-1] == CurLabel ) 
	PutN_Stack3D (i, j-1, k);
    if ( j < M-1 && LabImg[k][i][j+1] == CurLabel )
	PutN_Stack3D (i, j+1, k);
    if ( k > 0 && LabImg[k-1][i][j] == CurLabel )
	PutN_Stack3D (i, j, k-1);
    if ( k < L-1 && LabImg[k+1][i][j] == CurLabel )
	PutN_Stack3D (i, j, k+1);

    return 0;
}


/*--------------------------------------------------------------------------*/
/* FillRegion - Relabel a connected component with a new label              */
/*                                                                          */
/*     It fills a connected volume that contains the point (Si, Sj, Sk) and */
/* is labeled as CurLabel by NewLabel.                                      */
/*--------------------------------------------------------------------------*/

int
FillRegion( 
   unsigned char *** LabImg,  /* Labelled image to process */
   int N,              /* Dimensions of the image */
   int M,
   int L,
   int Si,             /* Coordinates of a seed pixel */
   int Sj,
   int Sk,
   int CurLabel,       /* Label to be replaced */
   int NewLabel )      /* Label to replace with */
{

    unsigned char tmpCurLabel, tmpNewLabel;


    tmpCurLabel = (unsigned char) CurLabel;
    tmpNewLabel = (unsigned char) NewLabel;
    return CC_flood_uchar (LabImg, L, N, M, Sk, Si, Sj,
                           &tmpCurLabel, &tmpNewLabel);
}


/*------------------------------------------------------------------------
get_eqLabel - ???

FIXME:  Apparently never used.  Prototype provided to keep gcc happy.
--------------------------------------------------------------------------*/

int get_eqLabel( int label, int * eqLabelTable );

int
get_eqLabel( int label,
             int * eqLabelTable )
{
    if (eqLabelTable[label] == label)
        return label;
    else
        return get_eqLabel (eqLabelTable[label], eqLabelTable);
}



/*------------------------------------------------------------------------
seedcmp - compare size of two regions defined by CC_reg_seed_t structs

This routines compares the size of two regions defined by CC_reg_seed_t
structs and is compatible with the doubly linked list routines defined
list.c.  It is used within ConnectComponent to maintain a sorted list of
regions.  The usual meaning is reversed, so that regions are sorted in 
decreasing order of size.
--------------------------------------------------------------------------*/

static int
seedcmp( CC_reg_seed_t * seed1,
         CC_reg_seed_t * seed2 )
{
  /* Note reversal of standard meaning, to sort in decreasing order */
  if (seed1->regSize < seed2->regSize)
    return 1;
  else if (seed1->regSize > seed2->regSize)
    return -1;
  else
    return 0;
}




/*-------------------------------------------------------------------------
ConnectComponent - Find and label connected components of a binary image

This function finds the connected components of a binary image and
labels them from 1 to MaxRegNo (which must be <= 253).  The total
number of components found is returned.  Regions of pixels that are
labelled as Target_V but containing fewer than MinRegSize pixels are
discarded by setting their pixels to OBJECT - Target_V, which is
presumably supposed to be BACKGROUND.  Regions composed of pixels
labelled anything other than Target_V are ignored.

Binary image may be a misnomer; the possible values relevant to this
program seem to be Target_V (always = OBJECT?) and bgLabel (always =
BACKGROUND?).  In any case, the output labelled image can contain
any unsigned char values as region labels, but skips over OBJECT,
bgLabel ( = OBJECT - Target_V) and Target_V.

The meaning of the argument Target_V is not known for certain.  It is
set to OBJECT everywhere that Michelle uses this function.  Tentatively,
it is used to identify the class for which connected components are to
be found.  If so, BACKGROUND pixels are not organized into a connected
component.

It's not clear whether or not BACKGROUND regions are included or excluded
as connected components.

The total number of regions found, and the number and size of the largest
is printed on standard output.

FIXME:  It might be useful to also produce a list of the regions found,
including size and a seed pixel (perhaps the centroid?).  If nothing else, 
this might make further processing more efficient.
---------------------------------------------------------------------------*/

int
ConnectComponent(
   unsigned char *** Binary_In,  /* Binary image to analyze */
   unsigned char *** Label_Out,  /* (out) Labelled image generated */
   int N,               /* Dimensions of the image */
   int M,
   int L,
   int MinRegSize,      /* Minimum region size */
   int Target_V,        /* Always OBJECT ??? */
   int MaxRegNo,	/* Maximum number of regions allowed */
   int * MaxV_Label )   /* (out) Label for largest background region ??? */
{
    int k, i; 
    int noRegions;
    int maxV_Size=0;
    unsigned char *ptr_o;
    unsigned char curLabel; 
    unsigned char newLabel; 
    unsigned char bgLabel;
    stack_t       *seedStack;
    list_t        *seedList;
    CC_reg_seed_t *seed;

    /* Initialize working variables */
    if (MaxRegNo > 253)  MaxRegNo = 253;
    bgLabel   = OBJECT - Target_V;
    newLabel  = 1;
    noRegions = 0;

    /* Copy the input binary image to the output labelled image */
    for ( k=0; k<L; k++ ) {
        memcpy( Label_Out[k][0], Binary_In[k][0],
	    N * M /* * sizeof(unsigned char) */ );
    }
    

  /*------------------------------------------------------------------------*/
  /*remove components with label!=Target_V or size<MinRegSize from seedStack*/
  /*------------------------------------------------------------------------*/

    /* Generate a stack of connected components */
    seedStack = CC_label_uchar( (void ***) Label_Out, L, N, M );
    LIST_resetStack( seedStack );

    /* Prepare an empty list of regions, to be sorted by size */
    seedList = LIST_newList( (listelemcompfunc) seedcmp,
			     (listelemfreefunc) defaultFreeContent );


    /* Loop over the regions contained in the stack */
    while ((seed = LIST_popStack (seedStack))) {
        curLabel=Label_Out[seed->slice][seed->row][seed->col];

        /* If the region is labelled Target_V and above the minimum
           size, insert it in the list */
        if (Binary_In[seed->slice][seed->row][seed->col] == Target_V &&
	    seed->regSize > MinRegSize) {
          LIST_insertList (seedList, seed);
        }

        /* If the region is labelled Target_V and below the minimum
           size, erase it (relabel it as background) */
        else if (Binary_In[seed->slice][seed->row][seed->col] == Target_V && 
	         seed->regSize <= MinRegSize && curLabel != bgLabel) {
          CC_flood_uchar (Label_Out, L, N, M, seed->slice, seed->row, 
	    	      seed->col, &curLabel, &bgLabel); 
          free (seed);
        }

        /* Otherwise just ignore it */
        else {
          free (seed);
        }
    }

    /* The sorted list now contains the regions found, in decreasing order
       of size. Now stuff the first MaxRegNo regions back onto the stack
       so that the stack now contain the largest regions in order
       of increasing size.  */
    LIST_topList (seedList);
    while ((seed = LIST_retrieveList (seedList))) {
        if (seedStack->stackSize < MaxRegNo)
            LIST_pushStack (seedStack, seed);
        else {
            curLabel=Label_Out[seed->slice][seed->row][seed->col];
            if (curLabel != bgLabel)
	        CC_flood_uchar (Label_Out, L, N, M, seed->slice, seed->row, 
		      	seed->col, &curLabel, &bgLabel); 
            printf( "WARNING: component with size = %d is removed.\n",
	            seed->regSize );
            free (seed);
        }
    }
    LIST_freeList (seedList);


/*  ------------------------------------------------------------------------*/
/*  relabel remaining components with label==Target_V.                      */
/*  ------------------------------------------------------------------------*/

    /* Loop over the remaining regions (from smallest to largest?) */
    LIST_resetStack (seedStack);
    while ((seed = LIST_popStack (seedStack))) {

        /* Find the next available pixel label */
        while (newLabel == OBJECT || newLabel == bgLabel ||
               newLabel == Target_V)
            newLabel++;
	
#if 0
        /* Print diagnostic information */
        printf ("%d: seed=(%d %d %d)\t size=%d\t label=%d \n",
	        newLabel, seed->slice, seed->row, seed->col, seed->regSize,
	        Label_Out[seed->slice][seed->row][seed->col]);
#endif

        /* Check for possible errors in previous calculation */
        if  ( seed->regSize <= MinRegSize ||
	      Binary_In[seed->slice][seed->row][seed->col] != Target_V)
            fatal ("connectComponent: incorrect new seedStack.");

        /* Label the current region with the newest label.  If it's
           the last (and largest) region, do the relabelling by
           hand for efficiency, and remember the label and size. */
        curLabel=Label_Out[seed->slice][seed->row][seed->col];
        if (seedStack->stackSize > 0)  
            CC_flood_uchar (Label_Out, L, N, M, seed->slice, 
	    	          seed->row, seed->col, &curLabel, 
		          &newLabel);
        else {                 /* the last region is the largest one */
            *MaxV_Label = newLabel;
            maxV_Size = seed->regSize;
            for ( k=0; k<L; k++ ) {
	        ptr_o = Label_Out[k][0];
	        for ( i=0; i<N*M; i++, ptr_o++ ) {
	            if ( *ptr_o == Target_V ) 
	              *ptr_o = newLabel;
	            }
                }
	}

        /* Finish up this region and prepare for the next */
        free (seed);
        noRegions ++;
        newLabel ++;
        if (newLabel-1 > MaxRegNo)
            fatal ("ConnectComponent: Too many connected components");
    }  /* Loop over regions */

    /* Clean up, print results, and exit with number of regions found */
    LIST_freeStack (seedStack);
    printf("Total No. of Regions = %d, MaxLabel = %d (%d)\n", 
	   noRegions, *MaxV_Label, maxV_Size);
    return noRegions;
}


/*--------------------------------------------------------------------------
TakeLabel - Conditionally change labels on a 3D image

This function examines and modifies each voxel of a 3D unsigned char image.
If the voxel is equal to Value_I, it is changed to Value_O; otherwise it
is changed to OBJECT - Value_O.  Value_O should be set to either OBJECT
or BACKGROUND, since this operation doesn't make a lot of sense otherwise.

FIXME:  This function could more reasonably be replaced by one that sets
the voxel to Value_A or Value_B, rather than Value_O or OBJECT - Value_O.
--------------------------------------------------------------------------*/

void
TakeLabel( unsigned char *** Label_IO,  /* Image to examine and modify */
	   unsigned char Value_I,       /* Value to test for */
	   unsigned char Value_O,       /* Either OBJECT or BACKGROUND */
	   int N,                       /* Dimensions of the image */
	   int M,
	   int L )
{
  int			i;
  unsigned char *	ptr_io;
  int			k;


  for (k = L - 1; k >= 0; k--) {
    ptr_io = Label_IO[k][0];
    for (i = N * M; i > 0; i--) {
      *ptr_io = (*ptr_io == Value_I ? Value_O : (OBJECT - Value_O));
      ptr_io++;
    }
  }
}



/*--------------------------------------------------------------------------
And2Mask - Compute a binary mask using an AND operation

This function compares two 3D unsigned char images Label_IO and Label_Ref
pixel by pixel.  If Label_IO is equal to Value_IO and Label_Ref is equal 
to Value_Ref, then Label_IO is set to OBJECT, otherwise to BACKGROUND.
--------------------------------------------------------------------------*/

void
And2Mask(
   unsigned char *** Label_IO,    /* Image to be modified */
   unsigned char *** Label_Ref,   /* Image to control modifications */
   int Value_IO,                  /* Comparison value for Label_IO */
   int Value_Ref,                 /* Comparison value for Label_Ref */
   int N,                         /* Dimensions of the image */
   int M,
   int L )
{

    int           i, k;
    unsigned char *ptr_io, *ptr_ref;

    for ( k=0; k<L; k++ ) {
	ptr_io = Label_IO[k][0];
	ptr_ref = Label_Ref[k][0];
	for ( i=0; i<N*M; i++, ptr_io++, ptr_ref++ ) {
	    if ( *ptr_io == Value_IO && *ptr_ref == Value_Ref ) {
		*ptr_io = OBJECT;
	    }
	    else {
		*ptr_io = BACKGROUND;
	    }
	}
    }
}



/*--------------------------------------------------------------------------
ExtractVOI - Create mask for head or brain in a PET or MRI image

If OuterFirstYN is set to "n", this function computes a mask for the
brain volume within a PET or T2 MRI volume image.  In these images the
scalp and skull are faint and the algorithm is tuned to match.

If OuterFirstYN is set to "y", this function computes a mask for the
head volume within an MRI T1 or proton volume image.  In this images the
scalp and skull have brightness similar to the brain parenchyma; the 
algorithm is tuned to match and does not attempt to separate brain from
scalp.

The name OuterFirstYN is not very well chosen.  The idea is that for
PET or T2 images (n), you can do the whole masking operation in one
step, whereas for T1 or proton images you must extract the head first
and then strip off the scalp and skull in a separate step.

The binary image input is expected to consist of pixel values OBJECT and
BACKGROUND only.  The output mask image is constructed to have the
value OBJECT where the desired image is present and BACKGROUND elsewhere.
--------------------------------------------------------------------------*/

void
ExtractVOI(
   unsigned char *** Binary_In,   /* Binary image to examine */
   unsigned char *** Mask_Out,    /* Output mask for head or brain */
   int N,                         /* Dimensions of the image */
   int M,
   int L,
   char * OuterFirst_YN )         /* Indicator to remove outer layer first */
{

  int           i, k, NonObject;
  int           NoRegion = 0, MaxNoReg = 253, MinSize = 5, MaxV_Label;
  unsigned char ***tmpMask_1, *ptr_1, *ptr_2, ***ptr3_1, ***ptr3_2;


  tmpMask_1 = alloc_3Dchar( N, M, L );
  NonObject = MaxNoReg + 1;


  /*------------------------------------------------------*/
  /*            Case 1 (OuterFirstYN == "n"):             */
  /*  Find the brain volume in PET or MRI T2 image        */
  /*------------------------------------------------------*/

  if (*OuterFirst_YN == 'n') {

    /* When the slice thickness is much larger than the in-plane pixel
       size, use the 3D morphological operations with care. */
    erode3D (Binary_In, tmpMask_1, N, M, L);
    ptr3_1 = tmpMask_1;
    ptr3_2 = Mask_Out;
    for (k = 0; k < L; k++)
      erode3D( ptr3_1++, ptr3_2++, N, M, 1 );

    NoRegion = ConnectComponent (Mask_Out, tmpMask_1, N, M, L, MinSize, 
				 OBJECT, MaxNoReg, &MaxV_Label); 
    dilateLabel3D (tmpMask_1, Mask_Out, N, M, L);
    for ( k=0, ptr3_1=tmpMask_1, ptr3_2=Mask_Out; k<L; 
	  k++, ptr3_1++, ptr3_2++ ) { 
      dilateLabel3D (ptr3_2, ptr3_1, N, M, 1);
      dilateLabel3D (ptr3_1, ptr3_2, N, M, 1);
      dilateLabel3D (ptr3_2, ptr3_1, N, M, 1);
    }
    And2Mask (tmpMask_1, Binary_In, MaxV_Label, OBJECT, N, M, L);

    /* Identify BACKGROUND & fill the holes inside of OBJECT region */
    for ( k=0, ptr3_1=tmpMask_1; k<L; k++, ptr3_1++ ) 
      FillRegion (ptr3_1, N, M, 1, 1, 1, 0, BACKGROUND, NonObject);

    TakeLabel( tmpMask_1,
	       (unsigned char) NonObject,
	       (unsigned char) BACKGROUND,
	       N, M, L );

    /* Remove very small OBJECT regions due to noise */
    for ( k=0, ptr3_1=tmpMask_1, ptr3_2=Mask_Out; k<L; 
	  k++, ptr3_1++, ptr3_2++ ) { 
      ConnectComponent (ptr3_1, ptr3_2, N, M, 1, MinSize, 
			OBJECT, MaxNoReg, &MaxV_Label);
      for ( i=0, ptr_1=Mask_Out[k][0]; i<N*M; i++, ptr_1++ ) {
	if ( *ptr_1 != BACKGROUND ) 
	  *ptr_1 = OBJECT;
      }
    }
  } /* *OuterFirst_YN = 'n' */



  /*------------------------------------------------------*/
  /*            Case 2 (OuterFirstYN == "y"):             */
  /*  Find the head volume in an MRI T1 or proton image   */
  /*------------------------------------------------------*/

  else if (*OuterFirst_YN == 'y') {

    dilate3D( Binary_In, tmpMask_1, N, M, L );
    erode3D( tmpMask_1, Mask_Out, N, M, L );

#if 0
    /* Commented out by MY */
    for (k = 0; k < L; k++) 
      memcpy( Mask_Out[k][0], Binary_In[k][0], 
	      N*M*sizeof(unsigned char) );
#endif

    ptr3_2 = Mask_Out;
    for (k = 0; k < L; k++) {
      FillRegion( ptr3_2++, N, M, 1, 1, 1, 0, BACKGROUND, NonObject );
    }
    TakeLabel( Mask_Out,
	       (unsigned char) NonObject,
	       (unsigned char) BACKGROUND,
	       N, M, L );

    NoRegion = ConnectComponent( Mask_Out, tmpMask_1, N, M, L, MinSize, 
				 OBJECT, MaxNoReg, &MaxV_Label );
    TakeLabel( tmpMask_1,
	       (unsigned char) MaxV_Label,
	       (unsigned char) OBJECT,
	       N, M, L );

    for ( k=0; k<L; k++ ) {
      ptr_1 = tmpMask_1[k][0];
      ptr_2 = Mask_Out[k][0];
      for (i = N * M; i > 0; i--) {
	if ( *ptr_1 != BACKGROUND ) 
	  *ptr_2 = OBJECT;
	else 
	  *ptr_2 = BACKGROUND;
	ptr_1++;
	ptr_2++;
      }
    }
  } /* *OuterFirst_YN == 'y' */

  /*------------------------------------------------------*/
  /*      Case 3 (Some other value of OuterFirstYN        */
  /*             Generate an error message.               */
  /*------------------------------------------------------*/
  else {
    fprintf( stderr, "ExtractVOI() called with unexpected OuterFirst_YN value!\n" );
    fprintf( stderr, "Expect y or n, got %c.\n", *OuterFirst_YN );
    exit( 1 );
  }

  /*--------------------- End of cases ---------------------*/

#if 0
  /*      temporary for testing  --- begin  */
  f_img = alloc_img (N, M);
  for ( k=0; k<L; k++ ) { 
    sprintf(str, "%02.2d", k+1);
    ptr_f = f_img[0];
    ptr_1 = tmpMask_1[k][0];
    for ( i=0; i<N*M; i++, ptr_f++, ptr_1++ ) {
      *ptr_f = (float) *ptr_1;
    }
    put_matrix (str, f_img, N, M, 1); 
  } 
  free_img (f_img);
  /*      temporary for testing  --- end  */
#endif

  /* Okay, now clean up and exit */
  free_3Dchar (tmpMask_1, L);
}





/*--------------------------------------------------------------------------
ExtractHead - Create mask for head MRI T1 image

This function computes a mask for the head volume within an MRI T1 or
proton volume image.  In this images the scalp and skull have
brightness similar to the brain parenchyma; the algorithm is tuned to
match and does not attempt to separate brain from scalp.

The algorithm used will sometimes label the interior of the eyeballs
and some regions connected to the ear canal as external, since they
are dark and sometimes visually connected to the outside.  Many
segmentation algorithms can evade this problem by processing only
slices above the erriferous regions.

This function is equivalent to the old ExtractVOI function with
OuterFirst_YN set to "y", which is now deprecated.

The binary image input is expected to consist of pixel values OBJECT
and BACKGROUND only.  The output mask image is constructed to have the
value OBJECT where the desired image is present and BACKGROUND
elsewhere.
--------------------------------------------------------------------------*/

void
ExtractHead(
   unsigned char *** Binary_In,   /* Binary image to examine */
   unsigned char *** Mask_Out,    /* Output mask for head or brain */
   int N,                         /* Dimensions of the image */
   int M,
   int L)
{
    int           i, k, NonObject;
    int           NoRegion = 0, MaxNoReg = 253, MinSize = 5, MaxV_Label;
    unsigned char ***tmpMask_1, *ptmp, *pmsk;


    tmpMask_1 = alloc_3Dchar( N, M, L );
    NonObject = MaxNoReg + 1;

    /* Do morphological close to remove small holes and nicks */
    dilate3D( Binary_In, tmpMask_1, N, M, L );
    erode3D ( tmpMask_1, Mask_Out,  N, M, L );

    /* Find a connected BACKGROUND region in each slice containing 
       the point (1,1) and label it as NonObject.  */
    for (k = 0; k < L; k++) {
        FillRegion( &Mask_Out[k], N, M, 1,  1, 1, 0, BACKGROUND, NonObject );
    }

    /* Convert every NonObject pixel to BACKGROUND, and every other
       pixel to OBJECT. */
    TakeLabel( Mask_Out, NonObject, BACKGROUND, N, M, L );

    /* ??? Find the connected components of type OBJECT, labelling
       each with a new region number in the output image tmpMask_1.  
       The largest connected component is assigned the maximum region 
       number, which is also placed in MaxV_Label.  */
    NoRegion = ConnectComponent( Mask_Out, tmpMask_1, N, M, L, MinSize, 
				 OBJECT, MaxNoReg, &MaxV_Label );

    /* Change every pixel == MaxV_Label to OBJECT, and all the others
       to BACKGROUND. */
    TakeLabel( tmpMask_1, MaxV_Label, OBJECT, N, M, L );

    /* Change every non-BACKGROUND pixel to OBJECT */
    for ( k=0; k<L; k++ ) {
        ptmp = tmpMask_1[k][0];
        pmsk = Mask_Out[k][0];
        for (i = N * M; i > 0; i--) {
	    if ( *ptmp++ != BACKGROUND )   *pmsk++ = OBJECT;
	    else                	   *pmsk++ = BACKGROUND;
        }
    }

    /* Okay, now clean up and exit */
    free_3Dchar (tmpMask_1, L);
}




/*--------------------------------------------------------------------------
InnerMask - Remove outer layer to extract inner VOI

This function apparently further processes an MRI T1 or proton image masked
head image generated by ExtractVOI to obtain a masked brain image.

It is almost certainly tuned to a specific task, although the details of the
tuning are still obscure.


Old comment (???): UpperT is on for input binary image.
--------------------------------------------------------------------------*/


void
InnerMask(
   unsigned char *** Binary_In,  /* (in) Input binary image to process */
   unsigned char *** Mask_IO,    /* (in) Initial mask image ??? */
   unsigned char *** Mask_Out,   /* (out) Improved mask image ??? */
   int N,                        /* Dimensions of the image */
   int M,
   int L )
{

  int	i, k;
  int	NoRegion = 0; 
  int   MaxNoReg = 253;
  int   MinSize = 5;
  int	MaxV_Label, DistT;
  int	*MaxDist;
  unsigned char	***tmpMask;
  unsigned char	*ptr_1, *ptr_2;
  unsigned char ***ptr3_1, ***ptr3_2;


  /* printf("InnerMask()\n"); */

  tmpMask = alloc_3Dchar( N, M, L );

  /* Compute Chamfer distance map according to the initial mask image
     Mask_IO and save in Mask_Out */
  DistMap (Mask_IO, Mask_Out, N, M, L);
  for ( k=0; k<L; k++ ) {
    memcpy (Mask_IO[k][0], Mask_Out[k][0], N*M*sizeof(unsigned char));
  }

  /* Mask the chamfer distance by the original binary image */
  for ( k=0; k<L; k++ ) {
    ptr_1 = Mask_Out[k][0];
    ptr_2 = Binary_In[k][0];
    for ( i=0; i<N*M; i++, ptr_1++, ptr_2++ ) {
      if ( *ptr_2 == BACKGROUND )
	*ptr_1 = BACKGROUND;
    }
  }
  if (0) put_volume_uchar("chamfer.img", Mask_Out, N, M, L);

  /*  Get a histogram and compute threshold for masked chamfer map */
  DistT = T_DistMap (Mask_Out, N, M, L);
  printf("Threshold for Chamfer map = %d\n", DistT);

  /*  Label each region & assign corresponding largest distance ??? */
  erode3D( Binary_In, Mask_Out, N, M, L );
  for ( k=0, ptr3_1=tmpMask, ptr3_2=Mask_Out; k<L; 
	k++, ptr3_1++, ptr3_2++ ) { 
    erode3D( ptr3_2, ptr3_1, N, M, 1 );
  }
  NoRegion = ConnectComponent (tmpMask, Mask_Out, N, M, L, MinSize, 
			       OBJECT, MaxNoReg, &MaxV_Label); 
  dilateLabel3D (Mask_Out, tmpMask, N, M, L);
  for ( k=0, ptr3_1=tmpMask, ptr3_2=Mask_Out; k<L; 
	k++, ptr3_1++, ptr3_2++ ) { 
    dilateLabel3D( ptr3_1, ptr3_2, N, M, 1 ); 
    dilateLabel3D( ptr3_2, ptr3_1, N, M, 1 );
    /*	dilateLabel3D( ptr3_1, ptr3_2, N, M, 1 ); */
  } 
  /*    for ( k=0; k<L; k++ ) 
	memcpy (tmpMask[k][0], Mask_Out[k][0], N*M*sizeof(unsigned char)); */

  /* ??? */
  MaxDist = (int *) mallock(NoRegion*sizeof(int));
  for ( k=0; k<NoRegion; k++ )
    MaxDist[k] = 0;
  for ( k=0; k<L; k++ ) {
    ptr_1 = tmpMask[k][0];
    ptr_2 = Mask_IO[k][0];
    for ( i=0; i<N*M; i++, ptr_1++, ptr_2++ ) {
      if ( *ptr_1>0 && *ptr_2>MaxDist[(*ptr_1)-1] ) 
	MaxDist[(*ptr_1)-1] = *ptr_2;
    }
  }

  /*  Assign OBJECT if distance greater than DistT */
  for ( k=0; k<L; k++ ) {
    for ( i=0, ptr_1=tmpMask[k][0], ptr_2=Mask_Out[k][0]; 
	  i<N*M; i++, ptr_1++, ptr_2++ ) {
      if ( *ptr_1>0 && MaxDist[(*ptr_1)-1]>DistT) 
	*ptr_2 = OBJECT;
      else 
	*ptr_2 = BACKGROUND;
    }
  }

  /* Save result for debugging */
  if (0) put_volume_uchar("inner.img", Mask_Out, N, M, L);

  /* Clean up and exit */
  free (MaxDist);
  free_3Dchar (tmpMask, L);
}



/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*   It refines the extracted inner VOI from InnerMask()                    */
/*--------------------------------------------------------------------------*/
/* UpperT is off for reference binary    */

void
CleanUpMask( unsigned char *** Binary_IO,
	     unsigned char *** Binary_Ref,
	     int N,
	     int M,
	     int L )
{
  int           i, k, NonObject;
  int           MaxNoReg = 253, MinSize = 5;
  int		MaxV_Label, HalfL;
  unsigned char ***tmpMask, *ptr_1, *ptr_2, ***ptr3_1;


  printf("CleanUpMask()\n");

  tmpMask = alloc_3Dchar( N, M, 1 );
  NonObject = MaxNoReg + 1;

/*--------------------------------------------------------------------------*/
/*  Recover brain shape & select mask reference & propagate                 */
/*--------------------------------------------------------------------------*/

  for ( k=0; k<L; k++ ) {
    ptr_1 = Binary_IO[k][0];
    ptr_2 = Binary_Ref[k][0];
    for ( i=0; i<N*M; i++, ptr_1++, ptr_2++ ) {
      if ( *ptr_2 == BACKGROUND )
	*ptr_1 = BACKGROUND;
    }
  }

  HalfL = L/2;
  if ( HalfL != 0 ) { 
    ptr3_1 = Binary_IO + HalfL;
    ConnectComponent (ptr3_1, tmpMask, N, M, 1, MinSize, 
		      OBJECT, MaxNoReg, &MaxV_Label);
    TakeLabel( tmpMask,
	       (unsigned char) MaxV_Label,
	       (unsigned char) OBJECT,
	       N, M, 1 ); 
    for ( k=HalfL; k>0; k-- ) {
      /*	    printf("%d-th mask propagation ....\n", (k-1)); */
      MaskPropagate (Binary_IO[k-1], tmpMask[0], N, M); 
      memcpy (tmpMask[0][0], Binary_IO[k-1][0], 
	      N*M*sizeof(unsigned char));
    }

    memcpy (tmpMask[0][0], Binary_IO[HalfL-1][0], 
	    N*M*sizeof(unsigned char));
    for ( k=HalfL-1; k<L-1; k++ ) {
      MaskPropagate (Binary_IO[k+1], tmpMask[0], N, M);
      memcpy (tmpMask[0][0], Binary_IO[k+1][0], 
	      N*M*sizeof(unsigned char));
    }
    /*      printf("finish Mask Propagation ....\n"); */
  }
    
/*--------------------------------------------------------------------------*/
/*  Identity BACKGROUND & fill holes inside of OBJECT region                */
/*--------------------------------------------------------------------------*/

#if 0
/*      temporary for testing  --- begin  */
  f_img = alloc_img (N, M);
  for ( k=0; k<L; k++ ) { 
    sprintf(str, "%02.2d", k+1);
    ptr_f = f_img[0];
    ptr_1 = Binary_IO[k][0];
    for ( i=0; i<N*M; i++, ptr_f++, ptr_1++ ) {
      *ptr_f = (float) *ptr_1;
    }
    if ( k==40 || k==41 || k==42 )
      put_matrix (str, f_img, N, M, 1); 
  } 
  free_img (f_img);
  /*      temporary for testing  --- end  */
#endif

  for ( k=0, ptr3_1=Binary_IO; k<L; k++, ptr3_1++ ) 
    FillRegion (ptr3_1, N, M, 1, 1, 1, 0, BACKGROUND, NonObject);
  TakeLabel( Binary_IO,
	     (unsigned char) NonObject,
	     (unsigned char) BACKGROUND,
	     N, M, L );

  free_3Dchar (tmpMask, 1);
}


/*--------------------------------------------------------------------------*/
/* purpose:                                                                 */
/*   It refines the test Mask using the reference mask.                     */
/*--------------------------------------------------------------------------*/

int
MaskPropagate( unsigned char ** Binary_IO,
	       unsigned char ** Mask_Ref,
	       int N,
	       int M )
{

    int           NoRegion=0, MaxNoReg=253, MinSize=5, NoObj,  
                  MaxV_Label, i;
    float         *IndexV, MaxIndexV, OverlapOld, OverlapNew;
    int           *ObjSize=0, *InterObjSize=0, UnionNo, InterNo;
    unsigned char ***tmpMask, *ptr_c1, *ptr_c2;


    tmpMask = alloc_3Dchar (N, M, 1);

    NoRegion = ConnectComponent (&Binary_IO, tmpMask, N, M, 1, MinSize, 
				 OBJECT, MaxNoReg, &MaxV_Label); 
    if ( NoRegion == 0 ) 
	return 0;

    IndexV = (float *) mallock(NoRegion*sizeof(float));
    ObjSize = (int *) mallock(NoRegion*sizeof(int));
    InterObjSize = (int *) mallock(NoRegion*sizeof(int));
    for ( i=0; i<NoRegion; i++ ) {
	ObjSize[i] = 0;
	InterObjSize[i] = 0;
    }
/*--------------------------------------------------------------------------*/
/*  compute index value for each label                                      */
/*--------------------------------------------------------------------------*/
    ptr_c1 = tmpMask[0][0];
    ptr_c2 = Mask_Ref[0];
    for ( i=0; i<N*M; i++, ptr_c1++, ptr_c2++ ) {
	if ( *ptr_c1>0 ) {
	    ObjSize[(*ptr_c1)-1] ++;
	    if ( *ptr_c2 == OBJECT )
		InterObjSize[(*ptr_c1)-1] ++;
	}
    }
    for ( i=0; i<NoRegion; i++ ) {
	if ( ObjSize[i] == 0 ) 
	    fatal ("MaskPropagate: inconsistent labels");
/*	printf("%d-th InterObjSize, ObjSize = %d %d\n", (i+1), 
	       InterObjSize[i], ObjSize[i]); */
	IndexV[i] = (float) InterObjSize[i]/ (float) ObjSize[i];
    }
    free (ObjSize);
    free (InterObjSize);
/*--------------------------------------------------------------------------*/
/*  identify OBJECT                                                         */
/*--------------------------------------------------------------------------*/
    OverlapNew = 0.0;
    NoObj = 0;
    for ( i=0, ptr_c2=Binary_IO[0]; i<N*M; i++, ptr_c2++ )
	*ptr_c2 = 0;

    do { 
	NoObj ++;
	OverlapOld = OverlapNew;

	MaxIndexV = IndexV[0];
	MaxV_Label = 1;
	for ( i=0; i<NoRegion; i++ ) {
	    if ( IndexV[i] > MaxIndexV ) {
		MaxIndexV = IndexV[i];
		MaxV_Label = i+1;
	    }
	}
	
	if ( MaxIndexV != 0.0 ) {
	    if ( MaxV_Label == 0 )
		fatal ("MaskPropagate: no OBJECT");
	    IndexV[MaxV_Label-1] = 0;

	    for ( i=0, ptr_c1=tmpMask[0][0], ptr_c2=Binary_IO[0];
		  i<N*M; i++, ptr_c1++, ptr_c2++ ) {
		if ( *ptr_c1 == MaxV_Label ) {
		    *ptr_c2 = OBJECT;
		}
	    }
	    
	    UnionNo = InterNo = 0;
	    ptr_c1 = Mask_Ref[0];
	    ptr_c2 = Binary_IO[0];
	    for ( i=0; i<N*M; i++, ptr_c1++, ptr_c2++ ) {
		if ( *ptr_c1 == OBJECT && *ptr_c2 == OBJECT ) {
		    InterNo ++;
		    UnionNo ++;
		}
		else if ( *ptr_c1 == OBJECT || *ptr_c2 == OBJECT )
		    UnionNo ++;
	    }
/*	    printf("%d-th: InterNo, UnioNo = %d %d \n", NoObj, 
		   InterNo, UnionNo); */
	    OverlapNew = (float) InterNo/ (float) UnionNo;

	    if ( OverlapNew < OverlapOld ) { 
		for ( i=0, ptr_c1=tmpMask[0][0], ptr_c2=Binary_IO[0];
		     i<N*M; i++, ptr_c1++, ptr_c2++ ) {
		    if ( *ptr_c1 == MaxV_Label ) {
			*ptr_c2 = BACKGROUND;
		    }
		}
	    } 
	}
	else 
	    OverlapNew = OverlapOld - 1.0;
    } while ( OverlapNew >= OverlapOld && NoObj < 20 );
		    
    free (IndexV);
    free_3Dchar (tmpMask, 1);
    return MaxV_Label;
}


/*--------------------------------------------------------------------------
DistMap - Compute 2D chamfer distance in a 3D image

Given a 3D image, this function computes the chamfer distance within each
slice.  (See ChamferDist2D for a definition of chamfer distance.)  The
"object" is found using find_mask and is probably tuned to some specific
(but unknown) task.

FIXME:  This function converts its input from uchar to float, computes
in float, and converts its output from float back to uchar.  Are all
these conversions truly necessary?
--------------------------------------------------------------------------*/

void
DistMap(
   unsigned char *** Binary_In,    /* Input binary image */
   unsigned char *** Map_Out,      /* Output distance image */
   int N,                          /* Dimensions of the image */
   int M,
   int L )
{
    float         **tmpBinary, **tmpMap, *ptr_f;
    int           i, k;
    int           ContourSize;
    unsigned char *ptr_c;


    tmpBinary = alloc_img(N, M);
    tmpMap = alloc_img(N, M);

    for ( k=0; k<L; k++ ) {
	ptr_c = Binary_In[k][0];
	ptr_f = tmpMap[0];
	for ( i=0; i<N*M; i++, ptr_c++, ptr_f++ ) {
	    *ptr_f = *ptr_c;
	}
	ContourSize = find_mask (tmpMap, tmpBinary, N, M, "CONTOUR");
	printf("%d-th slice Contour Size = %d\n", (k+1), ContourSize);
	
	if ( ContourSize != 0 ) { 
	    ChamferDist2D (tmpBinary, tmpMap, N, M);
	    ptr_c = Map_Out[k][0];
	    ptr_f = tmpMap[0];
	    for ( i=0; i<N*M; i++, ptr_c++, ptr_f++ ) {
		*ptr_c = (*ptr_f + 0.5);
	    }
	}
	else {
	    ptr_c = Map_Out[k][0];
	    for ( i=0; i<N*M; i++, ptr_c++) {
		*ptr_c = 0;
	    }
	}
    }

    free_img (tmpBinary);
    free_img (tmpMap);
}



/*--------------------------------------------------------------------------
T_DistMap - Compute threshold for a 3D distance map

This function computes the histogram of a 3D distance map and uses it
to determine the optimal threshold for some purpose yet to be determined;
it presumably has something to do with finding the optimal distance inward
from the scalp at which to separate the brain and skull, but this has yet
to be verified.
--------------------------------------------------------------------------*/

int
T_DistMap(                     /* Returns the optimal threshold */
   unsigned char *** Map_In,   /* The distance map to be analyzed */
   int N,                      /* Dimensions of the distance map */
   int M,
   int L )
{
    float         *Histo;
    int           DistT;

    Histo = (float *) callock(HISTO_L, sizeof(float));
    histogram3D (Map_In, Histo, N, M, L);
    DistT = select_threshold (Histo);

    free (Histo);
    return DistT;
}


/*--------------------------------------------------------------------------
ChamferDist2D - Compute a 2D Chamfer distance map for an OBJECT image

Given a image containing pixel labels BACKGROUND, BOUNDARY, and OBJECT
(see definition.h for their values), this function computes for each
point the Chamfer distance to BACKGROUND.  The Chamfer distance at
each point is defined as the distance from that point to the nearest
background pixel, along one of the eight principal directions (E, NE,
N, NW, W, SW, S, SE).
--------------------------------------------------------------------------*/

void
ChamferDist2D( 
   float ** Label_In,        /* (in) Label image */
   float ** Map_Out,         /* (out) Distance map */
   int N,                    /* (in) Dimensions of the image */
   int M )
{
    float    mask_v1 = 1.0, mask_v2 = 1.414;
    float    *ptr_1, *ptr_2; 
    float    min_v, v_lr, v_tb, v_TlBr, v_TrBl;
    float    large_v = 1000.0;
    int      i, j;

    /* Set boundary pixels to distance 1.0, and background to huge */
    for ( i=0, ptr_1=Label_In[0], ptr_2=Map_Out[0]; i<N*M; 
	  i++, ptr_1++, ptr_2++ ) {
	if ( *ptr_1 == BOUNDARY ) {
	    *ptr_2 = 1.0;
	}
	else {
	    *ptr_2 = large_v;
	}
    }

    /* Pass the forward mask over entire image */
    for ( i=1; i<N-1; i++ ) {
	for ( j=1; j<M-1; j++ ) {
	    if ( Label_In[i][j] == OBJECT ) {
		v_lr = Map_Out[i][j-1] + mask_v1;
		v_tb = Map_Out[i-1][j] + mask_v1;
		v_TlBr = Map_Out[i-1][j-1] + mask_v2;
		v_TrBl = Map_Out[i-1][j+1] + mask_v2;
		min_v = min(v_TrBl, min(v_TlBr, min(v_lr, v_tb)));
		Map_Out[i][j] = min (Map_Out[i][j], min_v);
	    }
	}
    }
    /* Pass the backward mask over entire image */
    for ( i=N-1; i>0; i-- ) {
	for ( j=M-1; j>0; j-- ) {
	    if ( Label_In[i][j] == OBJECT ) {
		v_lr = Map_Out[i][j+1] + mask_v1;
		v_tb = Map_Out[i+1][j] + mask_v1;
		v_TlBr = Map_Out[i+1][j+1] + mask_v2;
		v_TrBl = Map_Out[i+1][j-1] + mask_v2;
		min_v = min(v_TrBl, min(v_TlBr, min(v_lr, v_tb)));
		Map_Out[i][j] = min (Map_Out[i][j], min_v);
	    }
	}
    }

    /* Set the background pixels to zero distance */
    for ( i=0, ptr_1=Map_Out[0]; i<N*M; i++, ptr_1++ ) {
	if ( *ptr_1 == large_v ) {
	    *ptr_1 = 0.0;
	}
    }
}
	


/*--------------------------------------------------------------------------
RegionGrow3D - Grow and label a connected volume from a seed point

Given a seed point within an image, this function finds the connected
volume of points that are connected to the seed point and bear
CurLabel; it then relabels these points as NewLabel.  The number of
points in the volume is returned.  If the given seed point is not
CurLabel, then nothing is done and zero is returned.

FIXME:  This function is implemented recursively and can overflow the
stack if the connected volume is too large.
--------------------------------------------------------------------------*/

int
RegionGrow3D(
   unsigned char ***LabImg,   /* Labelled image to modify */
   int N,                     /* Dimensions of the image */
   int M,
   int L,
   int Si,		      /* Location of Seed */
   int Sj,
   int Sk,
   int CurLabel,	      /* Current label to look for */
   int NewLabel )             /* New label to be assigned */
{
    int size=1;

    /* If the given point has the wrong label, just exit */
    if ( LabImg[Sk][Si][Sj] != CurLabel ) {	
	return 0;
    }

    /* Otherwise label the point */
    LabImg[Sk][Si][Sj] = (unsigned char) NewLabel;
    /*    printf("RegionGrow3D: size = %d (%d %d %d) %d->%d (%d)\n", 
	   size, Sk, Si, Sj, CurLabel, NewLabel, LabImg[Sk][Si][Sj]); */

    /*  and recurse on its 6 nearest neighbors */
    if ( Sk > 0 ) 
	size+= RegionGrow3D(LabImg, N, M, L, Si, Sj, Sk-1, CurLabel, NewLabel);
    if ( Sj > 0 ) 
	size+= RegionGrow3D(LabImg, N, M, L, Si, Sj-1, Sk, CurLabel, NewLabel);
    if ( Si > 0 ) 
	size+= RegionGrow3D(LabImg, N, M, L, Si-1, Sj, Sk, CurLabel, NewLabel);
    if ( Si < N-1 ) 
	size+= RegionGrow3D(LabImg, N, M, L, Si+1, Sj, Sk, CurLabel, NewLabel);
    if ( Sj < M-1 )
	size+= RegionGrow3D(LabImg, N, M, L, Si, Sj+1, Sk, CurLabel, NewLabel);
    if ( Sk < L-1 ) 
	size+= RegionGrow3D(LabImg, N, M, L, Si, Sj, Sk+1, CurLabel, NewLabel);
    
    /*    printf("RegionGrow3D: size = %d (%d %d %d) %d->%d (%d) ***\n", 
	   size, Sk, Si, Sj, CurLabel, NewLabel, LabImg[Sk][Si][Sj]); */
    return size;
}
	
