/* ------------------------------------------------------------------------ 
split_rev.c - Split dual-echo image and optionally reverse slice order

Programmer:  Michelle Yan
Revisions:   Paul Hughett                                   

Copyright (c) 1998-2000 Trustees of the University of Pennsylvania
See the file COPYING for license terms.


    The source code is written for handling the two-echo (T2 and proton)  
MR images. Assume that a given MRI is stored in the order of proton and  
T2, and from the lower part of a brain to the top of the brain. The code 
splits a two-echo MRI into two image files: one is for T2, the other for 
for proton. The order of slices in the both files are reversed, i.e.     
from the top of a brain to the lower part of the brain.                  
------------------------------------------------------------------------*/

static const char rcsid[] = 
"$Id: split_rev.c,v 1.9 2000/08/30 18:24:36 hughett Exp $ Release $Name: segm-0-90 $";

#include <ctype.h>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>

#include <unistd.h>
#include <sys/stat.h>

#include "definition.h"
#include "errm.h"
#include "image_io.h"
#include "memm.h"


int
main (
      int argc,
      char *argv[])
{
  /* Argument definitions and defaults */
    char  FileName[128];       /* Input filename. */
    int   EchoNum = 2;         /* Number of echoes */
    char  Reverse_YN[] = "n";  /* Reverse slice order? */
    int   L = 0;               /* Number of slices. 0 => compute from size */
    int   N = 256, M = 256;    /* Number of rows and columes. */
    int   ImgType = 2;         /* Type of image data.  1 => unsigned char, */
                               /* 2 = unsigned short, -2 => signed short */
                               /* 4 = signed int,     -4 => float */

    /* Local variable definitions  */
    FILE     *File_In=0, *File_T2=0, *File_Pr=0;
    char     *Ptr; 
    unsigned char *slice = 0;  /* Buffer sufficient for one slice */
    int      k, NameLen, PixLen=0;
    long SliceBytes;           /* Number of bytes per slice */
    long offset;               /* Start of current slice in input file */
    int L2;                    /* Computed number of slices in input file */
    struct stat in_stat;       /* Inode info for the input file */
    int start;                 /* Starting value for slice number */
    int step;                  /* Step to next slice */
    int islice;                /* Current slice number */
    int err;

    /* Case 1: Prompt for arguments */
    if (argc == 1) {
      printf("File name of a MR image: ");
      scanf("%s", FileName);
      printf("File name: %s \n", FileName);
      printf("Single/Dual-echo (1/2) & Reverse slice order or not (y/n): \n");
      scanf("%d %s", &EchoNum, Reverse_YN);
      printf("Number of echo & Reverse slice order ----> %d, %s \n", 
             EchoNum, Reverse_YN); 
      printf("Image type, number of slices, rows & columns :\n");
      scanf("%d %d %d %d", &ImgType, &L, &N, &M);
      printf("Type, Slices, Rows & Cols --> %d %d, %dX%d\n", ImgType, L, N, M);
    }

    /* Case 2: Take arguments from the command line */
    else {
      if (strlen(argv[1]) > 128-1)
        fatal ("split_rev: filename is too long");
      strcpy(FileName, argv[1]);
      if (argc > 2)   EchoNum = atoi(argv[2]);
      if (argc > 3)   Reverse_YN[0] = tolower(*argv[3]);
      if (argc > 4)   ImgType = atoi(argv[4]);
      if (argc > 5)   L = atoi(argv[5]);
      if (argc > 6)   N = atoi(argv[6]);
      if (argc > 7)   M = atoi(argv[7]);
    }

    /* Check some arguments for validity  */
    if (EchoNum != 1 && EchoNum != 2)
      fatal("split_rev: Number of echoes must be 1 or 2");
    if (Reverse_YN[0] != 'y' && Reverse_YN[0] != 'n')
      fatal("split_rev: Reverse must be y or n");
    if (L < 0 || L > 1024)
      fatal("split_rev: Unreasonable number of slices");
    if (N < 1 || N > 1024)
      fatal("split_rev: Unreasonable number of rows");
    if (M < 1 || M > 1024)
      fatal("split_rev: Unreasonable number of columns");

    /* Compute number of bytes per pixel.  The image type codes
        originally used here were carefully chosen such that the
        absolute value of the code yielded the length in bytes on the
        Sun Sparc, but this is not necessarily portable.  For
        consistency, we retain the old codes but force the item length
        to be as specified, regardless of the platform.  */
    if ( ImgType < 0)  ImgType = -ImgType;
    if ( ImgType == 0 || ImgType > 16 )
      fatal("split_rev: Invalid number of bytes per pixel");
    PixLen = ImgType;

    /* Compute number of slices from size of file */
    /* FIXME: Warn if number of slices given does not equal number of
       slices in the file? */
    SliceBytes = PixLen * N * M;
    err = stat(FileName, &in_stat);
    if (err != 0)
      fatal("split_rev: Unable to stat the input file");
    L2 = in_stat.st_size / SliceBytes;
    if ( L2*SliceBytes != in_stat.st_size )
      fatal("split_rev: File is not an integral number of slices");
    if (L == 0) {
      L = L2;
      if (L < 1 || L > 1024)
        fatal("split_rev: Unreasonable number of slices was computed");
    }
    
    /* Open input file and allocate memory for one slice */
    File_In = open_file(FileName, "r");
    slice = mallock(SliceBytes);

    /* Compute output file names and open the output files */
    /* FIXME: Exactly the last 4 chars are removed to eliminate the
       nominal ".img" extension.  This is not very robust. */
    NameLen = strlen(FileName) - 4;
    if ( (Ptr = calloc(NameLen+9, sizeof( char ))) == NULL ) {
	fatal("split_rev: not enough memory for a file name.");
    }

    if ( EchoNum == 2 ) { 
	strncpy (Ptr, FileName, NameLen);
	strcat (Ptr, "_t2.img");
	File_T2 = open_file(Ptr, "w");

	strncpy (Ptr, " ", NameLen+8);
	strncpy (Ptr, FileName, NameLen);
	strcat (Ptr, "_pr.img");
	File_Pr = open_file(Ptr, "w");
    }
    else if ( EchoNum == 1 ) {
	strncpy (Ptr, FileName, NameLen);
	strcat (Ptr, "_rev.img");
	File_T2 = open_file(Ptr, "w");
    }
    free(Ptr);

    /* Set coefficients for normal or reverse order */
    if ( Reverse_YN[0] == 'y' ) {
      start = L-1;
      step  = -1;
    }
    else {
      start = 0 ;
      step  = +1;
    }

    /* All is ready, now shuffle them bytes */
    for ( k = 0; k < L; k++ ) {
      islice = start + k*step;
      offset = islice*SliceBytes;
      fseek(File_In, offset, SEEK_SET);
      err = fread(slice, 1, SliceBytes, File_In);
      if (err != SliceBytes)
        fatal ("split_rev: Error reading from the input file");
      if (EchoNum == 2 && islice % EchoNum == 0) {
        err = fwrite(slice, 1, SliceBytes, File_Pr);
        if ( err != SliceBytes )
          fatal("split_rev: Error writing to first output file");
      }
      else {
        err = fwrite(slice, 1, SliceBytes, File_T2);
        if ( err != SliceBytes )
          fatal("split_rev: Error writing to second output file");
      }
    }

    /* Release memory and close files */
    fclose (File_In);
    free(slice);

    if ( EchoNum == 2 ) {
	fclose (File_Pr);
	fclose (File_T2);
    }
    else {
	fclose (File_T2);
    }
    return 0;
}    
  
    
    
    
